/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.3.8)                              #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@bayesact.ca   bayesact.ca                                           #
#  Date  : April 2021                                                                            #
#  Copyright (C) 2017-2021  Jesse Hoey                                                           #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
# Test main function. Demonstrates reading a simulation file, setting up the simulation and then #
# running the simulation in user-query mode or automatic and printing the average state          #
# this differs from testbayesact.c as it allows for a specification of a series of ineractions   #
# to be read in from a file.                                                                     #
#************************************************************************************************/

#include "inputparser.h"
#include "fvarsample.h"

// functions to do planning
#include "forvalue.h"
// to write out the images
#include "imagewriter.h"


// uncomment this line to have a small image of each agent fundamental and transient for each event saved to a
// png image with this prefix. 
//#define WRITE_IMAGES
char *image_out_file="../output/image";
pair_type image_type=EP;

// for disruptive actions in simulation (marked with '+' at the end of the action)
#define DISRUPT_NOISE 0.1

// jesse trying to get the xx state to work properly for the PD game
//#define PDFUDGE 1
// if 1, will ask the user for all 4 actions/perceptions relevant in each interaction
// only one of which is actually user selectable in a noise-free environment.
#define SPECIFY_ALL_FOUR 0


int main(int argc, char **argv)
{
  // maximum number of iterations, unless overridden by the events file
  int max_iterations=50;

  // whether to use the default action at each step, or to sample from the action distribution
  int default_action=0;

  // whether to allow resets in identity to be done
  // hard resets when all weights are 0 are always done
  int do_resets=1;

  // use emotions at all - default is 0
  // if this flag is set, then emotions will be displayed, recorded in csv
  // and used to weight samples
  // to have emotions displayed, but not used to weight samples, look at
  // WEIGHT_AGENT_EMOTION and WEIGHT_CLIENT_EMOTION flags in fvarsample.h
  int use_emotions=0;
  // wether to use emotions during planning or not
  // that is, will emotions be chosen and sent for manipulative purposes
  int emotion_planning=0;

  // the number of samples to use for planning - set to zero to not do any planning
  int num_plan_samples=0;
  // planning timeout
  int timeout=0;

  // amount of noise disruption to do 
  double disrupt_noise=DISRUPT_NOISE;
   
  // verbosity (needed if doing user mode, otherwise tracks fvarsample.h/VERBOSE
  int verbosity=0;
#ifdef VERBOSE
  verbosity=1;
#endif
  // whether to do the mirror of each dyad automatically
  // this will always be true if the mirror dyad exists, false if not
  // it is used for convenience below
  // setting this to zero, although possible, is not advised, as it means you are doing
  // a simulation in which agent A takes action on agent B, but agent B is not acted upon,
  // a seeming contradiction that may not be all that far-fetched, because A may be performing
  // his action in his own head (and perceiving the effects of his action), but not actually
  // be performing the action in real life. For example, A might fantasize about making love to his
  // neighbour, but the neighbour is not aware of what happened at all. I'm guessing though that what
  // happens in this case is that a new "fantasy B" agent is created that A keeps alongside his original
  // "B" agent (model). When fantasizing, A uses "fantasy B", while in real life, A uses "B". When A is not
  // able to make this distinction, then trouble may be brewing, for A's model of B will diverge from what
  // B's model of B is. 
  int doing_mirror_dyad=1;

  // whether compute and show client from agent's perspective for each dyad
  // if this is 1 then there are four things to be computed each round -- what
  // each agent predicts he will do, what he predict his client to do,
  // and what that client (as agent in the mirror dyad) predicts that he will do,
  // and what that client predicts the agent will do (who is client to him)
  // in any case, when actually picking and taking action, this is not relevant
  int doing_client_persp=1;
  
  // the xx state. Not used unless you have some external state (outside of the identities of the agents)
  int use_xx_state=0;
  xostate xx_observation[2];


  // this is the simulation file that describes what is to be done - see testsim.txt
  char *simfile="testsim.txt";

  // method for emotion and action calculation
  // can be AVERAGE or EXPECTATION
  //
  // AVERAGED method for emotions and actions computes the average fundamental and transient across all samples and
  // computes the optimal emotion and action based on that
  //
  // EXPECTATION method for emotions and actions computes the expectation for emotion and optimal action 
  // (the emotion and action for each sample is computed, and then weighted by the sample's weight and summed)
  //
  AEMeth ae_method = EXPECTATION;
  
  // ================================================================================================
  // basic parameters
  // ================================================================================================
  // gamma - strength of the somatic transform to use by default
  // this may be over-ridden by the dictionary files if those have variances specified
  double gamma=-1.0;

  //  ================================================================================================
  // other variables and storage
  //  ================================================================================================
  // for interact simulations
  double tau_out[VD];
  // TO_DO these should also be a property of each agent !! 
  double **tau_interact;
  double **f_interact;
  int *first_iteration;
  double dist[VD];
  double tmpf[VD];
  double summed_distance;
  double summed_squared_distance;
  // TO BE REMOVED - all instances where this is mentioned now commented out and can be removed
  // except for in the testing of fundamentals
  identity *agent_id, *client_id;
  
  // some buffers for behaviours 
  char input_buffer[MAX_LABEL_LENGTH];
  // the best behaviours to take on each iteration, for each dyad involved in the interaction, for each of AGENT/CLIENT
  behaviour best_behaviours[2][2];
  // the actua behaviours taken on each iteration, for each dyad involved in the interaction, for each of AGENT/CLIENT
  behaviour the_behaviours[2][2];


  // indices to keep track of which agent in a dyad acted in a given event
  // need to have these indices so we can easily assign NULL
  int acted[2][2];
  d_entity_type who_acted;
  
  double totalweight;
  int iteration=0;
  int num_potential_actions, num_potential_emotions, best_action_from_average;
  int dointeract[2][2];
  // planning stuff
  // the plan tree
  treeNode *planTree[2];
  planTree[0] = planTree[1] = NULL;


  
  // number of best actions and emotions to show (in VERBOSE mode and when printing to CSV)
  int num_to_show=5;
  double total_emot_prob=0.0;
  double total_beh_prob=0.0;
  // dy is a key index for this code - it will either be 0 and 1, and indicates
  // the dyad (primary or mirror) being processed for each event - the index of the dyad being processed is
  // therefore dyads[the_dyad[dy]].dyad_index
  // ag is the index 0,1=AGENT,CLIENT and refers to an actor in the dyad
  int gotit,dy,ag;
  int the_behaviour_index, the_emotion_index;
  int best_behaviour_index;
  int best_emotion_index;
  char best_behaviour_label[MAX_LABEL_LENGTH];
  char best_emotion_label[MAX_LABEL_LENGTH];

  int best_act, best_emot;
  behaviour * beh_list=NULL;
  emotion * emot_list=NULL;
  double *emot = (double *) malloc(SSD*sizeof(double));;
  for (int i=0; i<SSD; i++) 
    emot[i] = 0.0;

  // the best emotions to take on each iteration, for each dyad involved in the interaction, for each of AGENT/CLIENT
  emotion best_emotions[2][2];
  // the actual emotion taken on each iteration, for each dyad involved in the interaction, for each of AGENT/CLIENT
  emotion the_emotions[2][2];
  double emot_prob;
  // this needs to be a pair of dyads - the_dyad + dyads[j].mirror_dyad_index since we will have to loop over the two
  // the_dyad[0] is the primary dyad for a particular event - this is the dyad whose agent is the one that comes first in the event
  // specification line. dyad[1] is the mirror dyad (if specified, if not, it is never used).
  int the_dyad[2];
  the_dyad[0]=the_dyad[1]=0;
  samplerState *sstate;
  
  // for header lines in CSV
  // make this ultra long as it might have many columns
  // TODO: calculate this from (1) if the mirror dyad exists and (2) how many samples the user wants to see (max_num_show_samples)
  static int max_line_length=65535;
  char * csv_header_line;
  csv_header_line = (char *) malloc(max_line_length*sizeof(char));
  char buffer[max_line_length];
  char *tmp_header_line;
  tmp_header_line = (char *) malloc(max_line_length*sizeof(char));
  // start us off
  sprintf(tmp_header_line,"");
  sprintf(csv_header_line,"");

  // some stuff for csv printing
  char * csvlabels[2];
  csvlabels[0]="AGENT";
  csvlabels[1]="CLIENT";
  


  // set these up to add an output csv file
  char *outfile="testout.csv";
  char outfile_tmp[1024]="";
  int output_csv=0;
  FILE *fpout=NULL;
  FILE *fpout_tmp=NULL;
  int max_num_show_samples=-1;  // print no samples by default

  char usagestring[1024];
  sprintf(usagestring,"usage: bayesactsim [-e] [-ep] [-x] [-h] [-p num] [-t num] [-o filename] [-s num] [-aa num] [-ab num] [-ac num] [-ba num] [-bc num] [simfile]\n\t -e : use emotions (default no)\n\t -x : use additional x state and observations (default no)\n\t -d : use default action only (default no)\n\t -h : print usage\n\t -p <num> : use num samples for plan tree generation (default no plan tree)\n\t -t <num> : timeout (in seconds) for planning (default 0 or 1 rollout)\n\t -ep: plan over emotions as well (default no)\n\t -o <filename>: output file name for CSV\n\t -s <num>: num is max number of raw samples to output (if not specified, but -o used, assume none, if zero, assume all)\n\t -ga gamma value [0.0025]\n\t [other parameters like alpha, beta, are now in the SIM file] \n\t -dn [DISRUPT_NOISE]: disrupt noise (if + used in simfile)\n\t simfile: simulation description file\n");
  

  // ================================================================================================
  // command line arguments
  // ================================================================================================
  int thearg=1;
  if (!(thearg<argc)) {
    printf("%s",usagestring);
    exit(1);
  }
  while (thearg<argc) {
    if (strncmp(argv[thearg],"-ep",3)==0) {
      emotion_planning=1;
    } else if (strncmp(argv[thearg],"-av",3)==0) {
      ae_method=AVERAGE;
    } else if (strncmp(argv[thearg],"-ex",3)==0) {
      ae_method=EXPECTATION;
    } else if (strncmp(argv[thearg],"-e",2)==0) {
      use_emotions=1;
    } else if (strncmp(argv[thearg],"-r",2)==0) {
      do_resets=1;
    } else if (strncmp(argv[thearg],"-dn",3)==0) {
      thearg++;
      disrupt_noise=atof(argv[thearg]);
    } else if (strncmp(argv[thearg],"-d",2)==0) {
      default_action=1;
    } else if (strncmp(argv[thearg],"-x",2)==0) {
      use_xx_state=1;
    } else if (strncmp(argv[thearg],"-v",2)==0) {
      verbosity=1;
    } else if (strncmp(argv[thearg],"-o",2)==0) {
      thearg++;
      outfile=argv[thearg];
      output_csv=1;
    } else if (strncmp(argv[thearg],"-s",2)==0) {  
      thearg++;
      max_num_show_samples=atoi(argv[thearg]);  // if zero, will print all samples
    } else if (strncmp(argv[thearg],"-ga",3)==0) {
      thearg++;
      gamma=atof(argv[thearg]);
    } else if (strncmp(argv[thearg],"-h",2)==0) {
      printf("%s",usagestring);
      exit(1);
      //    } else if (strncmp(argv[thearg],"-n",2)==0) {
      //      thearg++;
      //      num_samples=atoi(argv[thearg]);
    } else if (strncmp(argv[thearg],"-t",2)==0) {
      thearg++;
      timeout=atoi(argv[thearg]);
    } else if (strncmp(argv[thearg],"-p",2)==0) {
      thearg++;
      num_plan_samples=atoi(argv[thearg]);
      ae_method=PLANTREE;
    } else {
      simfile = argv[thearg];
    }
    thearg++;
  }

  // ================================================================================================
  // open output CSV file if needed
  // ================================================================================================
  if (output_csv) {
    // we write all the csv to a tmp file
    // that will be copied into the csv file at the end
    // after writing the header
    strcat(outfile_tmp,outfile);
    strcat(outfile_tmp,"tmp");
    fpout=fopen(outfile_tmp,"w");
    if (!fpout) {
      fprintf(stderr,"Error opening output file %s\n",outfile_tmp);
      exit(1);
    }
  }


  if (verbosity) {
    printf("reading from %s\n", simfile);
    printf("number of plan samples %d\n",num_plan_samples);
    if (use_emotions)
      printf("using emotions\n");
    if (emotion_planning)
      printf("planning over emotions is on\n");
    if (default_action)
      printf("will use default actions only\n");
  }

  // ================================================================================================
  // agents and dictionaries 
  // ================================================================================================
  agentState * agents = (agentState *) malloc(MAX_NUM_AGENTS*sizeof(agentState));
  int num_agents;
  dictionary * dictionaries = (dictionary *) malloc(MAX_NUM_DICTIONARIES*sizeof(dictionary));
  int num_dictionaries=0;
  // for convenience the_actors[dy][ag] will represent the agent who is acting in the dy dyad for an
  // event (so there are two, the primary and the mirror), and who is acting as either ag=AGENT or ag=CLIENT 
  agentState *the_actors[2][2];
  
  // ================================================================================================
  // read the simulation input file - this initializes everything
  // including all agents, dictionaries, and interactions
  // a dyad represents a particular agent's interaction with one other agent
  // and contains (1) an index to each agent participating in this interaction (dyad) - as "agent_index" and "client_index"
  // (2) an index to the "mirror dyad" if exists (e.g. if this dyad - primary - is agents A-B, the mirror is B-A)
  // if the mirror dyad (B-A) does not exist, then the client (B) can still "act", but if we ask it to
  // act "optimally" (according to BayesACT or interact), it can only do so through A's model
  // ================================================================================================
  // we statically allocate a fixed memory for the dyad structures, to make our lives a bit easier.
  // but note that this only limits the size of the dyad list, not the overall memory used
  // (which depends largely on the number of samples- each dyad has a set of numsamples samples)
  dyad * dyads = (dyad *) malloc(MAX_NUM_INTERACTIONS*sizeof(dyad));
  int num_dyads;
  // if 0 (BAYESACT), default behaviours are taken, -- this is not used anymore as BayesACT can be specified in the EVENTS file
  // if 1 (USER), behaviours are queried from the user
  // if 2 (EVENTS), events file is being used
  // setting USER as default, will be read from sim file
  query_type query_user=USER;
  event * events = NULL;

  if (verbosity) 
    printf("setting up agents and dyads\n");

  // this function does all the work - reads in the simulation setup, initializes all dictionaries and dyads
  // dyads have indices agent_index and client_index refeingr to agents in the list of agents
  // dyads also have a "mirror_dyad" that is the index of the mirror dyad. This is -1 if not assigned
  // MEMORY for individual dyads is all set up in here
  if (setUpAgentsDyads(simfile, agents, &num_agents, dyads, &num_dyads, &query_user, &max_iterations, gamma,
		       dictionaries, &num_dictionaries, &events) < 0) {
    fprintf(stderr,"Error setting up agents and dyads\n");
    exit(1);
  }
  // must see output to guide the user
  if (query_user==USER) 
    verbosity=1;

  // ================================================================================================
  // some bookkeeping and settings for the simulation
  // this is a set-up phase 
  // ================================================================================================
  // some temp storage 
  //set up tau_interact and f_interact
  tau_interact=(double **) malloc(num_dyads*sizeof(double *));
  for (int j=0; j<num_dyads; j++) 
    tau_interact[j] = (double *) malloc(VD*sizeof(double));
  f_interact=(double **) malloc(num_dyads*sizeof(double *));
  for (int j=0; j<num_dyads; j++) 
    f_interact[j] = (double *) malloc(VD*sizeof(double));
  
  // and a flag for first iteration - because we need to do the interact calculation on thefirst iteration
  // I suppose this first-iteration-interact-calculation thing could be moved outside the main loop?
  // should be char or uint at least 
  first_iteration = (int *) malloc(num_dyads*sizeof(int));
  for (dy=0; dy<num_dyads; dy++) 
    first_iteration[dy]=1;
  
#ifdef PDFUDGE
  for (dy=0; dy<num_dyads; dy++) {
    sstate = (dyads+dy)->dsstate;
    for (int j=0; j<sstate->numsamples;  j++) {
      setXXstate(&(sstate->samples[j].xx),0,dy);
    }
  }
#endif

  // ================================================================================================
  // main simulation loop
  // a lot of this should probably be encapsulated and moved to fvarsample.c or to some file of helper functions
  // this proceeds as follows
  // the REPL loop here is PREL: Print, Read, Evaluate
  // 1. select a dyad - either from events file or from user.
  //    Should be a BAYESACT option as well that selects the optimal dyad to work as in Group Simulator
  // 2. print some information to CSV and stdout
  // 3. compute emotions and actions by average, expectation or plantree method according to ae_method
  // 4. query user or consult events file for which action/emotion to take
  // 5. simulate xx observation and query user, or consult events file for which xx observation to make
  // 6. display what is to be done
  // 7. if the next event is to take an interact optimal action (specified by ! in the events file)
  //    then swap in the optimal interact action for the next agent to act. this is done in fvarsample.c
  //    at the start of updateFullStateTwoActors...
  //    these optimal interact actions are normally computed in step 9b below, so on first iteration we must compute it separately
  //    this is done in this file -- see "if interact optimal is expected to be taken.."
  //    but probably should be moved to fvarsample.c as well...
  // 8. update the state -- this is the main sequential update function in fvarsample.c updateFullStateTwoActorsXXX
  //    see bayesactv2.pdf for a description of how this function works.
  // 9. compute interact updates
  //    9a. compare action suggested by interact with the actual one taken in step 4.
  //    9b. based on action last taken by interact, update the interact state and compute the next actions to take
  //    NOTE: -step 9a is for when we run two BayesACT agents (B and I) against each other, but "force" one (I) to take the one that an
  //                   interact simulation with *fixed identities* would take at this stage of the simulation.
  //                   The other agent (B) takes BayesACT optimal actions. So far as agent B is concerned, it's interacting with
  //                   an interact agent. This allows us to see if it can "learn" what the optimal identities are. 
  //                   Note that is you do this with two dyads (including the mirror dyad of the first dyad), then the primary agent
  //                   of the mirror dyad is being "forced" to take Interact optimal actions, but is otherwise a BayesACT agent, and
  //                   so will also learn its own identity and the identity of the other agent. 
  //          -step 9b is updating that interact transient based on the behaviours actually taken and then computing the optimal
  //                   *next* actions to take that may be used in step 7 the next time an event for this dyad that has a "!" (interact optimal)
  //                   is encountered. 
  // 10. clean up
  // ================================================================================================
  while (query_user==USER || iteration < max_iterations) {
    //**********************************************************************************************************************
    // select a dyad to simulate next
    // this is either selected by the user, or is an "interaction" in an events file (a single line)
    // so this could probably be removed. 
    //**********************************************************************************************************************
    if (verbosity) {
      printf("available dyads: \n");
      for (int j=0; j<num_dyads; j++) 
	printf("%d: %s - %s (mirror %d)\n",j,agents[dyads[j].agent_index].name,agents[dyads[j].client_index].name,dyads[j].mirror_dyad_index);
      printf("\t select the one corresponding to the actor\n\t e.g. if A acted on B, choose dyad i: A - B :\n");
    }

    if (query_user==USER) {
      // ask the user which dyad to simulate next
      printf("select dyad to simulate (enter for same as last time %d): ",the_dyad[0]);
      the_dyad[0]=-1;
      while (the_dyad[0] < 0 || the_dyad[0] >=  num_dyads) {
	fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	// remove trailing newline
	input_buffer[strcspn(input_buffer, "\n")] = 0;
	if (strlen(input_buffer)!=0) {
	  the_dyad[0]=atoi(input_buffer);
	  if (the_dyad[0] < 0 || the_dyad[0]>num_dyads)
	    the_dyad[0]=-1;  // try again
	} else {
	  the_dyad[0]=0;
	}
      }
    } else if (query_user==EVENTS) {
      // or, get the event from the events file
      // always do the DYAD whose primary agent is the first listed in the events file
      // the first dyad is specified by the first agent in the event (in the events file)
      the_dyad[0] = events[iteration].dyad_index;
    }
    the_dyad[1] = dyads[the_dyad[0]].mirror_dyad_index;
    if (verbosity) {
      printf("the dyad selected : %d\n",the_dyad[0]);
      printf("mirror dyad       : %d\n",the_dyad[1]);
    }
    // I used to set the_dyad[1]=the_dyad[0] here but I should not need the_dyad[1] if that mirror dyad does not exist
    if (doing_mirror_dyad && the_dyad[1] < 0) 
      doing_mirror_dyad=0;

    // the agent and client in each dyad for this event
    // for convenience the_actors[dy][ag] will represent the agent who is acting in the dy dyad for an
    // event (so there are two, the primary and the mirror), and who is acting as either ag=AGENT or ag=CLIENT 
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      the_actors[dy][ID_AGENT] = agents+dyads[the_dyad[dy]].agent_index;
      the_actors[dy][ID_CLIENT] = agents+dyads[the_dyad[dy]].client_index;
    }
    //**********************************************************************************************************************
    // print information about each dyad to the CSV
    //**********************************************************************************************************************
    if (output_csv) {
      fprintf(fpout,"%d,%d,%d,",iteration,the_dyad[0],the_dyad[1]);
      if (iteration==0) {
	sprintf(tmp_header_line,"iteration,dyad 0, dyad 1,");
	strcat(csv_header_line,tmp_header_line);
      }
    }
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      if (verbosity) {
	printf("agent index %d client index %d\n",dyads[the_dyad[dy]].agent_index,dyads[the_dyad[dy]].client_index);
	printf("agent %s  client %s\n",the_actors[dy][ID_AGENT]->name,the_actors[dy][ID_CLIENT]->name);
	// really flag this shit up, we're getting moving finally, this is gonna be HUGE! 
	printf("***********************************************************************************************************\n");
	printf("***************************************************************** iteration %d ****************************\n",iteration);
	printf("***********************************************************************************************************\n");
      }
      // printSamplesCSV will only write the CSV out if fpout is not NULL,
      // and will do so only for the ID_AGENT
      // but also writes to stdout if VERBOSE is define-d
      sprintf(tmp_header_line,"");
      // returns tmp_header_line filled in with exactly what is in the csv on the first iteration
      printSamplesCSV(dyads+the_dyad[dy], the_actors[dy][ID_AGENT], the_actors[dy][ID_CLIENT], fpout, tmp_header_line, csvlabels[dy]);
      if (output_csv && iteration==0) 
	  strcat(csv_header_line,tmp_header_line);

      sstate = (dyads+the_dyad[dy])->dsstate;

      // compute and print deflections
      if (verbosity ) {
	printf("------------------------------------------------------------------\n");
	printf("deflections:\n");
      }
      // printDeflectionsCSV will only write the CSV out if fpout is not NULL,
      // also prints to the screen if fpout is NULL and VERBOSE is define-d
      sprintf(tmp_header_line,"");
      // returns tmp_header_line filled in with exactly what is in the csv on the first iteration
      printDeflectionsCSV(sstate, ae_method,fpout,tmp_header_line,csvlabels[dy]);
      if (output_csv && iteration==0) 
	strcat(csv_header_line,tmp_header_line);

#ifdef WRITE_IMAGES
      if (write_sample_image(image_out_file, the_dyad[dy], iteration, 640, 640, sstate->samples, sstate->numsamples,image_type)) {
	fprintf(stderr,"error in write_sample_image\n");
	exit(1);
      }
#endif
    }
    
    // **********************************************************************************************************************
    // compute emotions and actions by average, expectation or plantree method
    // this is set in ae_method which is specified at the start of this program
    // these are for information, and may be used in a simulation 
    // (using "*" in an events file or on the command line in USER mode)
    // **********************************************************************************************************************
    // in all of the following, we are computing dyad dy's agent's emotions and actions, for both itself and the client in its dyad
    // if not doing the mirror dyad, then only one dyad is computed. If not doing the client's perspective as well, then only one
    // agent is computed and displayed/saved.
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      sstate = (dyads+the_dyad[dy])->dsstate;
      // ag=0 is ID_AGENT
      // ag=1 is ID_CLIENT
      // ag represents the agent who acted, so really is redundant with who_acted...
      // but the way some of the functions are set up, some require ID_AGENT/ID_CLIENT
      // and some require BEH_AGENT/BEH_CLIENT  -- one day I am going to clean this mess up...
      for (ag=0; ag<(1+doing_client_persp); ag++) {
	// who_acts = BEH_AGENT if ag=0
	// who_acts = BEH_CLIENT if ag=1
	who_acted=BEH_AGENT;
	if (ag==ID_CLIENT)
	  who_acted=BEH_CLIENT;
	if (ae_method == AVERAGE) {
	  //
	  // emotion
	  //
	  // compute the expected emotion based on the state - the expectation of emotion
	  // now this is folded into getNextEmotion - this option will be used if sstate->use_emotion=0
	  // and is in the function getNextMostLikelyEmotion() in fvarsample.c
	  
	  if (use_emotions) {  
	    if (computeEmotionFromAverage(dyads+the_dyad[dy], emot, ag)) {
	      fprintf(stderr,"Error computing emotion\n");
	      exit(1);
	    }
	    // look up labels -- output in index gives the index into the dictionary of the emotion
	    somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &(best_emotions[dy][ag].index), emot, the_actors[dy][ag]->entity_dicts[EMOTION]);
	  } else {
	    getNullEmotion(dyads+the_dyad[dy], the_actors[dy][ag], &(best_emotions[dy][ag].index), &emot_prob, ag); 
	  }
	  // get label
	  strcpy(best_emotions[dy][ag].label,the_actors[dy][ag]->entity_dicts[EMOTION]->theDictionary[best_emotions[dy][ag].index].label);

	  // print out
	  if (use_emotions) {
	    if (verbosity) {
	      printf("=====================================================================\n");
	      printf("emotion for %s (as predicted by %s):\n",the_actors[dy][ag]->name,the_actors[dy][ID_AGENT]->name);
	      for (int j=0; j<SSD; j++)
		printf("%4.3g ",emot[j]);
	      printf("  label: %s\n",best_emotions[dy][ag].label);
	    }
	    if (output_csv) 
	      printEmotionsCSV(fpout,iteration,csv_header_line,emot,best_emotions[dy][ag].label,csvlabels[dy],csvlabels[ag]);
	  }

	  
	  // 
	  // behaviour
	  //
	  // encapsulate in a function! 
	  // computeActionAverageMethod(sstate, dyads[the_dyad[dy]].dyad_index, , ag);
	  
	  best_action_from_average = getNextActionFromAverageState(dyads+the_dyad[dy], the_actors[dy][ag], who_acted); 
	  // store as the optimal (default) action to take
	  if (best_action_from_average >= 0)  {
	    best_behaviours[dy][ag].index = best_action_from_average;
	  } else {
	    best_behaviours[dy][ag].index = 0;
	    printf("something went wrong - weights to zero in getNextActionFromAverageState average state computation - will use %s\n",
		   the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[0].label);
	  }
	  // get label
	  strcpy(best_behaviours[dy][ag].label,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[best_action_from_average].label);

	  // print out
	  if (verbosity) {
	    printf("=====================================================================\n");
	    printf("Best behaviour for %s (as predicted by %s): %s\n",
		   the_actors[dy][ag]->name,the_actors[dy][ID_AGENT]->name,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ag].index].label);
	    printf("=====================================================================\n");
	  }
	  if (output_csv) {
	    if (iteration==0) {
	      sprintf(tmp_header_line,"%s: best behaviour for %s,",csvlabels[dy],csvlabels[ag]);
	      strcat(csv_header_line,tmp_header_line);
	    }
	    fprintf(fpout,"%s,",the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ag].index].label);
	  }
	} else if (ae_method == EXPECTATION) {
	  //
	  // emotion
	  //
	  // now we can get a full list of emotions that could be displayed - this uses a somatic transform per sample, and
	  // aggregates the results in a distribution over the denotative emotion state
	  // this function is a bit messy because it allocates emot_list, which is why a pointer is passed in,
	  // and which must then be freed in the calling function
	  // However, I'm going to move all this into fvarsample.c so it should be encapsulated.

	  num_potential_emotions = getCurrentEmotionList(dyads+the_dyad[dy], the_actors[dy][ag], &emot_list, ag); 
	  // look up EPA value for the best one
	  somaticTransform(CONNOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &emot_list[0].index, emot, the_actors[dy][ag]->entity_dicts[EMOTION]);

	  // store the most probable one for potential later usage 
	  total_emot_prob=0.0;
	  if (num_potential_emotions > 0) {
	    best_emotions[dy][ag].index=emot_list[0].index;
	    // best one is on top
	    strcpy(best_emotions[dy][ag].label,the_actors[dy][ag]->entity_dicts[EMOTION]->theDictionary[emot_list[0].index].label);
	    if (verbosity) {
	      printf("=====================================================================\n");
	      printf("got %d potential current emotions for %s (from perspective of %s), top %d are:\n",
		     num_potential_emotions,the_actors[dy][ag]->name,the_actors[dy][ID_AGENT]->name,num_to_show);
	      printf("rank\t*\t cum. prob\t label (probability)\n");
	      printf("____________________________________________________________________\n");
	      //show top 5 - emot_list is sorted on return from getNextAction
	      for (int i=0; i< MIN(num_to_show,num_potential_emotions); i++)  {
		total_emot_prob += emot_list[i].prob;
		printf("%d\t*\t%4.3g\t%s (%4.3g)\n",i+1,total_emot_prob,
		       the_actors[dy][ag]->entity_dicts[EMOTION]->theDictionary[emot_list[i].index].label,emot_list[i].prob);
	      }
	    }
	  } else {
	    if (verbosity)
	      printf("num potential emotions was zero!\n");
	  }
	  // this is identical to above in AVERAGE method
	  if (output_csv) 
	    printEmotionsCSV(fpout,iteration,csv_header_line,emot,best_emotions[dy][ag].label,csvlabels[dy],csvlabels[ag]);
	
	  // ew... see above in getCurrentEmotionList
	  if (num_potential_emotions > 0)  
	    free(emot_list);
	  emot_list=NULL;

	  //
	  // action
	  //
	  // again, beh_list may be allocated inside this function .. ew
	  num_potential_actions = getNextAction(dyads+the_dyad[dy], the_actors[dy][ag], &beh_list, who_acted);

	  if (num_potential_actions > 0)  {
	    // store best one as the default option for usage or suggestion as optimal/default
	    best_behaviours[dy][ag].index = beh_list[0].index;
	    strcpy(best_behaviours[dy][ag].label,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[beh_list[0].index].label);
	    // could here also do a random draw from the action list 
	  } else {
	    // means something is wrong somewhere -- probably a reset
	    // this is a total HACK! I set this to be zero just because the above returned no actions ... TO BE FIXED
	    best_behaviours[dy][ag].index = 0;
	    strcpy(best_behaviours[dy][ag].label,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[0].label);
	  }

	  if (verbosity) {
	    if (num_potential_actions > 0)  {
	      printf("=====================================================================\n");
	      printf("got %d potential actions for %s (from perspective of %s), top %d are\n",
		     num_potential_actions,the_actors[dy][ag]->name,the_actors[dy][ID_AGENT]->name,num_to_show);
	      printf("rank\t*\t cum. prob\t label (probability)\n");
	      printf("____________________________________________________________________\n");
	      
	      //show top 5 - beh_list is sorted on return from getNextAction
	      total_beh_prob = 0.0;
	      for (int i=0; i< MIN(num_to_show,num_potential_actions); i++)  {
		total_beh_prob += beh_list[i].prob;
		printf("%d\t*\t%4.3g\t%s (%4.3g)\n",i+1,total_beh_prob,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[beh_list[i].index].label,beh_list[i].prob);
	      }
	    }
	  }

	  if (output_csv) {
	    if (iteration==0) {
	      sprintf(tmp_header_line,"%s: best behaviour for %s,",csvlabels[dy],csvlabels[ag]);
	      strcat(csv_header_line,tmp_header_line);
	    }
	    fprintf(fpout,"\"");
	    for (int i=0; i< MIN(num_to_show,num_potential_actions); i++)   {
	      fprintf(fpout,"%s",the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[beh_list[i].index].label);
	      if (i+1<MIN(num_to_show,num_potential_actions))
		fprintf(fpout,",");
	    }
	    fprintf(fpout,"\",");
	  }
	  // ew... again...see getNextAction(..) above
	  if (num_potential_actions > 0)
	    free(beh_list);
	  beh_list=NULL;
	
	} else if (ae_method == PLANTREE) {
	  //
	  // PLANTREE method for computing emotions and actions
	  // this consults the planning tree 
	  //
	  // get the next emotions and actions with the planTree
	  if (num_plan_samples > 0) {
	    // don't do this if this iteration is for the mirror dyad and this doesn't exist
	    if (dy==0 || doing_mirror_dyad) {
	      if (verbosity) 
		printf("PLANNING: getting next action to take for %s\n",the_actors[dy][ID_AGENT]->name);
	      int a_sample_index;
	      if (planTree[dy] == NULL) {
		if (verbosity) 
		  printf("PLANNNING: initializing with %d samples\n",num_plan_samples);
		initializeTreeNode(num_plan_samples, &planTree[dy]);
		for (int i=0; i<num_plan_samples; i++) {
		  a_sample_index = drawSample(sstate->r, sstate->samples, sstate->numsamples);
		  memcpy(planTree[dy]->samples+i, sstate->samples+a_sample_index, sizeof(sample));
		}
	      }
	      seekNextAction(dyads+the_dyad[dy], the_actors[dy][ID_AGENT], ag, planTree[dy], num_plan_samples, &best_act, &best_emot, timeout);
	      if (verbosity) 
		printf("PLANNING: best action to take %d (%s) best emotion %d (%s) has value %g\n",
		       best_act, the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label,
		       best_emot, the_actors[dy][ID_AGENT]->entity_dicts[EMOTION]->theDictionary[best_emot].label, planTree[dy]->value);
	      
	      
	      // store as default options, plantree mode
	      best_emotions[dy][ag].index = best_emot;
	      strcpy(best_emotions[dy][ag].label,the_actors[dy][ID_AGENT]->entity_dicts[EMOTION]->theDictionary[best_emot].label);
	      
	      best_behaviours[dy][ag].index = best_act;
	      strcpy(best_behaviours[dy][ag].label,the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label);
	      
	      if (query_user == USER) {
		printf("see the plan tree y/[N]?");
		fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
		// remove trailing newline
		input_buffer[strcspn(input_buffer, "\n")] = 0;
	      } 
	      if (query_user == EVENTS || strlen(input_buffer)>0)  {
		printTree(planTree[dy], the_actors[dy][ID_AGENT]);
		//for pretty printing in tikz use:
		//printTreeTikZ(planTree[dy], the_actors[ID_AGENT]);
	      }
	      if (verbosity) 
		printf("number of nodes in planTree after planning: %d\n",countTreeNodes(planTree[dy]));

	    }  else {
	      // the mirror dyad doesn't exist, so we'll need to find the actions for the client in the agent's planTree
	      // disable this mode if mirror dyad does not exist ? 
	      getBestNextActionsEmotions(planTree[1-dy], &best_act, &best_emot, &(best_behaviours[dy][ag].index), &(best_emotions[dy][ag].index));
	      strcpy(best_emotions[dy][ag].label,the_actors[dy][ID_CLIENT]->entity_dicts[EMOTION]->theDictionary[best_emotions[dy][ag].index].label);
	      strcpy(best_behaviours[dy][ag].label,the_actors[dy][ID_CLIENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ag].index].label);
	      if (verbosity) {
		printf("found agent action %d and emotion %d which is %s and %s\n",best_act,best_emot,the_actors[dy][ID_CLIENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label,the_actors[dy][ID_AGENT]->entity_dicts[EMOTION]->theDictionary[best_emot].label);
		printf("found client action %s and emotion %s\n",best_behaviours[dy][ag].label, best_emotions[dy][ag].label);
	      }
	    }
	  }
	}
      }
    }


    //**********************************************************************************************************************
    // query user or consult events file for which action/emotion to take
    //**********************************************************************************************************************
    // iterate over agent/client and get possible actions for each,
    // either (but not both) may be not specified
    // either query user for action to take, flag as a BAYESACT optimal action
    // or read from events file
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      for (ag=0; ag<(1+doing_client_persp); ag++) {
	// will this event be done using the interact optimal action
	// this action is the one that an interact agent would take at this point in the simulation
	// this requires the interact agent to be updated every round, see below after BayesACT update
	// (for testing: to be removed eventually)
// ----------------------------------------------------------------------------------------------- start of INTERACT STUFF
#ifdef DO_INTERACT
	// default is not to do it
	dointeract[dy][ag]=0;
#endif
// ----------------------------------------------------------------------------------------------- end of INTERACT STUFF
	// who acted for this dyad (can only be one of the two)
	acted[dy][ag]=0;
	if (query_user==EVENTS) {
	  // the behaviour_index of the event is for the primary dyad, so when doing the mirror dyad
	  // we need to remember this is the *client* acting, so they have to be reversed
	  // if we are doing the mirror dyad, then we have to look at the behaviour and emotion
	  // specified in the CLIENT slot in the events file for the AGENT, and in teh
	  // AGENT slot in the events file for the client for this dyad.
	  if (dy<1)  {
	    the_behaviour_index=events[iteration].behaviour_index[ag];
	    best_behaviour_index=best_behaviours[dy][ag].index;
	    strcpy(best_behaviour_label,best_behaviours[dy][ag].label);
	  } else  {
	    the_behaviour_index=events[iteration].behaviour_index[1-ag];
	    //the_emotion_index=events[iteration].emotion_index[1-ag];
	    best_behaviour_index=best_behaviours[1-dy][1-ag].index;
	    strcpy(best_behaviour_label,best_behaviours[1-dy][1-ag].label);
	    //best_emotion_index=best_emotions[1-dy][1-ag].index;
	    //strcpy(best_emotion_label,best_emotions[1-dy][1-ag].label);
	  }	
	  the_emotion_index=events[iteration].emotion_index[ag];
	  best_emotion_index=best_emotions[dy][ag].index;
	  strcpy(best_emotion_label,best_emotions[dy][ag].label);
	  // from events file
	  if (the_behaviour_index >= 0) {
	    acted[dy][ag]=1;
	    the_behaviours[dy][ag].index = the_behaviour_index; 
	    strcpy(the_behaviours[dy][ag].label,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[the_behaviours[dy][ag].index].label);
	  } else if (the_behaviour_index == NO_ACT) {  // empty so this agent didn't act  
	    acted[dy][ag]=0;
	    the_behaviours[dy][ag].index = -1;
	    strcpy(the_behaviours[dy][ag].label,"none");
	  } else if (the_behaviour_index == BAYES_ACT) {  //its stored in best_behaviours from above
	    acted[dy][ag]=1;
	    // changed April 9th
	    //the_behaviours[dy][ag].index=best_behaviours[the_dy][the_ag].index;
	    //strcpy(the_behaviours[dy][ag].label,best_behaviours[the_dy][the_ag].label);
	    the_behaviours[dy][ag].index=best_behaviour_index; //best_behaviours[dy][ag].index;
	    strcpy(the_behaviours[dy][ag].label,best_behaviour_label);
	  } else if (the_behaviour_index == INTER_ACT) {  //default do interact action  (for testing: to be removed)
	    acted[dy][ag]=1;
// ----------------------------------------------------------------------------------------------- start of INTERACT STUFF
#ifdef DO_INTERACT
	    dointeract[dy][ag]=1;
#endif
// ----------------------------------------------------------------------------------------------- end of INTERACT STUFF
	    // this will be replaced below - need it here?
	    // temporarily swap in the BayesACT option for now
	    the_behaviours[dy][ag].index=best_behaviours[dy][ag].index;
	    strcpy(the_behaviours[dy][ag].label,best_behaviours[dy][ag].label);
	  }
	  // emotions may be read in, but not used (if using_emotions=0)
	  if (the_emotion_index >= 0) {
	    the_emotions[dy][ag].index = the_emotion_index;
	  } else if (the_emotion_index == NO_ACT) {
	    // not sure what to do here - no emotion expressed but we are "using emotions" - so I guess the emotion index should be the "no emotion" which is at position 0 
	    the_emotions[dy][ag].index = 0;
	  } else if (the_emotion_index == BAYES_ACT) {
	    the_emotions[dy][ag].index = best_emotion_index;
	  } else if (the_emotion_index == INTER_ACT) {
	    // not done yet... 
	    the_emotions[dy][ag].index = 0;
	  }
	  strcpy(the_emotions[dy][ag].label,the_actors[dy][ag]->entity_dicts[EMOTION]->theDictionary[the_emotions[dy][ag].index].label);
	} else if (query_user==USER) {
	  // or  query the user for which behaviour to take
	  // will only accept one agent acting
	  gotit=-1;
	  if (SPECIFY_ALL_FOUR || (dy==0 && ag==0)) {
	    while (gotit<0) {
	      // dyad zero, agent zero is the only one we need to query
	      // the others are "none" except for the client in the mirror dyad whose action
	      // will be same as the agent in the primary dyad
	      printf("for dyad %d\n",dyads[the_dyad[dy]].dyad_index);
	      printf("behaviour to take for %s acting on %s (enter for none):  ",the_actors[dy][ag]->name, the_actors[dy][1-ag]->name);
	      fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	      // remove trailing newline
	      input_buffer[strcspn(input_buffer, "\n")] = 0;
	      if (strlen(input_buffer)==0) {
		if (ag==1 && !acted[dy][0]) 
		  printf("you said %s didn't act, so %s must! try again\n",the_actors[dy][ag]->name,the_actors[dy][1-ag]->name);
		else 
		  gotit=1;
	      } else {
		strcpy(the_behaviours[dy][ag].label,input_buffer);
		// also assigns the_behaviours[dy][ag].index
		gotit=lookUp(&the_behaviours[dy][ag],the_actors[dy][ag]->entity_dicts[BEHAVIOUR]);
		if (gotit>=0) {
		  acted[dy][ag]=1;
		}
	      }
	    }	    
	  } else {
	    // awkwardly copy over the user input into the related dyads
	    // the other way to do this is to actually ask the user to specify all four actions being taken or/perceived
	    // so if A acted on B, then we see A's perception of A's action, A's perception of B's action (which is none)
	    // B's perception of A's action (which may be different than A's perception of A's action), and B's perception
	    // of B's action (which is none).  To make the code work this way, just remove the entire else part of this if statement.
	    // I added the SPECIFY_ALL_FOUR #define that you can add that will revert to this input method
	    // which allows a user to specify noisy channels, for example.
	    if (dy==0) {
	      strcpy(the_behaviours[dy][ag].label,"");
	    } else {
	      if (ag==0) 
		strcpy(the_behaviours[dy][ag].label,"");
	      else 
		strcpy(the_behaviours[dy][ag].label,the_behaviours[0][0].label);
	    }
	    gotit=lookUp(&the_behaviours[dy][ag],the_actors[dy][ag]->entity_dicts[BEHAVIOUR]);
	    if (gotit>=0) {
	      acted[dy][ag]=1;
	    }
	  }
	  
	  // query the user for which emotion to display
	  if (use_emotions) {
	    gotit = -1;
	    while (gotit<0) {
	      printf("emotion to display for %s (enter for %s):  ",the_actors[dy][ag]->name,best_emotions[dy][ag].label);
	      fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	      // remove trailing newline
	      input_buffer[strcspn(input_buffer, "\n")] = 0;
	      if (strlen(input_buffer)==0) {
		gotit=1;
	      } else {
		strcpy(best_emotions[dy][ag].label,input_buffer);
		gotit=lookUp(&the_emotions[dy][ag],the_actors[dy][ag]->entity_dicts[EMOTION]);
		if (gotit<0)  
		  strcpy(the_emotions[dy][ag].label,best_emotions[dy][ag].label);
	      }
	    }
	  }
	}
      }
    }
    //**********************************************************************************************************************
    // simulate xx observation and query user, or consult events file for which xx observation to make
    //**********************************************************************************************************************
    if (use_xx_state) {
      // the mirror dyad only gets an xx observation if it exists...
      for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
	if (query_user==USER) {
	  //  TODO: this should also be done as above  - compute expecation, etc, and have specified in events file -
	  gotit = -1;
	  while (gotit<0) {
	    printf("observation of xx for %s:  ",the_actors[dy][ID_AGENT]->name);
	    fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	    // remove trailing newline
	    input_buffer[strcspn(input_buffer, "\n")] = 0;
	    gotit = stringToXXObs(input_buffer,&xx_observation[dy]);
	  }
	  printRawXXObs(&xx_observation[dy]);
	} else if (query_user==EVENTS) {
	  char *the_x_obs;
	  if (dy==0) 
	    the_x_obs=events[iteration].agent_x_obs;
	  else 			      
	    the_x_obs=events[iteration].client_x_obs;
	  if (the_x_obs[0]=='*') {
	    sampleNextXObservation(sstate, the_behaviours[dy][ID_AGENT].index , the_behaviours[dy][ID_CLIENT].index,xx_observation+dy);
	  } else {
	    gotit = stringToXXObs(the_x_obs,&xx_observation[dy]);
	    if (!gotit) {
	      printf("bad x observation specified in events file: %s\n",the_x_obs);
	      exit(1);
	    }
	  }
	}
      }
    }
// ----------------------------------------------------------------------------------------------- start of INTERACT STUFF
#ifdef DO_INTERACT    
    //**********************************************************************************************************************
    // if interact optimal is expected to be taken, then we swap in the interact optimal.
    // the interact optimal is computed after the state update, below
    // on first turn, therefore, we have to compute it from scratch
    // this should be encapsulated in a function, and then can be removed in any agent that does not wish to make use of it
    //**********************************************************************************************************************
    

    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      agent_id=dyads[the_dyad[dy]].agent_ids;  // uses the first ones
      client_id=dyads[the_dyad[dy]].client_ids;
      sstate = (dyads+the_dyad[dy])->dsstate;
      // here we copy in the *client* action as computed by interact on the *last* round 
      // this is because I do this computation after the belief update for the next round, and I do it for the agent that did *not* act (in the last round).
      // in the first round, we have to get the best actions to take
      if (first_iteration[dyads[the_dyad[dy]].dyad_index])  {  // was: cheap hack
	// grab the client action from the interact state for the first iteration
	// we are filling in client_act at the start when agent_id faces off with client_id
	// this involves calling computeInteractMean(sstate, dyad_index, BEH_CLIENT, tau_interact);
	// with tau_interact zset up so it has agent = agent and client=client for this dyad
	first_iteration[dyads[the_dyad[dy]].dyad_index]=0;
	// actually don't have to do this, but I put it here anyways
	for (int i=0; i<VD; i++)
	  tau_interact[dyads[the_dyad[dy]].dyad_index][i]=0.0;
	
	for (int i=0; i<SSD; i++)
	  tau_interact[dyads[the_dyad[dy]].dyad_index][i]=gsl_vector_get(the_actors[dy][ID_AGENT]->entity_dicts[AGENT]->theDictionary[agent_id->index].epa,i);
	
	for (int i=0; i<SSD; i++)
	  tau_interact[dyads[the_dyad[dy]].dyad_index][i+2*SSD]=gsl_vector_get(the_actors[dy][ID_AGENT]->entity_dicts[CLIENT]->theDictionary[client_id->index].epa,i);
	
#ifdef VERYVERBOSE
	if (verbosity) {	
	  printf("Tau interact for agent, client indices: %d, %d: ",agent_id->index,client_id->index);
	  for (int i=0; i<VD; i++)
	    printf("%g ",tau_interact[dyads[the_dyad[dy]].dyad_index][i]);
	  printf("\n");
	}
#endif
	// tau_interact[dyads[the_dyad[dy]].dyad_index][i+SSD] (behaviour) doesn't matter here as we are extracting it on the first iteration
	// fundamental=transient as its the first iteration
	computeInteractMean(sstate, BEH_AGENT, tau_interact[dyads[the_dyad[dy]].dyad_index], tau_interact[dyads[the_dyad[dy]].dyad_index]);
	// the interact mean is stored in sstate->interact_mean as a VD-D vector with optimal agent,behaviour,client from interact
	// we copy out the agent behaviour here
	for (int i=0; i<SSD; i++) 
	  dyads[the_dyad[dy]].interact_beh[ID_AGENT][i]=gsl_vector_get(sstate->sentiParams->interact_mean,i+SSD);
	
	// now do the client
	computeInteractMean(sstate, BEH_CLIENT, tau_interact[dyads[the_dyad[dy]].dyad_index], tau_interact[dyads[the_dyad[dy]].dyad_index]);
	// the interact mean is stored in sstate->interact_mean as a VD-D vector with optimal agent,behaviour,client from interact
	// we copy out the client behaviour here
	for (int i=0; i<SSD; i++) 
	  dyads[the_dyad[dy]].interact_beh[ID_CLIENT][i]=gsl_vector_get(sstate->sentiParams->interact_mean,i+SSD);
      } // was: cheap hack
	// now we swap in the interact optimal, computed either directly above on the first iteration (was: cheap hack)
	// or after the BayesACT update below for later iterations
      double *the_interact_beh;
      for (ag=0; ag<(1+doing_client_persp); ag++) {
	if (dointeract[dy][ag] && acted[dy][ag])  {
	  // have to do this funky thing because if the *client* in the *mirror dyad* was the one that acted, we want to use
	  // the interact behaviour from *agent* in the primary dyad (this will be the actual action taken for the interact agent
	  // for the client! This is also done below when computing the updates to the interact state
	  // this is still somewhat gross and confusing, but all this interact stuff should eventually be removed anyways GO BAYESACT GO! 
	  // if it was the client in this dyad that acted, then we need to use the client dyad's interact best action for AGENT, which is 1-ag
	  // otherwise, if it was the agent that acted, we use the agent dyad's interact best action for AGENT
	  if (acted[dy][1] && doing_mirror_dyad)
	    the_interact_beh = dyads[the_dyad[1-dy]].interact_beh[1-ag];
	  else
	    the_interact_beh = dyads[the_dyad[dy]].interact_beh[ag];
	  
	  somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &(the_behaviours[dy][ag].index), 
			   the_interact_beh, the_actors[dy][ag]->entity_dicts[BEHAVIOUR]);
	  strcpy(the_behaviours[dy][ag].label,the_actors[dy][ag]->entity_dicts[BEHAVIOUR]->theDictionary[the_behaviours[dy][ag].index].label);
	  if (verbosity) 
	    printf("TAKING INTERACT ACTION for %s in dyad %d: %d %s\n", ((ag==0) ? "AGENT" : "CLIENT"),dy,the_behaviours[dy][ag].index,the_behaviours[dy][ag].label);
	}
      }
    }
#endif  //DO_INTERACT
// ----------------------------------------------------------------------------------------------- start of INTERACT STUFF
   
    
   
    //**********************************************************************************************************************
    // possibly add noise to the action taken by only one actor who acted, if specified as disrupt_action flag
    // uses DISRUPT_NOISE as the variance of Gaussian noise added, find this in fvarsample.h
    // only works if doing EVENTS method
    //**********************************************************************************************************************
    double theact[SSD];
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      sstate = (dyads+the_dyad[dy])->dsstate;
      if (query_user==EVENTS && disrupt_noise > 0.0 && events[iteration].disrupt_action[dy]) {
	// if this dyad is disrupted, it means actually only that the "observations" the primary agent of the primary agent
	// in the mirror dyad's behaviours are disrupted. So we only need apply noise to one agent, the client
	// ok so normally I would iterated over both ID_AGENT and ID_CLIENT here (0/1) but this would disrupt the agents OWN action
	// that it takes as well. So if this is specified in the events file, I guess it should happen... so the events file 
	for (ag=0; ag<(1+doing_client_persp); ag++) { // ag=ID_CLIENT;
	  // pick a random action instead - so this REALLY disrupts things
	  // only danger of this method is that it might be disruptive (might have the same EPA)
	  //the_behaviours[dy].index = random(1,size(behaviour_dictionary))
	  // the other way is to get the epa vector of the index, add noise,
	  // then reconvert to another index (which might be the same again)
	  // whichever agent acted, his behaviour is disrupted, be it the self or the other
	  if (the_behaviours[dy][ag].index>=0) {
	    if (verbosity) 
	      printf("DISRUPTING ACTION with noise %4.3g dyad %d agent %s current action %s\n",disrupt_noise,dy,the_actors[dy][ag]->name,the_behaviours[dy][ag].label);
	     for (int i=0; i<SSD; i++)
	      theact[i]=gsl_vector_get(the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[the_behaviours[dy][ag].index].epa,i)+
		gsl_ran_gaussian(sstate->r,disrupt_noise);

	    somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &the_behaviours[dy][ag].index, theact, the_actors[dy][ag]->entity_dicts[BEHAVIOUR]);
	    strcpy(the_behaviours[dy][ag].label,the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[the_behaviours[dy][ag].index].label);
	    if (verbosity) 
	      printf("                                        new action %s\n",the_behaviours[dy][ag].label);
	  }
	}
      }
    }
    

    
    //**********************************************************************************************************************
    // now display what is to be done
    //**********************************************************************************************************************

    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      sstate = (dyads+the_dyad[dy])->dsstate;
#ifdef VERYVERBOSE
      if (verbosity) {
	for (ag=0; ag<(1+doing_client_persp); ag++) {
	  printf("Behaviour taken: %s\n", the_behaviours[dy][ag].label);
	  if (use_emotions) 
	    printf("emotion displayed: %s\n",the_emotions[dy][ag].label);
	  if (use_xx_state) {
	    xxObsToString(input_buffer,&xx_observation[dy]);
	    printf("x observation for dyad %d agent %s is : %s\n",dy,the_actors[dy][ag]->name, input_buffer);
	  }
	}
      }
#endif
      if (output_csv) {
	if (iteration==0) {
	  sprintf(tmp_header_line,"%s: behaviour taken (AGENT),%s: behaviour taken (CLIENT),",csvlabels[dy],csvlabels[dy]);
	  strcat(csv_header_line,tmp_header_line);
	}
	if (acted[dy][ID_AGENT]) 
	  fprintf(fpout,"%s,",the_behaviours[dy][ID_AGENT].label);
	else 
	  fprintf(fpout,"none,");
	
	if (acted[dy][ID_CLIENT]) 
	  fprintf(fpout,"%s,",the_behaviours[dy][ID_CLIENT].label);
	else 
	  fprintf(fpout,"none,");
	
	if (iteration==0) {
	  sprintf(tmp_header_line,"%s: emotion shown (AGENT), %s: emotion shown (CLIENT),",csvlabels[dy],csvlabels[dy]);
	  strcat(csv_header_line,tmp_header_line);
	}
	if (use_emotions) {
	  fprintf(fpout,"%s,",the_emotions[dy][ID_AGENT].label);
	  fprintf(fpout,"%s,",the_emotions[dy][ID_CLIENT].label);
	} else {
	  fprintf(fpout,"none,none,");
	}
      }	  
    }

    //**********************************************************************************************************************
    // write out the variances of fundamental and transients
    // as well as all the samples, if max_num_show_samples>0
    // if max_num_show_samples<0, it prints none,
    // if max_num_show_samples==0, it prints all the samples
    //**********************************************************************************************************************
    if (output_csv) {
      for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
	sstate = (dyads+the_dyad[dy])->dsstate;
	// clear this temporary char buffer
	sprintf(tmp_header_line,"");
	if (printRawVariancesCSV(sstate, the_actors[dy][ID_AGENT], the_actors[dy][ID_CLIENT], fpout, tmp_header_line, csvlabels[dy])) 
	  printf("something wrong in printRawVariancesCSV - perhaps you are using weighted samples?");
	if (iteration==0) 
	  strcat(csv_header_line,tmp_header_line);
	// clear this temporary char buffer
	sprintf(tmp_header_line,"");
	printRawSamplesCSV(sstate, the_actors[dy][ID_AGENT], the_actors[dy][ID_CLIENT], fpout, tmp_header_line, max_num_show_samples, csvlabels[dy]);
	if (iteration==0) 
	  strcat(csv_header_line,tmp_header_line);
      }
    }
    
    //************************************************************************************************************
    // update the state
    // this does the bulk of the work of propagating fundamentals, transients, and X forward 
    //************************************************************************************************************
    if (verbosity) 
      printf("performing the update - choose the same dyad again to see the results\n");
    
    // don't want to update the second dyad if it doesn't exist.
    for (dy=0; dy < (1+doing_mirror_dyad) ; dy++) {
      sstate = (dyads+the_dyad[dy])->dsstate;
      if (verbosity) {
	if (use_emotions) {
	  printf("%s is updating and perceives him/herself doing [%s] and showing [%s] to %s, who is doing [%s] and showing [%s] back\n",
		 the_actors[dy][ID_AGENT]->name,(acted[dy][ID_AGENT] ? the_behaviours[dy][ID_AGENT].label : "[doesn't act]"),the_emotions[dy][ID_AGENT].label,
		 the_actors[dy][ID_CLIENT]->name,(acted[dy][ID_CLIENT] ? the_behaviours[dy][ID_CLIENT].label : "[doesn't act]"),the_emotions[dy][ID_CLIENT].label);
	} else {
	  printf("[no_emotions]  %s is updating and perceives him/herself doing [%s] to %s, who is doing [%s] back\n",
		 the_actors[dy][ID_AGENT]->name,(acted[dy][ID_AGENT] ? the_behaviours[dy][ID_AGENT].label : "[doesn't act]"),
		 the_actors[dy][ID_CLIENT]->name,(acted[dy][ID_CLIENT] ? the_behaviours[dy][ID_CLIENT].label : "[doesn't act]"));
	}
      }

      // actually update the state
      // there are four cases depending if emotions and xx-state are used or not
      if (use_emotions)  { 
	if (use_xx_state) {
	  // most general case
	  totalweight = updateFullStateTwoActorsWithEmotionWithXXObs(dyads+the_dyad[dy], the_actors[dy][ID_AGENT],
								     (acted[dy][ID_AGENT] ? &(the_behaviours[dy][ID_AGENT]) : NULL),
								     (acted[dy][ID_CLIENT] ? &(the_behaviours[dy][ID_CLIENT]) : NULL),
								     &xx_observation[dy],
								     &(the_emotions[dy][ID_AGENT]), &(the_emotions[dy][ID_CLIENT]));
	} else {
	  totalweight = updateFullStateTwoActorsWithEmotion(dyads+the_dyad[dy], the_actors[dy][ID_AGENT],
							    (acted[dy][ID_AGENT] ? &(the_behaviours[dy][ID_AGENT]) : NULL),
							    (acted[dy][ID_CLIENT] ? &(the_behaviours[dy][ID_CLIENT]) : NULL),
							    &(the_emotions[dy][ID_AGENT]), &(the_emotions[dy][ID_CLIENT]));
	}	
      } else {
	if (use_xx_state) {
	  totalweight = updateFullStateTwoActorsWithXXObs(dyads+the_dyad[dy],the_actors[dy][ID_AGENT],
							  (acted[dy][ID_AGENT] ? &(the_behaviours[dy][ID_AGENT]) : NULL),
							  (acted[dy][ID_CLIENT] ? &(the_behaviours[dy][ID_CLIENT]) : NULL), &xx_observation[dy]);
	} else {
	  totalweight = updateFullStateTwoActors(dyads+the_dyad[dy], the_actors[dy][ID_AGENT],
						 (acted[dy][ID_AGENT] ? &(the_behaviours[dy][ID_AGENT]) : NULL),
						 (acted[dy][ID_CLIENT] ? &(the_behaviours[dy][ID_CLIENT]) : NULL));
	}
      }
	
      // prune the planTree - if this fails then planTree will return NULL and
      // will be re-initialized on the next pass above
      if (num_plan_samples > 0) {
	pruneTree(&planTree[dy], &the_behaviours[dy][ID_AGENT], &(the_emotions[dy][ID_AGENT]), &(the_behaviours[dy][ID_CLIENT]), &(the_emotions[dy][ID_CLIENT]), &xx_observation[dy]);
	if (verbosity) {
	  printf("number of nodes in planTree after pruning: %d\n",countTreeNodes(planTree[dy]));
	  printTree(planTree[dy], the_actors[dy][ID_AGENT]);
	}
      }
      // output to CSV the number of resets -- I adyach this to sstate-> direction but really I guess it should be adyached to the dyad[] structure
      // however, the dyad[] structure is an abstact thing, whereas the number of resets is an implementation thing, so it feels like it belongs here instead
      if (output_csv) {
	if (iteration==0) {
	  sprintf(tmp_header_line,"%s: resets, %s: resets successful, %s: resets agent, %s: resets client, %s: resets both, %s: resets hard,",
		  csvlabels[dy],csvlabels[dy],csvlabels[dy],csvlabels[dy],csvlabels[dy],csvlabels[dy]);
	  strcat(csv_header_line,tmp_header_line);
	}
	fprintf(fpout,"%d,",sstate->num_resets);
	fprintf(fpout,"%d,",sstate->num_resets_successful);
	fprintf(fpout,"%d,",sstate->num_resets_agent);
	fprintf(fpout,"%d,",sstate->num_resets_client);
	fprintf(fpout,"%d,",sstate->num_resets_both);
	fprintf(fpout,"%d,",sstate->num_hard_resets);
      }
    }

    // ----------------------------------------------------------------------------------------------- start of INTERACT STUFF
#ifdef DO_INTERACT
    //*************************************************************************************************************
    // update the interact state with the behaviour that was actually taken
    //*************************************************************************************************************
    // first, compute distance from best actions suggested by BayesACT and the optimal action (Interact)
    // this  is saved in agent_interact_beh/client_interact_beh
    // this is for testing - to be removed
    // ONLY works for 2 dyads - e.g. testsim9.txt
    double agent_dist, client_dist;
    int agent_beh_index, client_beh_index;
    agent_dist=0;
    client_dist=0;
    // first, we compute distances between the actions actually taken and the interact optimal suggestion
    
    for (dy=0; dy<(1+doing_mirror_dyad); dy++) {
      sstate = (dyads+the_dyad[dy])->dsstate;
      // from ~HERE~ to ~THERE~ can be removed once the interact testing is done
      // look up label of action suggested by Interact for this event, for both agent and client  
      somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &agent_beh_index, 
		       dyads[the_dyad[dy]].interact_beh[ID_AGENT], the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]);
      somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad[dy]].dyad_index, &client_beh_index, 
		       dyads[the_dyad[dy]].interact_beh[ID_CLIENT], the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]);
      
#ifdef VERYVERBOSE
      if (verbosity) {
	printf("AGENT optimal interact action (%s) compared to the actual action taken (%s)\n",
	       the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[agent_beh_index].label,
	       the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ID_AGENT].index].label);
	printf("CLIENT optimal interact action (%s) compared to the actual action taken (%s)\n",
	       the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[client_beh_index].label,
	       the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ID_CLIENT].index].label);
      }
#endif
      // distances bewteen those two things (sets of actions)
      for (int i=0; i<SSD; i++) {
	agent_dist += pow((dyads[the_dyad[dy]].interact_beh[ID_AGENT][i]-gsl_vector_get(the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ID_AGENT].index].epa,i)),2);
	client_dist += pow((dyads[the_dyad[dy]].interact_beh[ID_CLIENT][i]-gsl_vector_get(the_actors[dy][ID_AGENT]->entity_dicts[BEHAVIOUR]->theDictionary[best_behaviours[dy][ID_CLIENT].index].epa,i)),2);
      }

#ifdef VERYVERBOSE
      if (verbosity) {
	printf("label match indicator and distance between Interact optimal (computed last round) and BayesACT best for dyad %d:\n",dy);
	printf("AGENT action: %d,%4.3g\n",(agent_beh_index==best_behaviours[dy][ID_AGENT].index),sqrt(agent_dist));
	printf("CLIENT action: %d,%4.3g\n",(client_beh_index==best_behaviours[dy][ID_CLIENT].index),sqrt(client_dist));
      }
#endif
      // ~THERE~
      //
      // now, we do the interact update for this dyad
      // the ids are always fixed here as what they were in the simulation file
      // however, since agents can have more than one identity, we have to pick one
      // for now, pick the first one, but should be pick the most likely one

      agent_id=dyads[the_dyad[dy]].agent_ids;  // uses the first ones
      client_id=dyads[the_dyad[dy]].client_ids;

      if (acted[dy][ID_AGENT]) {
	// update the ongoing monitoring of the interact state for the AGENT of this dyad, tau_interact[dyads[the_dyad[dy]].dyad_index]
	// if it was the AGENT that acted last with behaviour the_behaviours[dy][ID_AGENT]
	doFullInteractUpdate(sstate, iteration, the_actors[dy][ID_AGENT],
			     BEH_AGENT, agent_id->index, client_id->index, &(the_behaviours[dy][ID_AGENT]),
			     tau_interact[dyads[the_dyad[dy]].dyad_index], f_interact[dyads[the_dyad[dy]].dyad_index],
			     tau_out, dyads[the_dyad[dy]].interact_beh[ID_AGENT], dyads[the_dyad[dy]].interact_beh[ID_CLIENT]);
	computeInteractMean(sstate, BEH_AGENT, tau_out, tau_out);
      } if (acted[dy][ID_CLIENT]) {
	// update the ongoing monitoring of the interact state for the AGENT of this dyad, tau_interact[dy]
	// if it was the AGENT that acted last with behaviour the_behaviours[dy][ID_AGENT]
	doFullInteractUpdate(sstate, iteration, the_actors[dy][ID_AGENT],
			     BEH_CLIENT, agent_id->index, client_id->index, &(the_behaviours[dy][ID_CLIENT]),
			     tau_interact[dyads[the_dyad[dy]].dyad_index], f_interact[dyads[the_dyad[dy]].dyad_index],
			     tau_out, dyads[the_dyad[dy]].interact_beh[ID_AGENT], dyads[the_dyad[dy]].interact_beh[ID_CLIENT]);
	computeInteractMean(sstate, BEH_CLIENT, tau_out, tau_out);
      }
      // interact mean in _sstate->sentiParams[dyads[the_dyad[dy]].dyad_index]->interact_mean
      // copy over the state
      for (int j=0; j<VD; j++) 
	tau_interact[dyads[the_dyad[dy]].dyad_index][j]=tau_out[j];
      
      
      calcExpectedDistanceToTransient(sstate, tau_interact[dyads[the_dyad[dy]].dyad_index], dist,1);
#ifdef VERYVERBOSE
      if (verbosity) {
	printf("Distance between Interact transient and BayesACT distribution:\n");
	printf("dist=");
      }
#endif
      summed_distance = 0;
      summed_squared_distance = 0;
      for (int j=0; j<VD; j++) {
	summed_distance += fabs(dist[j]);
	summed_squared_distance += dist[j]*dist[j];
#ifdef VERYVERBOSE
	if (verbosity) 
	  printf("%4.3g ",dist[j]);
#endif
      }
#ifdef VERYVERBOSE
      if (verbosity) {
	printf("\n AVERAGE of summed distance transient: %4.3g\n",(summed_distance/(2.0*SSD)));
	printf("sqrt(summed squared distance): %4.3g\n",sqrt(summed_squared_distance));
      }
#endif
      
      if (output_csv) {
	if (iteration==0) {
	  sprintf(tmp_header_line,"%s: bagent act correct?, %s: agent act dist, %s: client act correct?, %s: client act dist, %s: average distance,",
		  csvlabels[dy],csvlabels[dy],csvlabels[dy],csvlabels[dy],csvlabels[dy]);
	  strcat(csv_header_line,tmp_header_line);
	}
      	fprintf(fpout,"%d,%4.3g,",(agent_beh_index==the_behaviours[dy][ID_AGENT].index),sqrt(agent_dist));
	fprintf(fpout,"%d,%4.3g,",(client_beh_index==the_behaviours[dy][ID_CLIENT].index),sqrt(client_dist));
	fprintf(fpout,"%4.3g,",(summed_distance/(2.0*SSD)));
      }
    }
    // at the end of each iteration, we compare the fundamental sentiment (identities only) from bayesact 
    // to the actual identities that the *other* (mirror) agent/dyad *started with* to see if they match
    // this was wridyen for testing to see if BayesACT can track interact - to be removed
    // but can be used in general to see how well the two agents in a dyad are aligned
    // only works if dyad[1] actually exists (if doing_mirror_dyad=1)

    // compute mean
    double *f_avg = (double *) malloc(VD*sizeof(double));
    double *tau_avg = (double *) malloc(VD*sizeof(double));
    xxstate xx_avg;
    double f_avg_swap;
    dy=0;
    //getExpectedSample((dyads+the_dyad[dy])->dsstate, f_avg, tau_avg, &xx_avg);
    //printf("f_avg[0]: %g %g %g --- %g %g %g \n",    f_avg[0],f_avg[1],f_avg[2],f_avg[6],f_avg[7],f_avg[8]);
    //getExpectedSample((dyads+the_dyad[1-dy])->dsstate, f_avg, tau_avg, &xx_avg);
    //printf("1-dy f_avg[0]: %g %g %g --- %g %g %g \n",    f_avg[6],f_avg[7],f_avg[8],f_avg[0],f_avg[1],f_avg[2]);

    for (dy=0; dy<2; dy++) {
      getExpectedSample((dyads+the_dyad[1-dy])->dsstate, f_avg, tau_avg, &xx_avg);
      // swap agent/client in f_avg 
      for (int j=0; j<SSD; j++) {
	f_avg_swap=f_avg[j+SSD*2];
	f_avg[j+SSD*2]=f_avg[j];
	f_avg[j]=f_avg_swap;
      }
      calcExpectedDistanceToTransient((dyads+the_dyad[dy])->dsstate, f_avg, dist,0);

      summed_distance = 0;
      summed_squared_distance = 0;
      // only take the distance over identities not behaviour
      for (int j=0; j<SSD; j++) 
	dist[j+SSD]=0.0;
      for (int j=0; j<VD; j++) {
	summed_distance += fabs(dist[j]);
	summed_squared_distance += dist[j]*dist[j];
      }
      if (verbosity) {
	printf("\n Average summed absolute distance of fundamentals (mean amount of deflection per sentiment variable for agent %s): %4.3g\n",the_actors[dy][ID_AGENT]->name,(summed_distance/(2.0*SSD)));
	printf("sqrt(summed squared distance) of fundamentals (%s): %4.3g\n",the_actors[dy][ID_AGENT]->name,sqrt(summed_squared_distance));
      }
      if (output_csv) {
	if (iteration==0) {
	  sprintf(tmp_header_line,"%s: summed f dist, %s: sum square f dist,",csvlabels[dy],csvlabels[dy]);
	  strcat(csv_header_line,tmp_header_line);
	}
	fprintf(fpout,"%4.3g,",(summed_distance/(2.0*SSD)));
	fprintf(fpout,"%4.3g,",sqrt(summed_squared_distance));
      }
    }
    free(f_avg);
    free(tau_avg);
    

    
    if (0 && doing_mirror_dyad) {

      // do both, even though one way doesn't matter 
      for (dy=0; dy<2; dy++) {
	agent_id=dyads[the_dyad[1-dy]].client_ids;   // uses the first one
	client_id=dyads[the_dyad[1-dy]].agent_ids;
	for (int i=0; i<VD; i++) 
	  tmpf[i]=0.0;
	// actual identities of the other agent at the start
	for (int i=0; i<SSD; i++) 
	  tmpf[i]=gsl_vector_get(the_actors[1-dy][ID_AGENT]->entity_dicts[AGENT]->theDictionary[agent_id->index].epa,i);
	for (int i=0; i<SSD; i++) 
	  tmpf[i+SSD*2]=gsl_vector_get(the_actors[1-dy][ID_AGENT]->entity_dicts[CLIENT]->theDictionary[client_id->index].epa,i);
	// distance from the_dyad[dy]'s fundamental sentiment and the actual identities of the agent being model
	calcExpectedDistanceToTransient(sstate, tmpf, dist,0);
      
	// this is the distance between what the *other* agent in this dyad started out with as a vision of the situation
	// and what *this* agent ended up with. This is meant to be low if one agent uses BayesACT and reidentifies
	// and the other agent uses interact and stays the same
	summed_distance = 0;
	summed_squared_distance = 0;
	for (int j=0; j<SSD; j++) 
	  dist[j+SSD]=0.0;
	for (int j=0; j<VD; j++) {
	  summed_distance += fabs(dist[j]);
	  summed_squared_distance += dist[j]*dist[j];
	}
	if (verbosity) 
	  if (strcmp(the_actors[dy][ID_AGENT]->name,"tom")==0) {
	    printf("\n Summed distance of fundamentals: %4.3g\n",(summed_distance/(1.0*VD)));
	    printf("sqrt(summed squared distance) of fundamentals: %4.3g\n",sqrt(summed_squared_distance));
	  }

	if (output_csv) {
	  // cheap hack for my simulations only - we have to track one agent only and compare to the original of other one
	  // this is a cheap hack so that the columns stay consistent (otherwise the columns are first the agent that
	  // acted, and then the agent that didn't).
	  if (strcmp(the_actors[dy][ID_AGENT]->name,"tom")==0) {
	    if (iteration==0) {
	      sprintf(tmp_header_line,"agent tom summed f dist,agent tom sum square f dist,");
	      strcat(csv_header_line,tmp_header_line);
	    }
	    fprintf(fpout,"%4.3g,",(summed_distance/(1.0*VD)));
	    fprintf(fpout,"%4.3g,",sqrt(summed_squared_distance));
	  }
	}
      }
    }
#endif  // DO_INTERACT
// ----------------------------------------------------------------------------------------------- end of INTERACT STUFF
    if (output_csv) {
      if (iteration==0) {
	strcat(csv_header_line,"\n");
      }
      fprintf(fpout,"\n");
    }

    
    // ================================================================================================
    // end of BayesACT iteration
    // ================================================================================================
    iteration++;
      
  }
  // ================================================================================================
  // clean up
  // ================================================================================================
  // first close the csv file
  if (output_csv) 
      fclose(fpout);

  // now we must re-write fpout with csv_header_line at the head
  // re-open the temporary csv file
  if (output_csv) {
    //printf("opening %s\n",outfile);
    fpout=fopen(outfile,"w");
    //printf("opening %s\n",outfile_tmp);
    fpout_tmp=fopen(outfile_tmp,"r");
    if (!fpout) {
      fprintf(stderr,"Error opening output file %s\n",outfile);
      exit(1);
    }
    if (!fpout_tmp) {
      fprintf(stderr,"Error opening output file %s\n",outfile_tmp);
      exit(1);
    }
    fprintf(fpout,"%s",csv_header_line);
    // copy in all the csv
    while(fgets(buffer, max_line_length, fpout_tmp)) {
      fprintf(fpout,"%s",buffer);
    }
    fclose(fpout);
    fclose(fpout_tmp);
  }

  //free up memory
  free(first_iteration);
  for (int j=0; j<num_dyads; j++)
    free(tau_interact[j]);
  free(tau_interact);
  for (int j=0; j<num_dyads; j++)
    free(f_interact[j]);
  free(f_interact);
  free(csv_header_line);
  free(tmp_header_line);
  free(emot);
  for (int j=0; j<num_dictionaries; j++) {
    freeInstitutionFilterList(dictionaries+j,num_dyads);
    freeSentimentDictionary(dictionaries+j);
  }
  free(dictionaries);
  for (int j=0; j<num_dyads; j++) 
    freeParameters(dyads+j);
  free(dyads);
  // be free, agents.
  free(agents);
  if (query_user==EVENTS)
    free(events);
  for (dy=0; dy<2; dy++)
    freeTree(planTree[dy],NULL);
  
}

