/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : June 2016                                                                             #
#  Copyright (C) 2016  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#* forward value search 
#************************************************************************************************/
#include "forvalue.h"


// initialize a tree node
// there are two distinct types of level in the plan tree - one for action followed by one for observation
// 
void initializeTreeNode(int num_samples, treeNode **b)
{
  *b = (treeNode *) malloc(sizeof(treeNode));
  (*b)->samples = (sample *) malloc(num_samples*sizeof(sample));
  (*b)->children=NULL;
  (*b)->siblings=NULL;
  (*b)->value = 0.0;
  (*b)->prob = 0.0;
  (*b)->N = 0;
  (*b)->index.act_index=0;
  (*b)->index.emot_index=0;
  initializeXXObs(&((*b)->xo));
}

// copy a tree node -- only the allocated memory though
// this is used when creating a new observation node (2nd level)
// so the samples can be weighted by the observations
static void copyTreeNode(int num_samples, treeNode *bcopy, treeNode *b)
{
  memcpy(bcopy->samples,b->samples,num_samples*sizeof(sample));
  memcpy(bcopy->xo,b->xo,sizeof(xostate));
}

// free the allocated memory in a tree Node
static void freeTreeNode(treeNode *b)
{
  if (b != NULL) {
    free(b->samples);
    freeXXObs(&(b->xo));
  }
}
// returns a pointer out to the child of treeNode tree that matches act,emot and xo (if not NULL),
// or NULL if it is not a child of tree
// prev points to the previous node - either tree itself if tree has no children
// or to the previous sibling child in tree
static void getChild(treeNode *tree, treeNode **out, treeNode **prev, denotativeState *act, denotativeState *emot, xostate *xo)
{
  (*out) = tree->children;
  (*prev) = tree;
  // this loop keeps going so long as one of act, emot or xo (if specified) doesn't match with a child branch of the treeNode tree
  while ((*out) != NULL && ((xo != NULL && !compareXXObs(xo,(*out)->xo)) ||
			    act->index != (*out)->index.act_index ||
			    emot->index != (*out)->index.emot_index))
    {
      (*prev)=(*out);
      (*out) = (*out)->siblings;
    }
}

// compute the expected value across all children (first generation only) of a node
static void normalizeChildren(treeNode *tree)
{
  double total_prob = 0.0;
  treeNode * temp = tree->children;
  while (temp != NULL) {
    total_prob += temp->N;
    temp = temp->siblings;
  }
  temp = tree->children;
  while (temp != NULL) {
    temp->prob = temp->N/total_prob;
    temp = temp->siblings;
  }
  return;
}


// compute the expected value across all children (first generation only) of a node
static double computeExpectedValue(treeNode *tree)
{
  double value = 0.0;
  double total_prob = 0.0;
  int num_children=0;
  treeNode * temp = tree->children;
  while (temp != NULL) {
    total_prob += temp->prob;
    temp = temp->siblings;
    num_children++;
  }
  //printf("total prob %g\n",total_prob);
  // TODO  - check this - the check on total_prob being zero seems artificial
  temp = tree->children;
  double multfactor=1.0/num_children;
  if (total_prob > 0.0) 
    multfactor = 1.0/total_prob;

  while (temp != NULL) {
    value += temp->value*temp->prob;
    temp = temp->siblings;
  }
  return value*multfactor;
}

int getMaxDepth(treeNode *tree)
{
  if (tree==NULL)
    return 0;
  treeNode *temp = tree->children;
  int max_child_depth=-1;
  int child_depth;
  while (temp != NULL) {
    child_depth = getMaxDepth(temp);
    if (child_depth > max_child_depth) 
      max_child_depth=child_depth;
    temp = temp->siblings;
  }
  return max_child_depth+1;
}
int getMaxNumKids(treeNode *tree)
{
  if (tree==NULL)
    return 0;
  treeNode *temp = tree->children;
  int max_num_kids=-1;
  int kid_max_num_kids, num_kids=0;
  while (temp != NULL) {
    kid_max_num_kids = getMaxNumKids(temp);
    if (kid_max_num_kids > max_num_kids)
      max_num_kids = kid_max_num_kids;
    temp = temp->siblings;
    num_kids += 1;
  }
  
  return (max_num_kids > num_kids ? max_num_kids : num_kids);
}
int countTreeNodes(treeNode *tree)
{
  int count=1;
  if (tree==NULL) 
    return 0;
  treeNode *temp = tree->children;
  while (temp != NULL) {
    count += countTreeNodes(temp);
    temp = temp->siblings;
  }
  return count;
}

// print the whole tree recursively in a really simple way with spaces
static void printTreeHelper(treeNode *tree, agentState * the_agent, int * unfinished_parent_depths, int num_unfinished_parents, int depth)
{
  int i=0, j=0;
  while (i<depth) {
    if (i==unfinished_parent_depths[j]) {
      printf(" |");
      j++;
    } else {
      printf("  ");
    }
    i++;
  }
  printf(">");

  char xostr[16];
  xxObsToString(xostr,tree->xo);
  if (depth%2==0)
    printf("C:");
  else
    printf("A:");
  printf("Value of node %d (%s) %d (%s) xo: %s with prob %g N %d: %3.2g\n",
	 tree->index.act_index,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[tree->index.act_index].label,
	 tree->index.emot_index,the_agent->entity_dicts[EMOTION]->theDictionary[tree->index.emot_index].label,
	 xostr,tree->prob,tree->N,tree->value);
  treeNode *temp = tree->children;
  while (temp != NULL) {
    if (temp->siblings != NULL) {
      unfinished_parent_depths[num_unfinished_parents]=depth;
      printTreeHelper(temp,the_agent,unfinished_parent_depths,num_unfinished_parents+1,depth+1);
    } else {
      printTreeHelper(temp,the_agent,unfinished_parent_depths,num_unfinished_parents,depth+1);
      unfinished_parent_depths[num_unfinished_parents]=-1;
    }
    temp = temp->siblings;
  }
}

void printTree(treeNode *tree, agentState * the_agent)
{
  int parent_depths[256];
  for (int i=0; i<256; i++)
    parent_depths[i]=-1;
  if (tree != NULL) {
    printTreeHelper(tree,the_agent,parent_depths,0,0);
  }
}
void printTreeTikZHelper(treeNode *tree, agentState * the_agent, int heightfactor, int widthfactor, int startcol, int max_branch, int xval, int depth)
{
  char *action_belief_nodetype = "randomvarp";
  char *post_belief_nodetype = "randomvar";
  char *nodetype;
  int is_obs = 0;
  if (depth%2==1) {
    nodetype = post_belief_nodetype;
  }  else {
    nodetype = action_belief_nodetype;
    is_obs = 1;
  }
  printf("\\node (pa%d%d) at (%d, %d) [%s] {%4.3g};\n",depth,xval+startcol,widthfactor*xval,heightfactor*depth,nodetype,tree->value);

  treeNode *temp = tree->children;
  int xvalchildstep = pow(max_branch,depth-1);
  int xvalstep=startcol+xvalchildstep/2;
  int startxval = startcol;
  while (temp != NULL) {
    printTreeTikZHelper(temp,the_agent,heightfactor,widthfactor,startxval,max_branch,xvalstep,depth-1);
    if (is_obs) {
      /*printf("\\draw [->] (pa%d%d) -- (pa%d%d) [very thick] node [sloped, midway, above] {(%s,%s):%3.2g};\n",depth,xval+startcol,depth-1,xvalstep+startxval,
	     the_agent->entity_dicts[BEHAVIOUR]->theDictionary[temp->index.act_index].label,
	     the_agent->entity_dicts[EMOTION]->theDictionary[temp->index.emot_index].label,
	     temp->prob);
      */
      printf("\\draw [->] (pa%d%d) -- (pa%d%d) [very thick] node [scale=2.0,sloped, midway, above] {%s:%3.2g};\n",depth,xval+startcol,depth-1,
	     xvalstep+startxval,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[temp->index.act_index].label,
	     temp->prob);
    } else {
      /*printf("\\draw [->] (pa%d%d) -- (pa%d%d) [very thick] node [sloped, midway, above] {(%s,%s)};\n",depth,xval+startcol,depth-1,xvalstep+startxval,
	     the_agent->entity_dicts[BEHAVIOUR]->theDictionary[temp->index.act_index].label,
	     the_agent->entity_dicts[EMOTION]->theDictionary[temp->index.emot_index].label);
      */
      printf("\\draw [->] (pa%d%d) -- (pa%d%d) [very thick] node [scale=2.0,sloped, midway, above] {%s};\n",depth,xval+startcol,depth-1,xvalstep+startxval,
	     the_agent->entity_dicts[BEHAVIOUR]->theDictionary[temp->index.act_index].label);
    }
    temp = temp->siblings;
    xvalstep += xvalchildstep;
    startxval += xvalchildstep;
  }
}

// this will lay things out for a complete tree the depth of the tree being printed,
// and draw everything symmetrically
void printTreeTikZ(treeNode *tree, agentState * the_agent)
{
  if (tree == NULL)
    return;
  int max_depth=getMaxDepth(tree);
  int max_branch=getMaxNumKids(tree);
  int max_width=(int) pow(max_branch,max_depth+1);
  int heightfactor=max_depth*2;
  int widthfactor=max_branch;
  printf("\\begin{center}\\resizebox{\\textwidth}{!}{\n");
  printf("\\begin{tikzpicture}\n");
  printf("[\n");
  printf("randomvar/.style={circle, draw, scale=2.00, inner sep=1mm},\n");
  printf("randomvarp/.style={circle, draw, dashed, scale=2.00, inner sep=1mm},\n");
  printf("];\n");
  //printf("%%botleft (x,y) to top right (x',y')\n");
  //printf("\\draw (1,1) to[grid with coordinates] (%d,%d);\n",widthfactor*max_width,heightfactor*(max_depth+1));

  printTreeTikZHelper(tree,the_agent,heightfactor,widthfactor, 0,max_branch,max_width/max_branch,max_depth+1);

  printf("\\end{tikzpicture}\n");
  printf("}\n");
  printf("\\end{center}\n");
}

// print the values of all actions for a tree - first generation children only
void printActionValues(treeNode *tree, agentState *the_agent)
{
  treeNode *temp = tree->children;
  while (temp != NULL) {
    printf("Value of node %d (%s) %d (%s) : %3.2g\n",
	   temp->index.act_index,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[temp->index.act_index].label,
	   temp->index.emot_index,the_agent->entity_dicts[EMOTION]->theDictionary[temp->index.emot_index].label,
	   temp->value);
    temp = temp->siblings;
  }
}
// get the best value for all first generation children only
// best_index is the index to the best value which is returned
static double getBestValue(treeNode *tree, dindex *best_index)
{
  treeNode *temp = tree->children;
  int first=1;
  double best_value;
  while (temp != NULL) {
    if (first || temp->value > best_value) {
      best_value = temp->value;
      (*best_index) = temp->index;
      first=0;
    }
    temp = temp->siblings;
  }
  return best_value;
}

// gets the most likely branch of the tree - first generation children only
// best_index is the index to the most probable node which is returned
static double getBestProb(treeNode *tree, dindex *best_index)
{
  treeNode *temp = tree->children;
  int first=1;
  double best_prob;
  while (temp != NULL) {
    if (first || temp->prob > best_prob) {
      best_prob = temp->prob;
      (*best_index) = temp->index;
      first=0;
    }
    temp = temp->siblings;
  }
  return best_prob;
}

// gets the most valuable next action to take (for agent)
// and the most probable next action to take (for client)
void getBestNextActionsEmotions(treeNode *tree, int *agent_action, int *agent_emotion, int *client_action, int *client_emotion)
{
  treeNode *temp = tree->children;
  dindex best_index;
  // first find the best action in the tree
  getBestValue(tree,&best_index);
  *agent_action = best_index.act_index;
  *agent_emotion = best_index.emot_index;
  printf("best agent action found %d and emotion %d\n",*agent_action, *agent_emotion);
  // find that child again (this is a bit redundant and could be streamlined)
  while (temp->index.act_index != best_index.act_index || temp->index.emot_index != best_index.emot_index) 
    temp=temp->siblings;


  getBestProb(temp,&best_index);
  *client_action = best_index.act_index;
  *client_emotion = best_index.emot_index;
  printf("best client action found %d and emotion %d\n",*client_action, *client_emotion);
}

// get the best value for all first generation children only
// with values weighted by the prob weights
// best_index is the index to the best value which is returned
/* unused MARCH 2021
static double getBestValueWeighted(treeNode *tree, dindex *best_index)
{
  treeNode *temp = tree->children;
  int first=1;
  double best_value;
  while (temp != NULL) {
    if (first || temp->value*temp->prob > best_value) {
      best_value = temp->value*temp->prob;
      (*best_index) = temp->index;
      first=0;
    }
    temp = temp->siblings;
  }
  return best_value;
}
*/
// get the utility of a sample - this simply calls an application dependent function
static double  getUtility(sample *the_sample)
{
  
  double xx_utility = getXXUtility(&(the_sample->xx))*the_sample->weight;
  double deflection_utility = computeDeflectionOfSample(the_sample,AGENT);
  deflection_utility += computeDeflectionOfSample(the_sample,CLIENT);
  deflection_utility += computeDeflectionOfSample(the_sample,BEHAVIOUR);
  return (UTILITY_WEIGHT*xx_utility - DEFLECTION_WEIGHT*deflection_utility);
  //return -1.0*deflection_utility;
}

static double computeExpectedUtility(treeNode *tree, int num_samples)
{
  int i;
  double val=0.0, totweight=0.0;
  for (i=0; i<num_samples; i++) {
    val += getUtility(tree->samples+i)*tree->samples[i].weight;
    totweight += tree->samples[i].weight;
  }
  return val/totweight;

}
// sample an observation from a belief state represented by a set of samples the_samples
static void sampleObservationFromState(dyad *the_dyad,  agentState *the_agent,
				       sample * the_samples, int num_samples,
				       behaviour * the_client_behaviour,
				       emotion * the_client_emotion, xostate *xo)
{
  // draw a sample
  int a_sample_index = drawSample(the_dyad->dsstate->r, the_samples, num_samples);
  // sample an observation from that sample
  //printf("----- sampled observation:\n");
  sampleObservation(the_dyad, the_agent, the_samples+a_sample_index, the_client_behaviour, the_client_emotion, xo);
  //printRawSample(the_samples+a_sample_index);
  //printf("%d %d %d\n",the_client_behaviour->index,the_client_emotion->index, xo->obs_amount_won);
}

// the main simulation function
// this starts with the specified agent behaviour and emotion, then simulates the rest
void updateValue(dyad *the_dyad, agentState *the_agent, d_entity_type who_acted, treeNode *b,
		 behaviour * the_agent_behaviour, emotion *the_agent_emotion, int num_samples, int depth)
{
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index=the_dyad->dyad_index;
  int i;
  //printf("updating value at depth %d\n",depth);
  treeNode *temp=NULL, *temp2=NULL;
  treeNode *bp, *bpp;
  double samp_weight;
  d_entity_type whodat;
  // ***************************
  // action node
  // ***************************
  // see if a is a child already
  getChild(b, &temp, &temp2, the_agent_behaviour, the_agent_emotion, NULL);
  
  if (temp==NULL) { // not found
    //printf("adding node!\n");
    initializeTreeNode(num_samples, &bp);
    // here we need to also update the X-state but not weight by observations
    // need to make this a new function in fvarsample.c -- the remainder is below when "weight and resample" happens
    for (i=0; i<num_samples; i++)  {
      whodat = (gsl_rng_uniform(sstate->r) > 0.5 ? BEH_AGENT : BEH_CLIENT);
      if (sstate->use_emotions) {
	sampleNextState(the_dyad, whodat, the_agent, RESET_NONE,
			b->samples+i, bp->samples+i, the_agent_behaviour, the_agent_emotion, NULL, NULL);
      } else {
	sampleNextState(the_dyad, whodat, the_agent, RESET_NONE,
			b->samples+i, bp->samples+i, the_agent_behaviour, NULL, NULL, NULL);
      }
      // fill in the client behaviour with a denotative somatic transform!
      // we should only do this if there are no denotative dynamics that specify what xbc is -
      // that is, if updateXbcSample returns anything other than a random number !!
      if (whodat==BEH_CLIENT)  {
	somaticTransform(DENOTATIVE, sstate->r, dyad_index, &(bp->samples[i].x[BEH_CLIENT]), bp->samples[i].f+SSD*BEHAVIOUR, the_agent->entity_dicts[BEHAVIOUR]);
	// need to also fix up the xx state though! - this feels really hacky and should probably be done in some other way
	updateXXSample(sstate->r, &(bp->samples[i].xx), &(b->samples[i].xx), bp->samples[i].x[BEH_AGENT], bp->samples[i].x[BEH_CLIENT]);
      }
      // I think this should all be done by simply calling getNextAction???
      // no because that does the additional work of figuring out a list of best denotative actions
      /*
      behaviour * beh_list=NULL;
      int num_potential_actions = getNextAction(sstate, dyad_index, the_agent, &beh_list, BEH_CLIENT);
      printf("getting best action for client:\n");
      for (int i=0; i< num_potential_actions; i++) 
	printf("%d %g\n",beh_list[i].index,beh_list[i].prob);
      */
    }
    
    bp->index.act_index = the_agent_behaviour->index;
    bp->index.emot_index = the_agent_emotion->index;
    if (temp2 == b) 
      temp2->children = bp;
    else 
      temp2->siblings = bp;
  } else {
    bp = temp;
  }

  // ***************************
  // observation node
  // ***************************
  // bp is the next state after action the_agent_behaviour+the_agent_emotion
  // now get the next observation
  behaviour the_client_behaviour;
  emotion the_client_emotion;
  xostate xo;
  // sample an observation from the current state of bp
  sampleObservationFromState(the_dyad, the_agent, bp->samples, num_samples, &the_client_behaviour, &the_client_emotion, &xo);

  // see if it is in the tree
  getChild(bp, &temp, &temp2, &the_client_behaviour, &the_client_emotion, &xo);

  if (temp==NULL) { // not found
    //printf("adding node! at depth %d\n",depth);
    initializeTreeNode(num_samples, &bpp);
    // copy it over - all that changes are the weights, so we could save memory by keeping only
    // one set of samples, but two sets of weights
    copyTreeNode(num_samples, bpp, bp);
    bpp->index.act_index = the_client_behaviour.index;
    bpp->index.emot_index = the_client_emotion.index;
    memcpy(bpp->xo,&xo,sizeof(xostate));
    /*printf("**************************** computing weight for xo: ");
    printRawXXObs(&xo);
    printf("                                                  the_client_behaviour: %d\n",the_client_behaviour.index);
    printf("                                                  the_client_emotion %d\n",the_client_emotion.index);
    */
    //bpp->prob=0.0;
    for (i=0; i<num_samples; i++) {
      if (sstate->use_emotions)
	samp_weight = computeSampleWeight(sstate, dyad_index, the_agent, 
					  bpp->samples+i, &xo, &the_client_behaviour, &the_client_emotion, the_agent_emotion);
      else    
	samp_weight = computeSampleWeight(sstate, dyad_index, the_agent, 
					  bpp->samples+i, &xo, &the_client_behaviour, NULL, NULL);
      (bpp->samples+i)->weight *= samp_weight;
    }

    reSampleInPlace(bpp->samples,num_samples);

    // expected value of this new node is the expected utility of the samples, which are now unweighted
    bpp->value = 0.0;
    for (i=0; i<num_samples; i++) 
      bpp->value += getUtility(bpp->samples+i)/num_samples;


    if (temp2==bp)
      temp2->children = bpp;
    else
      temp2->siblings = bpp;
    // number of visits to this child, stored in the child node
    bpp->N = 1;
  } else {
    bpp = temp;
    // number of visits to this child, incremented
    bpp->N +=1;
  }
  // total number of visits to this node
  bp->N += 1;
  behaviour * beh_list=NULL, * best_beh=NULL;
  emotion * emot_list=NULL, * best_emot=NULL;
  int num_potential_actions, num_potential_emotions;

  // normalize probabilities of children here
  // this updates the "prob" variable according to N and N of each child
  normalizeChildren(bp);

  // now we can also normalize the action nodes according to how often they have been sampled
  // this can then be used to weight the action choices so that a really good looking action (value-wise)
  // that was only tried a small number of times (so is more deflecting) is down-weighted.
  normalizeChildren(b);

  // recursive step - get next action and emotion to take and call updateValue
  if (pow(DISCOUNT,depth) > TOLERANCE) {
    // TODO: replace this with a function that draws a single sample 
    num_potential_actions = getNextAction(the_dyad, the_agent, &beh_list, who_acted);

    // gets a list of the next emotions 
    // if use_emotions=0 then it returns a single null emotion in the list
    // if emotion_planning=0 then it returns a single (default) emotion in the list
    // otherwise returns a sorted list of best emotions
    num_potential_emotions = getCurrentEmotion(the_dyad, the_agent, &emot_list, AGENT);

    //printf("num_potential_emotions: %d\n",num_potential_emotions);
    //for (int ii=0; ii<num_potential_emotions; ii++) 
    //printf("emotion %d prob %g\n",emot_list[ii].index,emot_list[ii].prob);

    //best_beh = drawRandomProb(sstate, beh_list);
    // see also on line 651
    best_beh = drawWeightedProb(sstate->r, beh_list, WEIGHTED_ACTION_FACTOR);
    //printf("best behaviour drawn: %d %g\n",best_beh[0].index,best_beh[0].prob);
    best_emot = drawRandomProb(sstate->r, emot_list);
    updateValue(the_dyad, the_agent, who_acted, bpp, best_beh, best_emot, num_samples, depth+1);
    free (beh_list);
    free (emot_list);
  }
  // expected value of child node is the expectation over observations
  bp->value = DISCOUNT*computeExpectedValue(bp);
  
  // expected value of node is the argmax over actions
  dindex best_index;

  // two options - one weighted and one not - also need to change in seekNextAction at the end where the best
  // action is actually chosen
  // standard version with no action probability weights
  b->value = getBestValue(b,&best_index)+computeExpectedUtility(b,num_samples);

  // version that weights on the action choices according to how likely they have been tried
  
  //b->value = getBestValueWeighted(b,&best_index)+computeExpectedUtility(b,num_samples);

  return;
}
// free the tree b but skip child skip (unless NULL)
// calling frame *must* maintain a pointer to the skip node
// could instead have this function return this pointer, or set b to it
void freeTree(treeNode *b, treeNode *skip)
{
  treeNode *temp, *temp2;
  // free everything else
  if (b==NULL)
    return;
  freeTreeNode(b);
  temp = b->children;
  temp2 = b->siblings;
  free(b);
  // free all the children!
  while (temp != NULL) {
    // save pointer to next sibling (may be NULL)
    temp2 = temp->siblings;
    // don't free the skip node
    if (temp != skip) 
      // free this child and all its descendants,
      // but not the skip node
      freeTree(temp,skip);
    // move on
    temp = temp2;
  }
}

// prune the planTree b with behaviours and emotions and xobservations as given
// if a branch is not found, then the planTree b is set to NULL and returned,
// otherwise, b points to the remaining tree after the prune
void pruneTree(treeNode **b, behaviour * the_agent_behaviour, emotion *the_agent_emotion,
		 behaviour * the_client_behaviour, emotion * the_client_emotion, xostate *xo)
{
  // see if a is a child already
  treeNode *temp=NULL, *temp2=NULL;
  treeNode *bp, *bpp;
  getChild(*b, &temp, &temp2, the_agent_behaviour, the_agent_emotion, NULL);
  if (temp==NULL) { // not found
    printf("agent action or emotion not found %d %d\n",the_agent_behaviour->index, the_agent_emotion->index);
    // start all over again
    freeTree(*b,NULL);
    *b=NULL;
    return;
  }
  bp=temp;
  getChild(bp, &temp, &temp2, the_client_behaviour, the_client_emotion, xo);
  if (temp==NULL) {
    printf("client action or emotion not found %d %d xo: ",the_client_behaviour->index, the_client_emotion->index);
    printRawXXObs(xo);
    freeTree(*b,NULL);
    // start all over again
    *b=NULL;
    return;
  }
  // points to the new root node
  bpp=temp;

  printf("freeing next level\n");
  freeTree(*b,bpp);

  *b=bpp;
}

// the primary function that is used to find a next action to take
// best_act is the output best action to take
// best_emotion is the output best emotion to display
// secs_timeout gives the number of seconds more than one to plan for: setting this as zero means the planner will run for 1 second
void seekNextAction(dyad *the_dyad, agentState *the_agent, d_entity_type who_acted, treeNode *b, int num_samples,
		    int * best_act, int * best_emotion, int secs_timeout)
{
  samplerState *sstate = the_dyad->dsstate;
  int timeout=0;
  int num_rollouts=0;
  behaviour * beh_list=NULL;
  emotion * emot_list=NULL;
  behaviour * best_beh=NULL;
  emotion * best_emot=NULL;
  // get a distribution over actions, or a single action if sstate->default_action is set
  //int num_potential_actions =
  getNextAction(the_dyad, the_agent, &beh_list, who_acted);

  //for (int i=0; i< num_potential_actions; i++) 
  //  printf("%d %g\n",beh_list[i].index,beh_list[i].prob);
    
  // get current emotion - can be one of three cases
  // either a sample over the distribution over emotions
  // or the most likely emotion only
  // or a null emotion (if emotions are not being used at all)
  //int num_potential_emotions =
  getCurrentEmotion(the_dyad, the_agent, &emot_list, AGENT);
  //printf("num_potential_emotions: %d\n",num_potential_emotions);
  //for (int ii=0; ii<num_potential_emotions; ii++) 
  //printf("emotion %d prob %g\n",emot_list[ii].index,emot_list[ii].prob);

  struct timeval time_start;
  struct timeval time_end;
  int secs_so_far=0;
  int first_pass=1;
  gettimeofday(&time_start, NULL);        
  while (!timeout) {
    if (MOST_LIKELY_BEH_FIRST && first_pass > 0) {
      // take the most likely behaviour and emotion on the first pass
      best_beh = beh_list;
    } else {
      // draw a sample from beh_list - weights are beh_list[i].prob
      //best_beh = drawRandomProb(sstate, beh_list);
      // set WEIGHTED_ACTION_FACTOR 0.0 in forvalue.h to remove the action factor
      // this is an arbitrary factor added to the most likely value given by
      // the tree, in conjunction with MOST_LIKELY_BEH_FIRST, this ensures that
      // at least, and potentially a lot more, the first action returned will
      // be aligned or not
      best_beh = drawWeightedProb(sstate->r, beh_list, WEIGHTED_ACTION_FACTOR);
    }
    
    if (MOST_LIKELY_EMOT_FIRST && first_pass > 0) {
      best_emot = emot_list;
    } else {
      // draw a sample from emot_list - weights are emot_list[i].prob
      best_emot = drawRandomProb(sstate->r, emot_list);
      //printf("drawing from emot_list: %d\n",best_emot->index);
    }
    if (first_pass > 0)
      first_pass=0;
    /*printf("rolling out %d with behaviour %d (%s) and emotion %d (%s)\n",num_rollouts,
	   best_beh->index,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[best_beh->index].label,
	   best_emot->index,the_agent->entity_dicts[EMOTION]->theDictionary[best_emot->index].label);
    */
    // descend the plan tree b using that behaviour and emotion, exetending the tree by a single node
    // if possible, otherwise  updating the value
    updateValue(the_dyad, the_agent, who_acted, b, best_beh, best_emot, num_samples, 0);
    //printTree(b,the_agent);
    num_rollouts++;
    gettimeofday(&time_end, NULL);
    secs_so_far = time_end.tv_sec-time_start.tv_sec;
    //printf("time taken so far: %d timeout: %d\n",secs_so_far,secs_timeout);
    //if (num_rollouts > 20)
    // if secs_timeout is 0 then we just do a single iteration
    if (secs_timeout < 1 || secs_so_far > secs_timeout) 
      timeout=1;
    //printActionValues(b);
  }
  //printTree(b,the_agent);
  //printf("number of rollouts: %d\n",num_rollouts);
  //printActionValues(b,the_agent);
  // get argmax_a  b.children[a].value
  dindex best_action;
  // two options
  // standard version with no action probability weights
  
  getBestValue(b,&best_action);
  // version that weights on the action choices according to how likely they have been tried
  //getBestValueWeighted(b,&best_action);
  (*best_act) = best_action.act_index;
  (*best_emotion) = best_action.emot_index;
  free (beh_list);
  free (emot_list);
  return;
}


