/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : June 2016                                                                             #
#  Copyright (C) 2016  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#* plan tree header - forward value search functions
#************************************************************************************************/
#ifndef __FORVALUE
#define __FORVALUE
#include <stdio.h>
#include <sys/time.h>

#include "fvarsample.h"

// the tree will stop expanding when DISCOUNT^depth > TOLERANCE
// so max depth = log(TOLERANCE)/log(DISCOUNT)
#define TOLERANCE 0.92
//#define TOLERANCE 0.75
#define DISCOUNT 0.95

//relative weight of deflection vs utility - a number from 0 to 1 
#define DEFLECTION_WEIGHT 0.0
#define UTILITY_WEIGHT (1.0-DEFLECTION_WEIGHT)

// amount to add to the zeroth element of the action selection distribution
// default of 0 means actions selected at random from the NAB distribution.
#define WEIGHTED_ACTION_FACTOR  1.0    
// if defined, when building the tree, will add the most likely action first
// according to the affect control principle.  If not defined (or defined as zero), 
// will draw a sample from the distribution given by the normative action bias
#define MOST_LIKELY_BEH_FIRST 1
#define MOST_LIKELY_EMOT_FIRST 0

typedef struct deno_index dindex;
struct deno_index {
  // the index of the action (agent *or* client for action / observation nodes resp.)
  int act_index;
  // the index of the emotion (again for agent *or* client for action/observation nodes resp.)
  int emot_index;
};

// a node in the plan tree
// a Plan Tree is implemented as a node with a pointer (tree->children) to a linked list of its children (siblings of tree->children)
// node -- NULL  (or siblings possibly here too at the root...a forest)
//  |
// node->children -- node->children->siblings -- node->children->siblings->siblings ... -- NULL
//  |
// node->children->children -- node->children->children->siblings ... -- NULL
//  |
// NULL
//
// sample: belief state at this treeNode
// value: expected value of that belief state
// prob: probability of the observation that *led to* this belief - awkward to have it this way
//       the alternative is to have an array of probabilities in the parent
// children: a pointer to the first child of this node
// siblings: a pointer to the next sibling of this node
typedef struct treenode treeNode;
struct treenode {
  sample * samples;
  double value;
  double prob;
  int N;
  dindex index;
  // the xstate observation (for observation nodes only)
  xostate * xo;  
  
  treeNode * children;
  treeNode * siblings;
};

// utility functions
int countTreeNodes(treeNode *tree);
int getMaxDepth(treeNode *tree);
int getMaxNumKids(treeNode *tree);
void getBestNextActionsEmotions(treeNode *tree, int *agent_action, int *agent_emotion, int *client_action, int *client_emotion);
void printActionValues(treeNode *tree, agentState *the_agent);
void printTree(treeNode *tree, agentState *the_agent);
void printTreeTikZ(treeNode *tree, agentState * the_agent);

// main functions
void initializeTreeNode(int num_samples, treeNode **b);

void seekNextAction(dyad *the_dyad, agentState *the_agent, d_entity_type who_acted, treeNode *b, int num_samples,
		    int * best_act, int * best_emotion, int secs_timeout);

void updateValue(dyad *the_dyad, agentState *the_agent, d_entity_type who_acted, treeNode *b,
		 behaviour * the_agent_behaviour, emotion *the_agent_emotion, int num_samples, int depth);


void pruneTree(treeNode **b, behaviour * the_agent_behaviour, emotion *the_agent_emotion,
	       behaviour * the_client_behaviour, emotion * the_client_emotion, xostate *xo);

void freeTree(treeNode *b, treeNode *skip);
#endif
