/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.3.8)                              #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@bayesact.ca         bayesact.ca                                     #
#  Date  : April 2021                                                                            #
#  Copyright (C) 2017-2021  Jesse Hoey                                                           #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  Main functions that do all the work of sampling etc
#************************************************************************************************/

#include "fvarsample.h"

// the following lists must be in the same order as the corresponding enums 
// some labels to map from agent_type to string descriptions of them
// entity_type
int num_c_entity_labels=NGT;
char * c_entity_labels[] = {"AGENT","BEHAVIOUR","CLIENT","EMOTION"};
int num_d_entity_labels=NGTD;
char * d_entity_labels[] = {"ID_AGENT","ID_CLIENT","BEH_AGENT","BEH_CLIENT"};

// dict_type
int num_dtype_labels=6;
char * dtype_labels[] = {"MEAN","SD","COV","SAMPLE"};

// helper functions for printing matrices and vectors
void printMatrix(gsl_matrix * m, char * name, int sizex, int sizey)

{
  int i,j;
  printf("%s\n",name);
  for (i = 0; i < sizex; i++) {
    for (j = 0; j < sizey; j++)
      printf ("%4.3g ", gsl_matrix_get (m, i, j));
    printf("\n");
  }
}
void printVector(gsl_vector * m, char * name, int sizex)
{
  int i;
  printf("%s\n",name);
  for (i = 0; i < sizex; i++) 
      printf ("%4.3g ", gsl_vector_get (m, i));
  printf("\n");
}
// all this does is figure out how many lines are in the dynamics or emotions file - this is the number of features
static int getNumFeaturesFromDynamicsFile(char *fname) {
  FILE *fp = fopen(fname,"r");
  if (!fp)
    return 0;
  int lines=0;
  char ch;
  while(!feof(fp)) {
    ch = fgetc(fp);
    if(ch == '\n') {
      lines++;
    }
  }
  return lines;
}

// function to compute the Euclidean distance between two double arrays of size nd 
double euclidean_distance(double *epa1, double *epa2, int nd)
{
  int i;
  double dist=0.0;
  for (i=0; i<nd; i++) {
    dist += pow(epa1[i]-epa2[i],2);
  }
  return sqrt(dist);
}

// function to compute the cosine distance between two double arrays of size nd 
double cosine_distance(double *epa1, double *epa2, int nd)
{
  int i;
  double cdist1=0.0;
  double cdist2=0.0;
  double dprod=0.0;
  for (i=0; i<nd; i++) {
    cdist1 += pow(epa1[i],2);
    cdist2 += pow(epa2[i],2);
    dprod += epa1[i]*epa2[i];
  }
  return dprod/(sqrt(cdist1)*sqrt(cdist2));
}

// looks up a string label in dictionary
// assigns index of s to be the index in the dictionary if it exists, -1 otherwise
// dictionary has n elements
// this will be *much* slower than Python as its a simple linear search
int lookUpLabel(char * slabel, dictionary * the_dictionary)
{
  for (int i=0; i<the_dictionary->num_items; i++) {
    if (strcmp(the_dictionary->theDictionary[i].label,slabel)==0) {
      //printf("found it at %d %s %s\n",i,the_dictionary->theDictionary[i].label,s->label);
      return i;
    }
  }
  return -1;
}
int lookUp(denotativeState *s, dictionary * the_dictionary)
{
  s->index = lookUpLabel(s->label,the_dictionary);
  return s->index;
}

// helper functions to map from strings to enumerated types for
// turn (lookUpActor), entitites (lookUpEntity) and Dictionary types (lookUpDictTypes)
// map char label to index (AGENT/CLIENT -> "agent"/"client")
int lookUpActor(char *label) {
  return lookUpGeneric(label,num_c_entity_labels,c_entity_labels);
}
int lookUpEntity(char *label) {
  return lookUpGeneric(label,num_c_entity_labels,c_entity_labels);
}
int lookUpDictType(char *label) {
  return lookUpGeneric(label,num_dtype_labels,dtype_labels);
}
// look up a label in a list and return index
int lookUpGeneric(char *label, int num_labels, char **label_list)
{
  int index=0;
  while (index<num_labels && !(strcmp(label_list[index],label)==0)) 
    index++;
  return index;
}

// compare two denotative states according to probability
// used to sort the arrays over behaviours and identities
int denotativeStateCompare( const void *s1, const void * s2) {
  denotativeState ds_s1 = *((denotativeState *) s1);
  denotativeState ds_s2 = *((denotativeState *) s2);

  if (ds_s1.prob > ds_s2.prob)
    return -1;
  if (ds_s1.prob < ds_s2.prob)
    return 1;
  return 0;
}


//open the input file for emotions (modifiers)
//each row of which contains a binary number prefaced with the character "Z"
//giving the terms in (fae,fap,faa,ea,ep,ea) where (ea,ep,ea) is the emotions vector
//followed by 3 coefficients
//the first row gives the constant d
//the next three rows give the 3x3 matrix E
//the next three rows give the 3x3 matrix R
//the remaining rows give the diagonal of the diagonal matrix Q
//the rest is the Q matrix, which is read generally as follows:

// old way of understanding this .. but was wrong I think
//each line is are the diagonal elements of a diagonal matrix Q
//Z100100 means Q_e and multiples a matrix I_Re that has fa_e along the diagonal  -->
//Z100010 means Q_e and multiples fa_p
//Z010001 means Q_p and multiplies fa_a
//Z001100 means Q_a and multiplies fa_e ??

// new way:
// each line has two 0..2 indices as specified by the 1s
// the second gives which of Q_e,Q_p,Q_a this refers to
// the first gives the column of that matrix, so:
// Z100100 means first columns of Q_e
// Z001100 means last column of Q_e
// Z010001 means middle column of Q_a
// Z010010 means middle column of Q_p
//...etc
// in the java code he calls the first index the “modifier index” and the second one the “noun index”
//which means the first position in Zxxxxxx is the numbers that will multiply e_e (the e component of the emotion), and this is the first column of Q_j  where j is the f_e,f_p, or f_a depending on where the “1” sits in the last three positions in Zxxxxxx
// Q_e is the matrix that "combines" f_ae with the emotion vector, and so contains the terms for
// fa_e*Q_e*(e_e, e_p, e_a) --> i.e. Q_e(j,j) is for the term fa_e*Q_e(j,j)*e_j
//incorrect:  we will assume only one fa and one e multiplier per matrix row  (i.e. the matrices Q are diagonal)
//we will put this all into a single double * array dREQ which,
// where dREQ[i*SSD...(i+1)*SSD] is a SSD-vector (double array of length SSD) that contains the term that
// multiplies the features dreq_index[i][0...dreq_num_feeatures[i]-1]
//where dreq_index[i][j] is an index into [ee,ep,ea...(more if SSD>3), fe,fp,fa...(more if SSD>3)]
// num_features is the number of lines in the file - get it with the function getNumFeaturesFromDynamicsFile
// and allocate dREQ to be this big before calling this function
static int readModifiersDynamicsFile(int num_features, int ** dreq_index, int *dreq_num_features, double *dREQ, char *fname)
{
  char c;
  int i,j,nnfeat;
  double x;

  int nn=SSD*2;  // number of features on each row

  char *tmp = (char *) malloc(nn*sizeof(char));

  //num_features needs to be passed in
#ifdef VERBOSE
  printf("number of features: %d\n",num_features);
  printf("reading from %s\n",fname);
#endif
  FILE *fp = fopen(fname,"r");
  if (!fp)
    return 0;
  int current_feature=0;
  while (fscanf(fp,"%c",&c) != EOF) {
    if (c != 'Z') // each line starts with a "Z"
      return 0;
    // these are the features
    nnfeat=0;  // arity of the polynomial feature
    for (i=0; i<nn; i++) {
      fscanf(fp,"%c",tmp+i);
      if (tmp[i]=='1')
	nnfeat++;
    }
    dreq_num_features[current_feature]=nnfeat;
    dreq_index[current_feature] = (int *) malloc(dreq_num_features[current_feature]*sizeof(int));
    j=0;
    for (i=0; i<nn; i++) 
      if (tmp[i]=='1')
	dreq_index[current_feature][j++]=i;
    
    // this is the matrix "dREQ" coefficients
    for (i=0; i<SSD; i++) {
      fscanf(fp,"%lg",&x);
      dREQ[current_feature*SSD+i]=x;
    }
    //read newline  
    fscanf(fp,"%c",&c);
    current_feature++;
  }
  // the matrices E and R are transposed so we tranpose them
  // TOCHECK: remove this for interesting results?  - see also below in computeEmotion
  double tmpv[SSD*SSD];
  for (i=0; i<SSD; i++) 
    for (j=0; j<SSD; j++) 
      tmpv[j*SSD+i] = dREQ[(i+1)*SSD+j];
  for (i=1; i<=SSD; i++) 
    for (j=0; j<SSD; j++) 
      dREQ[i*SSD+j]=tmpv[(i-1)*SSD+j];
  for (i=0; i<SSD; i++) 
    for (j=0; j<SSD; j++) 
      tmpv[j*SSD+i] = dREQ[(i+4)*SSD+j];
  for (i=4; i<=SSD+3; i++) 
    for (j=0; j<SSD; j++) 
      dREQ[i*SSD+j]=tmpv[(i-4)*SSD+j];


  
  free(tmp);
  fclose(fp);
  return 1;
}
// function to read the dynamics file (e.g. tdynamics-male.dat)
// this text file should have the following format. 
// Each line contains a single feature in the impression formation feature vector (polynomial features)
// first character: Z
// next num_feature_terms characters: 0 or 1 (denoting if that grammar term [e.g. tae, tap, taa, tbe, tbp, tba, tce, tcp, tca] is in the feature)
// next num_feature_terms floating point numbers: elements of M matrix M[i][j] for the ith feature, jth grammar term
//
// num_feature_terms and num_features could be figured out from this file, but it would require two passes
// see getNumFeaturesfromDynamicsFil
static int readDynamicsFile(int num_feature_terms, int ** m_index, int *m_num_features, double *M, char *fname)
{
  char c;
  int i,j,nnfeat;
  double x;
  // num_feature_terms needs to be passed in - this is the number of columns
  char *tmp = (char *) malloc(num_feature_terms*sizeof(char));

#ifdef VERBOSE
  printf("reading from %s\n",fname);
#endif
  FILE *fp = fopen(fname,"r");
  if (!fp)
    return 0;
  int current_feature=0;
  while (fscanf(fp,"%c",&c) != EOF) {
    if (c != 'Z') 
      return 0;
    // these are the features
    nnfeat=0;  // arity of the polynomial feature
    for (i=0; i<num_feature_terms; i++) {
      fscanf(fp,"%c",tmp+i);
      if (tmp[i]=='1')
	nnfeat++;
    }
    m_num_features[current_feature]=nnfeat;
    m_index[current_feature] = (int *) malloc(m_num_features[current_feature]*sizeof(int));
    j=0;
    for (i=0; i<num_feature_terms; i++) 
      if (tmp[i]=='1')
	m_index[current_feature][j++]=i;
    
    // this is the matrix "M" coefficients
    for (i=0; i<num_feature_terms; i++) {
      fscanf(fp,"%lg",&x);
      M[current_feature*num_feature_terms+i]=x;
    }
    //read newline  
    fscanf(fp,"%c",&c);
    current_feature++;
  }
  free(tmp);
  fclose(fp);
  return 1;
}

// inverts m and puts result in minv
// zero means success, anything else failure
static int my_invert_matrix(const gsl_matrix * m, gsl_matrix * minv, int n)
{

  gsl_matrix *work = gsl_matrix_alloc(n,n); 
  
  int s, status;
  gsl_permutation * p = gsl_permutation_alloc (n);
  gsl_matrix_memcpy(work, m);
  status = gsl_linalg_LU_decomp (work, p, &s);
  if (!status)
    status = gsl_linalg_LU_invert (work, p, minv);
  

  gsl_permutation_free(p);
  gsl_matrix_free(work);
  return status;
}

// draw a random sample from id - according to id[*].prob
// which must be normalized
denotativeState * drawRandomProb(gsl_rng *rg, denotativeState * id)
{
  //double r=((double) rand())/((double) RAND_MAX);
  double r = gsl_rng_uniform_pos(rg);
  double p=0;
  int i=0;
  while (p<r) {
    p += id[i].prob;
    i++;
  }
  return id+i-1;
}
// draw a random sample from id - according to id[*].prob, but weighted such that id[0].prob <- min(1,id[0].prob+weight) 
// which must be normalized
denotativeState * drawWeightedProb(gsl_rng *rg, denotativeState * id, double weight)
{
  //double r=((double) rand())/((double) RAND_MAX);
  double r = gsl_rng_uniform_pos(rg);
  double p=0;
  int i=0;
  // the math here works - we can do this.
  //printf("p: %g   r: %g\n",p,r);
  while (p<r) {
    // max(0,-i) = (0>-i :? 0 : 1) means it only applies at i=0
    p += (id[i].prob+((0>-i)?0:1)*weight)/(1.0+weight);
    //printf("%d %g %g %g\n",i,id[i].prob,weight,p);
    i++;
  }
  return id+i-1;
}

// draw a random index according to prob, which must be normalized
int drawRandomProbDouble(gsl_rng *r, double *prob)
{
  //double r=((double) rand())/((double) RAND_MAX);
  double ther = gsl_rng_uniform_pos(r);
  double p=0;
  int i=0;
  while (p<ther) {
    p += prob[i];
    i++;
  }
  return i-1;
}

// initiailize the sampler state  - return the random seed used
void initializeBayesACT(dyad *the_dyad, agentState *the_agent)
//int numsamples, double alphaa, double alphab, double alphac,
//				  double betaa, double betac, int do_resets,
//				  int default_action, int use_emotions, int emotion_planning)
{
  the_dyad->dsstate  = (samplerState *) malloc(sizeof(samplerState));
  samplerState * sstate = the_dyad->dsstate;
 
  sstate->current_num_interactions = 0;
  sstate->sentiParams = (sentimentParams *) malloc(sizeof(sentimentParams));
  // remains unallocated
  // sstate->samples = (sample **) malloc(sizeof(sample *));
  sstate->numsamples=the_agent->numsamples;
  sstate->roughenNoise=pow(the_agent->numsamples,(-1.0/3.0));
  sstate->alphaa = the_agent->alphaa;
  sstate->alphab = the_agent->alphab;
  sstate->alphac = the_agent->alphac;
  sstate->betaa = the_agent->betaa;
  sstate->betac = the_agent->betac;
  

  sstate->num_resets=0;
  sstate->num_resets_agent=0;
  sstate->num_resets_client=0;
  sstate->num_resets_both=0;
  sstate->num_hard_resets=0;

  sstate->default_action = the_agent->default_action;
  sstate->do_resets = the_agent->do_resets;
  sstate->use_emotions = the_agent->use_emotions;
  sstate->emotion_planning = the_agent->emotion_planning;

  // set up stuff for random number generation
  gsl_rng_env_setup();
  sstate->T = gsl_rng_default;  // change the random number generator here
  sstate->r = gsl_rng_alloc (sstate->T);

  struct timeval time; 
  gettimeofday(&time,NULL);
  
  if (the_dyad->random_seed != 0)
    sstate->random_seed = the_dyad->random_seed;
  else  
    sstate->random_seed = (time.tv_sec * 1000) + (time.tv_usec) ;
  gsl_rng_default_seed = sstate->random_seed;
  gsl_rng_set(sstate->r, sstate->random_seed);
  
  sstate->T = gsl_rng_default;  // change the random number generator here
  sstate->r = gsl_rng_alloc (sstate->T);

  unsigned long testrn = 0;
  testrn=gsl_rng_get(sstate->r);
  // test for consistency
#ifdef VERBOSE
  printf ("generator type: %s\n", gsl_rng_name (sstate->r));
  printf ("seed = %lu\n", sstate->random_seed);
  printf ("first value = %lu\n", testrn);
#endif
  
  // useful to leave this on to debug errors because you can set a breakpoint in the
  // error handler, but otherwise should be off - see
  // https://www.gnu.org/software/gsl/manual/gsl-ref.html#Error-Handling
  gsl_set_error_handler_off();
}
// set up a Dyadic Interaction by setting up the necessary matrices (alpha, beta, etc).
// and then creating a sample set
// creates and intitializes a new interaction as a sentimentParams + a set of samples
// the index of the interaction is returned such that sstate->sentiParams[index] and
// sstate->samples[index] give the parameters and samples, respectively
int setUpDyadicInteraction(dyad *the_dyad, agentState * the_agent)
{
  // gets allocated in intitializeBayesACT
  //the_dyad->dsstate = (samplerState*) malloc(sizeof(samplerState));
  samplerState *sstate  = the_dyad->dsstate;
  sstate->sentiParams = (sentimentParams *) malloc(sizeof(sentimentParams));
  if (setupParameters(sstate, the_agent->dyn_fname, the_agent->mod_fname))
    return -1;

  // MAIN ALLOCATION of memory for samples for this dyad
  // sample is statically allocated (e.g has a fixed size)
  sstate->samples = (sample *) malloc(sstate->numsamples*sizeof(sample));
  
  // make sure all identities exist in the dictionary - otherwise bail out
  for (int i=0; i<the_agent->num_persona_ids; i++) {
    if (lookUp(the_agent->persona_ids+i,the_agent->entity_dicts[AGENT])<0) {
      printf("identity <%s> not in dictionary\n",the_agent->persona_ids[i].label);
      return -1;
    }
#ifdef VERBOSE
    printf("********identity <%s> for agent %d in dictionary at position %d\n",the_agent->persona_ids[i].label,i,the_agent->persona_ids[i].index);
#endif
  }

  
  for (int i=0; i < the_dyad->num_agent_ids; i++) {
    if (lookUp(the_dyad->agent_ids+i,the_agent->entity_dicts[AGENT])<0) {
      printf("identity <%s> not in dictionary\n",the_dyad->agent_ids[i].label);
      return -1;
    }
#ifdef VERBOSE
    printf("********identity <%s> for agent %d in dictionary at position %d\n",the_dyad->agent_ids[i].label,i,the_dyad->agent_ids[i].index);
#endif
  }
  for (int i=0; i<the_dyad->num_client_ids; i++) {
    if (lookUp(the_dyad->client_ids+i,the_agent->entity_dicts[CLIENT])<0) {
      printf("identity <%s> not in dictionary\n",the_dyad->client_ids[i].label);
      return -1;
    }
#ifdef VERBOSE
    printf("*******identity <%s> for agent %d in dictionary at position %d\n",the_dyad->client_ids[i].label,i,the_dyad->client_ids[i].index);
#endif
  }

  getSampleSet(the_dyad,
	       the_dyad->agent_ids, the_dyad->num_agent_ids,
	       the_dyad->client_ids, the_dyad->num_client_ids, the_agent);
  return 0;

}

// gets a set of samples by drawing random samples from the distribution over agent_ids in agent_id
// and that over client ids in client_id, then drawing random samples for those identities using
// the dictionary
// TODO: make this more general so agent_id and client_id are in a single denotativeState ** array
// because we may more generally want to include e.g. settings or other denotative state here
void getSampleSet(dyad *the_dyad,
		  identity * agent_id, int num_agent_ids,
		  identity * client_id, int num_client_ids,
		  agentState * the_agent)
{
  int i,k;
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  
  //double sig=0.1;
  sample * samples = sstate->samples;
  int num_samples = sstate->numsamples;

  for (k=0; k<num_samples; k++) {
    // draw an agent and client id
    // if the distributions passed in are NULL, it means the id is unknown, so
    // a random id is chosen from which a sample will be drawn. This will yield
    // a random sample from the distributions over all id sentiments in the dictionary.
    // this should be actually a property of the APP Function?
    // indeed it should - because for example if I am initializing an agent with two possible identity-pairs (e.g. citizen-citizen and environmentalist-culprit)
    // then I don't want samples with citizen-culprit ... or maybe thye will get ruled out anyways as being so unlikely
    // but in general, the method for drawing the initial sample set should be APP dependent
    if (agent_id==NULL) 
      samples[k].x[ID_AGENT] = gsl_rng_uniform_int(sstate->r, the_agent->entity_dicts[AGENT]->num_items);
    else
      samples[k].x[ID_AGENT] = drawRandomProb(sstate->r,agent_id)->index;

    if (client_id==NULL)
      samples[k].x[ID_CLIENT] = gsl_rng_uniform_int(sstate->r, the_agent->entity_dicts[CLIENT]->num_items);
    else {
      // omg horrible hack  ---- this is to ensure that pairs are matched up
      if (samples[k].x[ID_AGENT] == agent_id[0].index) {  // citizen;
	samples[k].x[ID_CLIENT] = client_id[0].index;  // relies on citizen being first
      } else {
	samples[k].x[ID_CLIENT] = client_id[1].index;
      }
    }
    //printf("index for agent and client is %d %d %g\n",indices[0],indices[1],client_id[0].prob);
    //draw a fundamental sentiment for agent (j=0) and then for client (j=1)
    somaticTransform(CONNOTATIVE, sstate->r, dyad_index, samples[k].x+ID_AGENT, samples[k].f+SSD*AGENT, the_agent->entity_dicts[AGENT]);
    somaticTransform(CONNOTATIVE, sstate->r, dyad_index, samples[k].x+ID_CLIENT, samples[k].f+SSD*CLIENT, the_agent->entity_dicts[CLIENT]);
    
    // the behaviour component doesn't matter for the initial state
    for (i=0; i < SSD; i++) 
      samples[k].f[i+BEHAVIOUR*SSD]=0.0;
    samples[k].x[BEH_AGENT]=-1; // not set
    samples[k].x[BEH_CLIENT]=-1; // not set

    // we are setting up samples, so we don't know who acted for them yet. Assume AGENT but this is more of a placeholder
    samples[k].who_acted=BEH_AGENT;
    // initialize the rest of the xx state
    initializeXXstate(&(samples->xx));

    // the transient is the same as the fundamental
    memcpy(samples[k].tau,samples[k].f,VD*sizeof(double));

    samples[k].weight=1.0;

  }
}
// function that sets up the various parameters.
// Arguments are:
//  -dyn_fname:  dynamcis file name -- need to have one for males and one for females
//  -mod_fname:  modifiers dynamics file name
// Produces:
//  -agent index: indexes an element of sstate so this agent can be used
//  returns -1 on error 
int setupParameters(samplerState * sstate, char * dyn_fname, char *mod_fname)
{
  //double alpha, beta, alpha2, beta2;
  double alphaa = sstate->alphaa;
  double alphab = sstate->alphab;
  double alphac = sstate->alphac;
  double betaa = sstate->betaa;
  double betac = sstate->betac;
  
  int i,j;
  sentimentParams * __agent_params = sstate->sentiParams;

  __agent_params->num_features = getNumFeaturesFromDynamicsFile(dyn_fname);
  if (mod_fname !=NULL) {
    __agent_params->num_features_modifiers = getNumFeaturesFromDynamicsFile(mod_fname); 
  } else {
    __agent_params->num_features_modifiers = 0;
  }


  //covariance matrices
  __agent_params->sig_a = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_a = gsl_matrix_alloc (VD, VD);
  __agent_params->sig_f = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_f = gsl_matrix_alloc (VD, VD);
  //versions for reset step
  __agent_params->sig_a_reset = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_a_reset = gsl_matrix_alloc (VD, VD);
  __agent_params->sig_f_reset_agent = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_f_reset_agent = gsl_matrix_alloc (VD, VD);
  __agent_params->sig_f_reset_client = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_f_reset_client = gsl_matrix_alloc (VD, VD);
  __agent_params->sig_f_reset_both = gsl_matrix_alloc (VD, VD);
  __agent_params->isig_f_reset_both = gsl_matrix_alloc (VD, VD);
  

  
  double alphaa2 = alphaa*alphaa;
  double alphab2 = alphab*alphab;
  double alphac2 = alphac*alphac;
  double betaa2 = betaa*betaa;
  double betac2 = betac*betac;
  
#ifdef VERBOSE
  printf("num_features read in: %d\n",__agent_params->num_features);
  printf("vector dimensionality: %d\n",VD);
  printf("num_features read in for modifiers: %d\n",__agent_params->num_features_modifiers);
  //printf("epsilon_reset: %f\n",epsilon_reset);
  printf("alphaa: %g  alphab: %g alphac: %g  betaa: %g betac: %g \n",alphaa, alphab, alphac, betaa, betac);
  //printf("RESET values alphaa: %g  alphab: %g alphac: %g betaa: %g betac: %g \n",alpha_valuea_reset, alpha_valueb_reset, alpha_valuec_reset, beta_valuea_reset, beta_valuec_reset);
  printf("RESET THRESHOLD:  %g\n",RESET_THRESHOLD);
#endif


  //all initialized diagonal
  for (i=0; i<VD; i++) {
    for (j=0; j<VD; j++) {
      if (i!=j) {
	gsl_matrix_set(__agent_params->sig_a, i, j, 0.0);
	gsl_matrix_set(__agent_params->sig_f, i, j, 0.0);
	gsl_matrix_set(__agent_params->sig_a_reset, i, j, 0.0);
	gsl_matrix_set(__agent_params->sig_f_reset_agent, i, j, 0.0);
	gsl_matrix_set(__agent_params->sig_f_reset_client, i, j, 0.0);
	gsl_matrix_set(__agent_params->sig_f_reset_both, i, j, 0.0);
      } else {
	if (i < SSD) {
	  gsl_matrix_set(__agent_params->sig_a, i, j, alphaa2);
	  gsl_matrix_set(__agent_params->sig_a_reset, i, j, alphaa2);
	  gsl_matrix_set(__agent_params->sig_f, i, j, betaa2);
	  //CHANGES MARCH 2021: on agent reset, we want this to be very small as we will be "shoehorning" in the ACT optimal
	  gsl_matrix_set(__agent_params->sig_f_reset_agent, i, j, 1e-11); 
	  // CHANGES MARCH 2021: on client reset, we want the agent id to stay the same, let the client do all the resetting
	  gsl_matrix_set(__agent_params->sig_f_reset_client, i, j, 1e-11);
	  //CHANGES MARCH 2021: on both reset, we want this to be very small as we will be "shoehorning" in the ACT optimal
	  gsl_matrix_set(__agent_params->sig_f_reset_both, i, j, 1e-11); 
	} else if (i < SSD*2) {
	  gsl_matrix_set(__agent_params->sig_a, i, j, alphab2);
	  gsl_matrix_set(__agent_params->sig_a_reset, i, j, alphab2);
	  gsl_matrix_set(__agent_params->sig_f, i, j, INFINITY);  // used to have to set this really small and "shoehorn in" f_b on agent turn
	  // CHANGES MARCH 2021: this is correct, as we want the "shoehorned in" 
	  gsl_matrix_set(__agent_params->sig_f_reset_agent, i, j,  1e-11); 
	  gsl_matrix_set(__agent_params->sig_f_reset_client, i, j, 1e-11); 
	  gsl_matrix_set(__agent_params->sig_f_reset_both, i, j,   1e-11); 
	} else {
	  gsl_matrix_set(__agent_params->sig_a, i, j, alphac2);
	  gsl_matrix_set(__agent_params->sig_a_reset, i, j, alphac2);
	  gsl_matrix_set(__agent_params->sig_f, i, j, betac2);
	  // CHANGES MARCH 2021: on agent reset, we want the client id to stay the same, let the agent do all the resetting
	  gsl_matrix_set(__agent_params->sig_f_reset_agent, i, j, 1e-11);
	  //CHANGES MARCH 2021: on client reset, we want this to be very small as we will be "shoehorning" in the ACT optimal
	  gsl_matrix_set(__agent_params->sig_f_reset_client, i, j, 1e-11);
	  //CHANGES MARCH 2021: on both reset, we want this to be very small as we will be "shoehorning" in the ACT optimal
	  gsl_matrix_set(__agent_params->sig_f_reset_both, i, j, 1e-11); 
	}
      }
    }
  }	

  // inverted versions -- these may be singular since I use INFINITY for some of them
  // so I ignore the error check 
  int status;
  status=my_invert_matrix(__agent_params->sig_a,__agent_params->isig_a,VD);
  status=my_invert_matrix(__agent_params->sig_f,__agent_params->isig_f,VD);
  status=my_invert_matrix(__agent_params->sig_a_reset,__agent_params->isig_a_reset,VD);
  status=my_invert_matrix(__agent_params->sig_f_reset_client,__agent_params->isig_f_reset_client,VD);
  status=my_invert_matrix(__agent_params->sig_f_reset_agent,__agent_params->isig_f_reset_agent,VD);
  status=my_invert_matrix(__agent_params->sig_f_reset_both,__agent_params->isig_f_reset_both,VD);


  __agent_params->mu_n = gsl_vector_alloc (VD);
  __agent_params->sig_n = gsl_matrix_alloc (VD, VD);

  __agent_params->interact_mean = gsl_vector_alloc (VD);

  __agent_params->curlyK = gsl_matrix_alloc(VD,VD);
  __agent_params->curlyKinv = gsl_matrix_alloc(VD,VD);
  __agent_params->curlyKa = gsl_matrix_alloc(VD,VD);
  __agent_params->curlyKainv = gsl_matrix_alloc(VD,VD);
  __agent_params->curlyKc = gsl_matrix_alloc(VD,VD);
  __agent_params->curlyKcinv = gsl_matrix_alloc(VD,VD);

  __agent_params->curlyH = gsl_matrix_alloc(VD,SSD);
  __agent_params->curlyHa = gsl_matrix_alloc(VD,SSD);
  __agent_params->curlyHc = gsl_matrix_alloc(VD,SSD);

  __agent_params->curlyC = gsl_vector_alloc(VD);
  __agent_params->curlyCa = gsl_vector_alloc(VD);
  __agent_params->curlyCc = gsl_vector_alloc(VD);
  gsl_matrix_set_identity(__agent_params->curlyK);
  gsl_matrix_set_identity(__agent_params->curlyKa);
  gsl_matrix_set_identity(__agent_params->curlyKc);
  gsl_matrix_set_zero(__agent_params->curlyH);
  gsl_matrix_set_zero(__agent_params->curlyHa);
  gsl_matrix_set_zero(__agent_params->curlyHc);
  gsl_vector_set_zero(__agent_params->curlyC);
  gsl_vector_set_zero(__agent_params->curlyCa);
  gsl_vector_set_zero(__agent_params->curlyCc);
  
  __agent_params->m_index = (int **) malloc(__agent_params->num_features*sizeof(int*));
  __agent_params->m_num_features = (int *) malloc(__agent_params->num_features*sizeof(int));
  __agent_params->M = (double *) malloc(__agent_params->num_features*VD*sizeof(double));

  if (readDynamicsFile(VD,__agent_params->m_index,__agent_params->m_num_features,__agent_params->M,dyn_fname) < 1) {
    fprintf(stderr,"error reading dynamics file");
    return -1;
  }

  if ( __agent_params->num_features_modifiers > 0) {
    __agent_params->dreq_index = (int **) malloc(__agent_params->num_features_modifiers*sizeof(int*));
    __agent_params->dreq_num_features = (int *) malloc(__agent_params->num_features_modifiers*sizeof(int));
    __agent_params->dREQ = (double *) malloc(__agent_params->num_features_modifiers*SSD*sizeof(double));
    
    if (readModifiersDynamicsFile(__agent_params->num_features_modifiers,__agent_params->dreq_index,
				  __agent_params->dreq_num_features,__agent_params->dREQ,mod_fname) < 1) {
      fprintf(stderr,"error reading modifiers dynamics file");
      return -1;
    }
  }

  return 0;
}

// must call freeParameters on all allocated agents first
// frees up core state and random number generator
void freeBayesACT(samplerState *sstate)
{
  free(sstate->samples);
  free(sstate->sentiParams);
  gsl_rng_free(sstate->r);
  free(sstate);
}

// free they dyad structures
static int freeDyads(dyad * the_dyad)
{
  if (the_dyad->num_agent_ids > 0) 
    free(the_dyad->agent_ids);
  if (the_dyad->num_client_ids > 0)
    free(the_dyad->client_ids);
  return 0;
}

//free the parameters overall from sstate
int freeParameters(dyad *the_dyad)
{
  freeParametersFromIndex(the_dyad->dsstate);
  freeBayesACT(the_dyad->dsstate);
  freeDyads(the_dyad);
  return 0;
}

// free up parameters set up above
// takes one argument - the agent index
int freeParametersFromIndex(samplerState * sstate)
{
  
  sentimentParams * __agent_params = sstate->sentiParams;

  gsl_matrix_free (__agent_params->sig_a);
  gsl_matrix_free (__agent_params->sig_f);
  gsl_matrix_free (__agent_params->isig_a);
  gsl_matrix_free (__agent_params->isig_f);

  gsl_matrix_free (__agent_params->sig_a_reset);
  gsl_matrix_free (__agent_params->isig_a_reset);
  gsl_matrix_free (__agent_params->sig_f_reset_agent);
  gsl_matrix_free (__agent_params->isig_f_reset_agent);
  gsl_matrix_free (__agent_params->sig_f_reset_client);
  gsl_matrix_free (__agent_params->isig_f_reset_client);
  gsl_matrix_free (__agent_params->sig_f_reset_both);
  gsl_matrix_free (__agent_params->isig_f_reset_both);

  gsl_matrix_free (__agent_params->curlyK);
  gsl_matrix_free (__agent_params->curlyKa);
  gsl_matrix_free (__agent_params->curlyKc);
  gsl_matrix_free (__agent_params->curlyH);
  gsl_matrix_free (__agent_params->curlyHa);
  gsl_matrix_free (__agent_params->curlyHc);
  gsl_matrix_free (__agent_params->curlyKainv);
  gsl_matrix_free (__agent_params->curlyKcinv);
  gsl_matrix_free (__agent_params->curlyKinv);
  gsl_vector_free (__agent_params->curlyC);
  gsl_vector_free (__agent_params->curlyCa);
  gsl_vector_free (__agent_params->curlyCc);


  // free the impression features
  for (int i=0; i<__agent_params->num_features; i++) 
    free(__agent_params->m_index[i]);
  free(__agent_params->m_num_features);
  free(__agent_params->m_index);
  free(__agent_params->M);

  // free the modifiers
  if ( __agent_params->num_features_modifiers > 0) {
    for (int i=0; i < SSD; i++) 
      free(__agent_params->dreq_index[i]);
    free(__agent_params->dreq_num_features);
    free(__agent_params->dreq_index);
    free(__agent_params->dREQ);
  }
  
  gsl_vector_free(__agent_params->mu_n);
  gsl_matrix_free(__agent_params->sig_n);
  gsl_vector_free(__agent_params->interact_mean);

  return 0;
  
}
// computes the somatic transform between an index (denotative label)
// and a connotative EPA vector.
// usage 1. compute the P(connotative_epa | sentiment profile for the behaviour indexed by denotative_index in dictionary
// in general, somaticTransform(I,F,dict,dtype) should:
// I is pointer to denotative index  - this must be an index in the dictionary
// F is pointer to the connotative vector - this must be an array of length SSD
// SATtype defines which of these will be performed
// if I,F are both non-null (sat_type == WEIGHT)
//     compute P(F|I) -- returns a double "weight" or probability density (point) estimate
// if F is null and I is not  (sat_type == CONNOTATIVE)
//     draw a sample from P(F|I) --- could return the weight as well if needed? 
// if I is null and F is not  (sat_type == DENOTATIVE)
//     draw a sample from P(I|F)  <--- this is the nearest neighbour stuff
// -------- possibly further complications, but these may be better suited in a different function that calls this one
// if I is an array and F is an EPA vector
//     draw a sample from P(I|F)
//----------------
double somaticTransform(SATtype sat_type, gsl_rng *r, int dyad_index,			
			int *denotative_index, double * connotative_epa,
			dictionary * dict)
{
  int i;
  sentiment * the_dictionary = dict->theDictionary;

  double weight=0.0;
  // if I,F are both non-null:
  //     compute P(F|I) -- returns a double "weight" or probability density (point) estimate
  if (sat_type == WEIGHT) {
    // ****************************************************************************************
    // WEIGHT Somatic Transform -- compute the distance from connotative_epa to denotative_index in the dictionary
    // taking into account the distribution for denotative index in the dictionary
    // ****************************************************************************************
    weight=1.0;
    if (dict->dtype==MEAN) {
      for (int j=0; j < SSD; j++)
	weight *= exp(-0.5*pow(gsl_vector_get(the_dictionary[*denotative_index].epa,j)-connotative_epa[j],2)/(the_dictionary[*denotative_index].var));
    } else if (dict->dtype==SD) {
      for (int j=0; j < SSD; j++)
	weight *= exp(-0.5*pow(gsl_vector_get(the_dictionary[*denotative_index].epa,j)-connotative_epa[j],2)/
		      (gsl_vector_get(the_dictionary[*denotative_index].variance,j)));
    } else if (dict->dtype==COV) {
      gsl_vector * x = gsl_vector_alloc(SSD);
      gsl_vector * work = gsl_vector_alloc(SSD);
      for (int j=0; j < SSD; j++) 
	gsl_vector_set(x,j,connotative_epa[j]);
#ifdef GSL_VERSION_2
      gsl_ran_multivariate_gaussian_pdf (x, the_dictionary[*denotative_index].epa, the_dictionary[*denotative_index].cholesky, &weight, work);
#else
      weight=dmvnorm(SSD,x,the_dictionary[*denotative_index].epa, the_dictionary[*denotative_index].covariance);
#endif
      gsl_vector_free(x);
      gsl_vector_free(work);
    }
  } else if (sat_type == CONNOTATIVE) {
    // ****************************************************************************************
    // CONNOTATIVE Somatic Transform -- get connotative_epa from denotative index by looking it up in the the_dictionary and
    // drawing a sample from the resulting distribution
    // and we want to draw a sample from denotative_index
    // in an area of high density for P(connotative_epa | denotative_index)
    // ****************************************************************************************
    if (dict->dtype==MEAN) {
      for (i=0; i < SSD; i++) 
	connotative_epa[i] = gsl_vector_get(the_dictionary[*denotative_index].epa,i) +
	  gsl_ran_gaussian (r,the_dictionary[*denotative_index].var);
    } else if (dict->dtype==SD) {
      for (i=0; i < SSD; i++) 
	connotative_epa[i] = gsl_vector_get(the_dictionary[*denotative_index].epa,i) +
	  gsl_ran_gaussian (r,gsl_vector_get(the_dictionary[*denotative_index].variance,i));
    } else if (dict->dtype==COV) {
      gsl_vector *fsample = gsl_vector_alloc(SSD);
      // draw from a Gaussian distribution
#ifdef GSL_VERSION_2
      gsl_ran_multivariate_gaussian (r,the_dictionary[*denotative_index].epa,the_dictionary[*denotative_index].cholesky,fsample);
#else
      rmvnorm(r, SSD, the_dictionary[*denotative_index].epa,the_dictionary[*denotative_index].covariance,fsample);
#endif      
      memcpy(connotative_epa,gsl_vector_ptr(fsample,0),SSD*sizeof(double));
      gsl_vector_free(fsample);
    }
    weight=0.0;
  } else if (sat_type == DENOTATIVE) {
    // ****************************************************************************************
    // DENOTATIVE Somatic Transform -- get denotative index from connotative epa by comparing to 
    // each the_dictionary entry and finding the closest one
    // ****************************************************************************************
    double how_far, min_how_far; //, max_weight;
    int best_index,new_best_index;
    int num_items_to_do,converted_i;
    // here an exhaustive computation over all the_dictionary items is rather unfortunate
    // TODO: replace exhaustive computation with something smarter
    // This could be e.g. a nearest neighbours computation in EPA space for example - get the 10 closest neighbours and randomly
    // sample from those
    // this works in two ways - and here, I'm choosing based only on mimic_iteract, but we might want to use the eucidean approach
    // for other reasons as well
    // (1) compute the euclidean distance only and return the closest one
    // (2) compute P(connotative_epa | denotative_index) and draw a sample from that
    //       so for dict->dtype==MEAN, this is euclidean distance
    //                           SD/COV, this is weighted euclidean connotative'*COV^{-1}*connotative
    //                           SAMPLE, this would be distance to the nearest sample? or a density estimate would be better
    //       in general, its just another WEIGHT somatic transform! 
    // we could also return a complete sorted list based on distance (could do that here since we're computing it all anyways!)
    double *prob_items;
    // this is a problem. The somatic tansform slows this down by a factor of 2. Therefore, we do this hack where we
    // essentially assume a variance of 1.0 - so compute euclidean distance.
    // flick this to 1 in order to speed up by a factor of 2, but not have the DENOTATIVE Somatic Transform use the actual variance on the concept or \gamma setting
    // denotative somatic transform is used to "copy out" the x_b label that describes the sentiment of the optimal action prediction as computed by the connotative system
    // also on reset to label newly assigned identities. So using a variance of 1.0 on these is perhaps not quite such a big deal.
    // It means the ranks of the labels in BayesACT that describe the ongoing event may be off, but I expect the same behaviours to be present.
    // ALSO: always_do_euclidean=1 SHOULD track Interact better, but this has still to be verified...
    int always_do_euclidean=0;
    // if using a MEAN the_dictionary, then this doesn't make a difference other than to not use whatever setting of gamma was chosen when computing labels.
    // can always do euclidean (since its faster) if dtype is MEAN since the answers will always be the same (the distance may change though) 
    if (dict->dtype==MEAN) 
      always_do_euclidean=1;
    weight=0.0;
    // if this dictionary has no institutionally correct items, we will just do all of them
    num_items_to_do=dict->num_items;
    if (dict->num_ic_items!=NULL && dict->num_ic_items[dyad_index]>0)
      num_items_to_do=dict->num_ic_items[dyad_index];
    
    if (!always_do_euclidean) 
      prob_items = (double *) malloc(num_items_to_do*sizeof(double));
    best_index=1;
    //fprintf(stderr,"best index: %d  ic_converter %d\n",best_index,dict->list_ic_converter[dyad_index][best_index]);

    // the dictionary needs to be implemented so this loop can be avoided 
    for (i=0; i < num_items_to_do; i++) {

      converted_i = i;
      if (dict->num_ic_items!=NULL && dict->num_ic_items[dyad_index]>0)
	converted_i = dict->list_ic_converter[dyad_index][i];
    
      
      if (always_do_euclidean) {
	how_far=euclidean_distance(connotative_epa,gsl_vector_ptr(the_dictionary[converted_i].epa,0),SSD);
	//how_far=cosine_distance(connotative_epa,gsl_vector_ptr(the_dictionary[i].epa,0),SSD);
	if (i==0 || how_far < min_how_far) {
	  min_how_far = how_far;
	  best_index=i;
	}
	// factor of 10.0 here totally arbitrary
	weight=exp(-10.0*min_how_far);
      } else {
	// this is the correct method, but slows things down a LOT (TODO: why?)
	// POSSIBLY here - if this identity passes institutional filters, then
	// we do the somaticTransform, otherwise, we assign weight 0
	
	prob_items[i] = somaticTransform(WEIGHT, r, dyad_index, &converted_i, connotative_epa, dict);
	weight += prob_items[i];
      }
    }
    if (!always_do_euclidean) {
      if (weight > MIN_SAMPLE_TOTAL_WEIGHT) {
	for (i=0; i < num_items_to_do; i++) 
	  prob_items[i] /= weight;
      } else {
	for (i=0; i < num_items_to_do; i++) 
	  prob_items[i]  = 1.0/((double) num_items_to_do);
      }
      best_index = drawRandomProbDouble(r, prob_items);
      free(prob_items);
    }
    new_best_index=best_index;
    if (dict->num_ic_items!=NULL && dict->num_ic_items[dyad_index]>0)
      new_best_index = dict->list_ic_converter[dyad_index][best_index];

    if (new_best_index > dict->num_items || new_best_index < 0) {
      fprintf(stderr,"ERROR in dict_ic_converter for dyad %d best index was %d and coverted to %d\n",dyad_index,best_index,new_best_index);
      //crash
      exit(1);
    }
    *denotative_index = new_best_index;
  }
  return weight;
}

// roughen the sample - add some random white noise in [-roughenNoise,+roughenNoise]
// the identity components of the fundamental sentiment only
void roughenSample(samplerState *sstate, sample samp, double roughenNoise) 
{
  double r;
  for (int i=0; i<SSD; i++) {
    r = (2*gsl_rng_uniform(sstate->r)-1.0)*roughenNoise;
    samp.f[i] += r;
    r = (2*gsl_rng_uniform(sstate->r)-1.0)*roughenNoise;
    samp.f[i+2*SSD] +=r;
  }
}
void roughenAllSamples(samplerState *sstate, double roughNoise) 
{
  for (int i=0; i<sstate->numsamples; i++)
    roughenSample(sstate,sstate->samples[i],roughNoise);
}

// resample part of the samples from the persona_id
void resampleFromPersona(samplerState *sstate, int dyad_index, agentState *the_agent, double fraction) 
{
  for (int i=0; i<sstate->numsamples; i++) {
    if (gsl_rng_uniform(sstate->r) < fraction) {
      sstate->samples[i].x[ID_AGENT] = drawRandomProb(sstate->r,the_agent->persona_ids)->index;
      somaticTransform(CONNOTATIVE,sstate->r, dyad_index, sstate->samples[i].x+ID_AGENT,
		       sstate->samples[i].f+SSD*AGENT, the_agent->entity_dicts[AGENT]);
      memcpy(sstate->samples[i].tau+SSD*AGENT, sstate->samples[i].f+SSD*AGENT, SSD*sizeof(double));

      // cheat and use agent persona as client's
      /*
      sstate->samples[dyad_index][i].x[ID_CLIENT] = drawRandomProb(sstate->r,the_agent->persona_ids)->index;
      somaticTransform(CONNOTATIVE,sstate->r,dyad_index,sstate->samples[dyad_index][i].x+ID_CLIENT,
      		       sstate->samples[dyad_index][i].f+SSD*CLIENT,the_agent->entity_dicts[CLIENT]);
      memcpy(sstate->samples[dyad_index][i].tau+SSD*CLIENT,sstate->samples[dyad_index][i].f+SSD*CLIENT,SSD*sizeof(double));
      */
      sstate->samples[i].reset_flag=2;  // mark same as a successfull reset
    }
  }
}
		   
// re-sample a weighted set of samples, to get an unweighted set new_samples
// the method used is to calculate the cumulative weights, then to uniformly
// pick samples according to these cumulative weights
void reSampleInPlace(sample * samples, int numsamples)
{
  sample *new_samples = (sample *) malloc(numsamples*sizeof(sample));
  reSample(samples, new_samples, numsamples);
  memcpy(samples, new_samples, numsamples*sizeof(sample));
  free(new_samples);
}

void reSample(sample * samples, sample * new_samples, int numsamples)
{
  // resample - put resampled samples back in new_samples
  if (numsamples==0)
    return;
  double * cum_weights = (double *) malloc(numsamples*sizeof(double));
  double weight_so_far=0.0;
  double totalweight=0.0;
  double rsn=0;
  int i,k,j=0;

  // get the total weight of all samples
  for (i=0; i<numsamples; i++) 
    totalweight += samples[i].weight;
  //printf("totalweight: %g\n",totalweight);
  // compute the cumulative weight cum_weights[i]=weight[i]+cum_weights[i-1] ... 
  for (i=0; i<numsamples; i++) {
    weight_so_far += samples[i].weight;
    cum_weights[i] = weight_so_far/totalweight;
  }
  
  for (i=0; i<numsamples; i++) {
    while (rsn>=cum_weights[j]) {
      j++;
    }
    rsn += 1.0/numsamples;
    memcpy(new_samples[i].f,samples[j].f,VD*sizeof(double));
    memcpy(new_samples[i].tau,samples[j].tau,VD*sizeof(double));
    for (k=0; k<NGTD; k++) 
      new_samples[i].x[k]=samples[j].x[k];
    //XX:
    copyXXstate(&(new_samples[i].xx),&(samples[j].xx));
    new_samples[i].weight=1.0;
    new_samples[i].who_acted = samples[j].who_acted;
    new_samples[i].reset_flag = samples[j].reset_flag;
  }
  free(cum_weights);
  
}
// this version will do two resamplings, one for each who_acted
// so the number of samples that were used for each actor stay the same and
// are resampled as such. The function works by first sorting the samples
// so all the ones with who_acted=BEH_AGENT come first, then calling reSample on each part
void reSampleByActor(sample * samples, sample * new_samples, int numsamples)
{
  int i,j;
  int numsamples_agent=0, numsamples_client=0;
  sample tmp_sample;
  i=0;  // is the first client sample, may be negative if they are all client, and may be numsamples if they are all agent
  j=numsamples-1;  // is the last agent sample, may be numsamples if they are all agent
  while (i<j) {
    while (i<=j && samples[i].who_acted==BEH_AGENT)
      i++;
    
    while (j>=i && samples[j].who_acted==BEH_CLIENT)
      j--;
    if (i < j) {
      // swap samples[i] and samples[j]     
      memcpy(&tmp_sample,samples+j,sizeof(sample));
      memcpy(samples+j,samples+i,sizeof(sample));
      memcpy(samples+i,&tmp_sample,sizeof(sample));
      i++;
      j--;
    }
  }
  numsamples_agent = i;
  numsamples_client = numsamples-numsamples_agent;
#ifdef VERYVERBOSE
  printf("numsamples_agent: %d numsamples_client: %d\n",numsamples_agent,numsamples_client);
#endif
  reSample(samples, new_samples, numsamples_agent);
  reSample(samples+numsamples_agent, new_samples+numsamples_agent, numsamples_client);
  
}
// draw a single sample
int drawSample(gsl_rng *r, sample * samples, int numsamples)
{
  double weight_so_far=0.0;
  double totalweight=0.0;
  int i;

  double rsn = gsl_rng_uniform_pos(r);

  // get the total weight of all samples
  for (i=0; i<numsamples; i++) 
    totalweight += samples[i].weight;

  if (totalweight < MIN_SAMPLE_TOTAL_WEIGHT) {
    // draw a random sample
    return ((int) floor(numsamples*rsn));
  }
  // draw the sample
  for (i=0; i<numsamples; i++) {
    weight_so_far += samples[i].weight;
    if (weight_so_far/totalweight > rsn) 
      return i;
  }
  return -1;
}
// worker function to actually sample a sentiment from a set of sentimentParams (which must have been initialized)
// puts the results in myfout and mytout
int sampleASentiment(samplerState *sstate, double *myfout, double *mytout)
{
  sentimentParams * __agent_params = sstate->sentiParams;

  gsl_vector *fsample = gsl_vector_alloc(VD);
  gsl_vector *workv = gsl_vector_alloc(VD);

  gsl_matrix *workm = gsl_matrix_alloc(VD,VD);
  // this is the part that actually does the sample draw from the multivariate normal
  gsl_matrix_memcpy(workm,__agent_params->sig_n);
  //printMatrix(__agent_params->sig_n,"sig_n",VD,VD);
  int status=1;
#ifdef GSL_VERSION_2
  status = gsl_linalg_cholesky_decomp1(workm);
#else
  status = gsl_linalg_cholesky_decomp(workm);
#endif

  if (!status) { // no error
    
    //printVector(__agent_params->mu_n,"mu",VD);
    //printMatrix(__agent_params->sig_n,"sig",VD,VD);
#ifdef GSL_VERSION_2
    gsl_ran_multivariate_gaussian (sstate->r,__agent_params->mu_n,workm,fsample);
#else
    rmvnorm(sstate->r, VD, __agent_params->mu_n,workm,fsample);
#endif
    // copy out fsample to myfout
    memcpy(myfout,gsl_vector_ptr(fsample,0),VD*sizeof(double));

    // copy out curlyC to workv
    gsl_vector_memcpy(workv, __agent_params->curlyC);

    // now compute the tau sample 
    gsl_vector_view fb = gsl_vector_subvector(fsample,SSD,SSD);
    gsl_blas_dgemv(CblasNoTrans, 1.0, __agent_params->curlyH, &fb.vector, 1.0, workv);

    // copy out curlyC to output mytout
    memcpy(mytout,gsl_vector_ptr(workv,0),VD*sizeof(double));
  }
  
  gsl_vector_free(fsample);
  gsl_vector_free(workv);
  gsl_matrix_free(workm);


  return status;
  
}
// compute the deflection of a single sample
double computeDeflectionOfSample(sample * the_sample, c_entity_type who_acts)
{
  double deflection=0.0;
  for (int k=0; k < SSD; k++) 
    deflection += pow((the_sample->f[(who_acts)*SSD+k] - the_sample->tau[(who_acts)*SSD+k]),2);
  deflection *= the_sample->weight;
  return deflection;
}

// computes the deflection - computes the deflection only over those components specified by who_acts
double computeDeflection(samplerState *sstate, c_entity_type who_acts)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;
  double ldef=0.0;
  double deflection=0.0;
  double totalweight=0.0;

  for (int i=0; i<numsamples; i++) {
    ldef = computeDeflectionOfSample(the_samples+i, who_acts);
    deflection += ldef;
    //for (int k=0; k < SSD; k++) 
    //deflection += pow((the_samples[i].f[who_acts*SSD+k] - the_samples[i].tau[who_acts*SSD+k]),2)*the_samples[i].weight;
    totalweight += the_samples[i].weight;
  }
  return deflection/totalweight;
}
// computes the variance of deflection - computes the variance of deflection only over those components specified by who_acts
double computeDeflectionVariance(samplerState *sstate,  c_entity_type who_acts)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;
  double ldef=0.0;
  double deflection=0.0;
  double deflectionvar=0.0;
  double totalweight=0.0;
  
  for (int i=0; i<numsamples; i++) {
    ldef = computeDeflectionOfSample(the_samples+i, who_acts);
    deflection += ldef;
    //for (int k=0; k < SSD; k++) 
    //deflection += pow((the_samples[i].f[who_acts*SSD+k] - the_samples[i].tau[who_acts*SSD+k]),2)*the_samples[i].weight;
    totalweight += the_samples[i].weight;
    deflectionvar += ldef*ldef/(the_samples[i].weight);  // only one weight!
  }
  // naive version
  deflectionvar = deflectionvar/totalweight;
  deflection = deflection/totalweight;
  return (deflectionvar-deflection*deflection);
}
// computes binned deflection - computes the deflection only over those components specified by who_acts
// returns a structure
// NOT DONE YET!!
double computeDeflectionStatistics(samplerState *sstate, c_entity_type who_acts)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;
  double deflection=0.0;
  double totalweight=0.0;

  for (int i=0; i<numsamples; i++) {
    for (int k=0; k < SSD; k++) 
      deflection += pow((the_samples[i].f[who_acts*SSD+k] - the_samples[i].tau[who_acts*SSD+k]),2)*the_samples[i].weight;
    totalweight += the_samples[i].weight;
  }
  return deflection/totalweight;
}
// compute the deflection of the average sentiments and impressions.
// this should be closer to what Interact would compute
double computeDeflectionOfAverage(samplerState *sstate, c_entity_type who_acts)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;
  double deflection=0.0;
  double totalweight=0.0;

  double * f_avg, * tau_avg;
  f_avg = (double *) malloc(SSD*sizeof(double));
  tau_avg = (double *) malloc(SSD*sizeof(double));

  for (int i=0; i<SSD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }

  
  for (int i=0; i<numsamples; i++) {
    for (int j=0; j<SSD; j++) {
      f_avg[j] = f_avg[j]+the_samples[i].f[who_acts*SSD+j]*the_samples[i].weight;
      tau_avg[j] = tau_avg[j]+the_samples[i].tau[who_acts*SSD+j]*the_samples[i].weight;
    }
    totalweight += the_samples[i].weight;
  }

  for (int j=0; j < SSD; j++) {
    f_avg[j] = f_avg[j]/totalweight;
    tau_avg[j] = tau_avg[j]/totalweight;
  }

  for (int j=0; j < SSD; j++) 
    deflection += pow((f_avg[j] - tau_avg[j]),2);


  free(f_avg);
  free(tau_avg);
  
  
  return deflection;
}

//updateFullState and updateFullStateWithXXObs are just handler functions that set up calls to the other fuctions with NULL
// as arguments. In fact, however, these two first functions (single actor) are never used, as I think I moved to TwoActors
// when trying to get simultaneous actions to work
double updateFullState(dyad *the_dyad, agentState *the_agent,
		       behaviour * the_behaviour, identity * who_acted)
{
  return updateFullStateWithXXObs(the_dyad, the_agent, the_behaviour, who_acted, NULL);
}
// does a full state update, including weighting and re-sampling (and roughening - not implemented yet)
// and taking care of resets (re-identification) on a sample-based level
// the_samples are replaced with the new samples once done
// the normal update with a single actor

double updateFullStateWithXXObs(dyad *the_dyad, agentState *the_agent,
		       behaviour * the_behaviour, identity * who_acted, xostate *xx_observation)
{
  if (who_acted->index==AGENT) {
    return updateFullStateTwoActorsWithXXObs(the_dyad, the_agent, the_behaviour, NULL, xx_observation);
  } else {
    return updateFullStateTwoActorsWithXXObs(the_dyad, the_agent, NULL, the_behaviour, xx_observation);
  }
}
// does the full state update with an action by both agent and client (one may be NULL)
double updateFullStateTwoActors(dyad *the_dyad, agentState *the_agent,
				behaviour * agent_behaviour, behaviour * client_behaviour_observation)
{
  return updateFullStateTwoActorsWithXXObs(the_dyad, the_agent, agent_behaviour, client_behaviour_observation, NULL);
}
double updateFullStateTwoActorsWithXXObs(dyad *the_dyad, agentState *the_agent,
				behaviour * agent_behaviour, behaviour * client_behaviour_observation, xostate * xx_observation)
{
  return updateFullStateTwoActorsWithEmotionWithXXObs(the_dyad, the_agent,
						      agent_behaviour, client_behaviour_observation, xx_observation, NULL, NULL);
}
// does the full state update with an action and emotion for each agent and client
double updateFullStateTwoActorsWithEmotion(dyad *the_dyad, agentState *the_agent,
					   behaviour * agent_behaviour, behaviour * client_behaviour_observation,
					   emotion * agent_emotion, emotion * client_emotion)
{
  return updateFullStateTwoActorsWithEmotionWithXXObs(the_dyad, the_agent,
						      agent_behaviour, client_behaviour_observation, NULL,
						      agent_emotion, client_emotion);
}
double updateFullStateTwoActorsWithEmotionWithXXObs(dyad *the_dyad, agentState *the_agent,
						    behaviour * agent_behaviour, behaviour * client_behaviour_observation,
						    xostate *xx_observation,
						    emotion * agent_emotion, emotion * client_emotion)
{
  double totalweight=0.0;

  samplerState *sstate = the_dyad->dsstate;
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;
  sample *new_samples = (sample *) malloc(numsamples*sizeof(sample));

  for (int i=0; i<numsamples; i++) 
    new_samples[i].x[ID_CLIENT]=-1.0;

  double reset_threshold=RESET_THRESHOLD;
  //reset_type reset_tests[] = {RESET_NONE,RESET_CLIENT,RESET_AGENT,RESET_BOTH};
  //char * reset_labels[] = {"NONE","CLIENT","AGENT","BOTH"};
  reset_type who_to_reset=RESET_NONE;
  int num_reset_tests=1;
  if (sstate->do_resets) 
    num_reset_tests = 3;

#ifdef ALWAYS_RESET_BOTH
  num_reset_tests = 2;
#endif

#ifdef NEVER_RESET
  num_reset_tests=1;
#endif
  
  int reset_test, num_resets=0, num_resets_successful=0, num_resets_agent=0, num_resets_client=0, num_resets_both=0, reset_success=0;
  double allweight=0.0;
  double allweight_agent=0.0;
  double allweight_client=0.0;
  d_entity_type who_acted;

  // compute interact mean
  double *f_interact = (double *) malloc(VD*sizeof(double));
  double *tau_interact = (double *) malloc(VD*sizeof(double));
  
  xxstate xx_avg;

  
  // get the expectation of the current state from interact
  // INTERACT STUFF

#ifdef DO_INTERACT
  //  sentimentParams * __agent_params = sstate->sentiParams+(the_dyad->dyad_index);
  //  if (__agent_params->do_interact) {
    getExpectedSample(sstate, f_interact, tau_interact, &xx_avg);

    // shoehorn in the actual behaviour of agent 
    if (agent_behaviour) {
      for (int i=0; i<SSD; i++)
	f_interact[i+SSD]=gsl_vector_get(the_agent->entity_dicts[BEHAVIOUR]->theDictionary[agent_behaviour->index].epa,i);
    } else if (client_behaviour_observation) {
      for (int i=0; i<SSD; i++)
	f_interact[i+SSD]=gsl_vector_get(the_agent->entity_dicts[BEHAVIOUR]->theDictionary[client_behaviour_observation->index].epa,i);
    }
    // set tau to be exactly f
    memcpy(tau_interact,f_interact,VD*sizeof(double));
    // compute the mean by interact
    // this is stored in sstate->sentimentParams->interact_mean and will be moved to dyad->sstate->sentimentParms->interact_mean
    if (agent_behaviour) 
      computeInteractMean(sstate, BEH_AGENT, f_interact, tau_interact);
    else if (client_behaviour_observation) 
      computeInteractMean(sstate, BEH_CLIENT, f_interact, tau_interact);
    
#ifdef VERYVERBOSE
    printInteractMean(sstate);
#endif
    
    free(f_interact);
    free(tau_interact);
    //  }
  // INERACT STUFF
#endif
  
#ifdef OUTPUT_ALL_WEIGHTS
  double *all_weights_save;
  all_weights_save=(double *) malloc(numsamples*sizeof(double));
#endif
  
  for (int i=0; i<numsamples; i++) {

    // run through reset tests - the first is the standard update
    // the next three are potential re-identifications
    // don't want to reset if the weight -> 0 because of the turn (-1 is returned from sampleWeighNextSentiment in this case)
    reset_test=0;
    who_to_reset=RESET_NONE;
    totalweight=-1.0;
    reset_success=0;
    who_acted = ID_AGENT;  // we use ID_AGENT just as a flag here to mean it has not been set yet
    while (reset_test < num_reset_tests  && totalweight <= reset_threshold) {

      // sample the next state and compute the weight for this sample
      totalweight = sampleWeightNextSentimentWithEmotion(the_dyad, who_to_reset, the_agent, &who_acted,
							 the_samples+i, new_samples+i,
							 agent_behaviour, client_behaviour_observation,
							 xx_observation,
							 agent_emotion, client_emotion);
      if (who_to_reset != RESET_NONE && totalweight > reset_threshold) {
	reset_success=1;
	// used to check here as well and copy f->x for agent and client id if reset -- now moved to SampleNextState
      }
      // if the update failed, we will need to re-try with a reset
      if (totalweight <= reset_threshold) {
	// figure out who to reset next
	if (reset_test == 0) 
	  who_to_reset=RESET_CLIENT;

	if (reset_test == 1) 
	    who_to_reset = RESET_BOTH;

#ifdef ALWAYS_RESET_BOTH
	// overwrite
	who_to_reset = RESET_BOTH;
#endif

#ifdef NEVER_RESET
	who_to_reset=RESET_NONE;
#endif
      }
      reset_test++;
    }  //  while

    // reset_flag is used for computing emotions only
    // num_resets_XX is printed below
    // no resets: reset_test=1
    // one reset : reset_test=2
    // two resets : reset test=3
    // reset_flag=0 : zero resets
    // reset_flag=2 : one reset,  success
    // reset_flag=1 : two resets
    new_samples[i].reset_flag = 0;
    if (reset_test>1) {
      num_resets++;
      new_samples[i].reset_flag = 1;
      if (reset_success) {
	new_samples[i].reset_flag = 2;
	num_resets_successful++;
	if (who_to_reset == RESET_AGENT)
	  num_resets_agent++;
	if (who_to_reset == RESET_CLIENT)
	  num_resets_client++;
	if (who_to_reset == RESET_BOTH)
	  num_resets_both++;
      }
    }
    if (totalweight > 0 && totalweight > reset_threshold) {
      allweight += totalweight;
      if (who_acted==BEH_AGENT) 
	allweight_agent += totalweight;
      else if (who_acted == BEH_CLIENT)
	allweight_client += totalweight;
    }
#ifdef OUTPUT_ALL_WEIGHTS
    all_weights_save[i]=totalweight;
#endif
  } // loop over samples
#ifdef VERBOSE
  printf("number of reset attempts: %d successes: %d\n",num_resets, num_resets_successful);
  printf("number of resets for agent: %d client: %d both: %d \n",num_resets_agent, num_resets_client, num_resets_both);
#endif
  
  sstate->num_resets=num_resets;
  sstate->num_resets_successful=num_resets_successful;
  sstate->num_resets_agent=num_resets_agent;
  sstate->num_resets_client=num_resets_client;
  sstate->num_resets_both=num_resets_both;

#ifdef VERBOSE
  printf("total weight of all samples in update %g\n",allweight);
  printf("total weight of all samples for agent in update %g\n",allweight_agent);
  printf("total weight of all samples for client in update %g\n",allweight_client);
#endif

#ifdef OUTPUT_ALL_WEIGHTS
  FILE *fp=fopen(ALL_WEIGHTS_SAVE_FILE,"a");
  for (int i=0; i<numsamples; i++) 
    fprintf(fp,"%4.3g,",all_weights_save[i]);
  fprintf(fp,"\n");
  free(all_weights_save);
  fclose(fp);
#endif
  
  double roughNoise=sstate->roughenNoise;
  if (allweight <= reset_threshold) {
    // we will only copy over who acted
    d_entity_type *save_who_acted=(d_entity_type *) malloc(sstate->numsamples*sizeof(d_entity_type));
#ifdef VERBOSE
    printf("----reset failed: weight %g threshold %g\n",allweight,reset_threshold);
#endif
    sstate->num_hard_resets = sstate->num_hard_resets+1;
    // TODO this will fail if allweight==0.0
    // actually it does not but we should take care of this case - otherwise it just resamples the index=0 sample only
    // the way it is now
    // temporary hack - copy over all samples - possibly do this and then add more roughening noise
    // WHY was I copying new_samples -> the_samples here? 
    for (int i=0; i<sstate->numsamples; i++) 
      save_who_acted[i]=new_samples[i].who_acted;
    memcpy(new_samples,the_samples,numsamples*sizeof(sample));
    // have to also set all weights to 1.0
    for (int i=0; i<sstate->numsamples; i++) {
      new_samples[i].who_acted=save_who_acted[i];
      new_samples[i].weight=1.0;
    }
    // and increase roughening noise - really shake things up
    roughNoise *= 2.0;
    free(save_who_acted);
  } 
  reSampleByActor(new_samples,the_samples,numsamples);

#ifdef VERYVERBOSE
  printf("number of hard resets: %d\n",sstate->num_hard_resets);
#endif
  // resample from the persona
  if (the_agent->num_persona_ids > 0) {
    printf("resampling from persona\n");
    resampleFromPersona(sstate, the_dyad->dyad_index, the_agent, PERSONA_FRACTION);
  }

  // add roughening
  roughenAllSamples(sstate, roughNoise);
  free(new_samples);
  // why? 
  return totalweight;
}


// function to sample a next sentiment and compute the weight,
// and has input:
//   agent_behaviour_action: action taken by agent
//   client_behaviour_observation: action of client observed by agent
// output:
// on agent action (next_sample->who_acted=BEH_AGENT)
//   agent_behaviour: will be exactly agent_behaviour_action
//   client_behaviour: draw from P(x_bc' | x)
//   next_sample->xx : draw from P(x' | x_c', x_a', x_x
// on client action (next_sample->who_acted=BEH_CLIENT)
//   agent behaviour: sampled from P(x_ba' | x)
//   client behaviour: will be exactly the observation of client behaviour client_behaviour_observation
//   next_sample->xx: draw form P(x' | x_bc',x_ba', x_x)
// updateXSample(next_sample,agent_behaviour_action, &agent_behaviour, &client_behaviour,client_behaviour_observation)
void updateXSample(gsl_rng *r, dyad *the_dyad, sample * next_sample, sample * current_sample, agentState *the_agent,
		   behaviour *agent_behaviour, behaviour * client_behaviour_observation)
{
  // identity mapping for denotative identities
  // note that this will be "overridden" if a reset is done
  // the behaviour component is overridden right afterwards
  //if (!DO_DENOTATIVE_IDENTITY_UPDATES)
   for (int i=0; i < NGTD; i++) 
     next_sample->x[i] = current_sample->x[i];

    
  // set the x_b to be exactly the behaviour observed.
  // This could be a distribution of behaviour labels (indices)
  // and would be a random draw from this distribution,
  //which would then need to be weighted by the actual actions taken
  if (agent_behaviour != NULL) 
    updateSampleXba(r, &(next_sample->x[BEH_AGENT]), current_sample->xx, current_sample->x, agent_behaviour->index, the_dyad, the_agent);  
  else // REMOVED WHEN IMPLEMENTING INSTITUTIONAL FILTERS (may have been just to avoid the calculation lol) :  if (next_sample->who_acted==BEH_CLIENT) 
    updateSampleXba(r, &(next_sample->x[BEH_AGENT]), current_sample->xx, current_sample->x, -1, the_dyad, the_agent); 
  
  // draw a sample from the predicted next client behaviour distribution
  // this should not depend on client_behaviour_observation but we make it so to save samples
  if (client_behaviour_observation != NULL) 
    updateSampleXbc(r, &(next_sample->x[BEH_CLIENT]), current_sample->xx, current_sample->x, client_behaviour_observation->index, the_dyad, the_agent);
  else //REMOVED WHEN IMPLEMENTING INSTITUTIONAL FILTERS if (next_sample->who_acted==BEH_AGENT)
    updateSampleXbc(r, &(next_sample->x[BEH_CLIENT]), current_sample->xx, current_sample->x, -1, the_dyad, the_agent);

  if (DO_DENOTATIVE_IDENTITY_UPDATES) {
    // this function does not exist yet...is a stub in defaultAPP.c
    // see the implementation in bayesactAPP.c
    updateSampleXac(r, &(next_sample->x[ID_AGENT]), current_sample->x[ID_AGENT],
		    &(next_sample->x[ID_CLIENT]), current_sample->x[ID_CLIENT],
		    next_sample->x[BEH_AGENT], next_sample->x[BEH_CLIENT],the_dyad,the_agent);
  }
  
  // this is user defined and updates the denotive additional state xx based on previous state and the two agent behaviours
  updateXXSample(r, &(next_sample->xx), &(current_sample->xx), next_sample->x[BEH_AGENT], next_sample->x[BEH_CLIENT]);
}
// weight the sample according to the observation of xx
double weightXSample(gsl_rng *r, sample * next_sample, xostate * xx_observation)
{
  // possibly weight the identities here - pass in observations of these as well
  if (xx_observation == NULL)
    return 1.0;
 
  return weightXXSample(r, &(next_sample->xx), xx_observation);
}
// weight the sample according to the observation of x_bc (the client behaviour)
double weightXbcSample(gsl_rng *r, sample * next_sample, behaviour *client_behaviour_observation)
{
  // possibly weight the identities here - pass in observations of these as well
  if (client_behaviour_observation == NULL)
    return 1.0;
 
  return weightXXbcSample(r, &(next_sample->x[BEH_CLIENT]), client_behaviour_observation->index);
}

// the_behaviour is a denotativeState giving the observed behaviour
// really, this should be a triple of observation values for x_a, x_b and x_c.
// as it is now, its only one behaviour observation (of x_b) with a deterministic observation function x_b==\Omega_{x_b}
double sampleWeightNextSentimentWithEmotion(dyad *the_dyad, reset_type rtype,
					    agentState *the_agent, d_entity_type *who_acted,
					    sample * current_sample, sample * next_sample,
					    behaviour * agent_behaviour, behaviour * client_behaviour_observation,
					    xostate *xx_observation,
					    emotion * the_agent_emotion, emotion * the_client_emotion)
{
  samplerState *sstate = the_dyad->dsstate;
  //double beh_weight,aid_weight,cid_weight;
  double weight;
  // select how to update this sample - fails if no one acted at all
  if (agent_behaviour==NULL && client_behaviour_observation==NULL) {
    next_sample->weight=0.0;
    return -1;
  }
  // otherwise, select the one who acted, or randomly if both acted -  not used anymore
  // if who_acted is specified (on resets to be one of BEH_AGENT or BEH_CLIENT) then this takes precedence
  // its ID_AGENT if not set yet - its used as a flag only
  if (*who_acted == ID_AGENT) {  // we use ID_AGENT just as a flag here to mean it has not been set yet
    if (agent_behaviour==NULL) {
      *who_acted = BEH_CLIENT;
    } else if (client_behaviour_observation==NULL) {
      *who_acted = BEH_AGENT;
    } else {
      // else both are not null, so the choice is random (made above)
      // this will only happen now if neither agent acts, then the sampleNextState call is like
      // it is in getNextAction, in which who_acted is random, so the next state sample is some kind
      // of weird average in which you don't know who acted on the last round. This seems like a bad idea
      // April 20, 2021, so I changed it back to just be BEH_AGENT - so the agent imagines it acts in any
      // round in which neither agent acts
      //*who_acted = (gsl_rng_uniform(sstate->r) > 0.5 ? BEH_AGENT : BEH_CLIENT);
      *who_acted=BEH_AGENT;
    }
  }

  // sample the next sentiment and x state, weight by somatic transform for identity and behaviour only
  sampleNextState(the_dyad, *who_acted, the_agent, rtype,
		  current_sample, next_sample, agent_behaviour, the_agent_emotion, client_behaviour_observation, the_client_emotion);

  // weight the new sample for x observation and emotions
  weight=computeSampleWeight(sstate, the_dyad->dyad_index, the_agent, next_sample,
			     xx_observation, client_behaviour_observation, the_client_emotion, the_agent_emotion);
  
  next_sample->weight *= weight;
  
  return next_sample->weight;
}
// get the weight of the sample based on the observations of xx, x_bc and \epsilon_c
// these are xx_observation, client_behaviour_observation, and the_client_emotion, respectively
double computeSampleWeight(samplerState *sstate, int dyad_index, agentState * the_agent, 
			   sample * next_sample, xostate * xx_observation, 
			   behaviour * client_behaviour_observation,
			   emotion * the_client_emotion, emotion * the_agent_emotion)
{
  double weight, xbc_weight,ae_weight,ce_weight;
  double tmp_emotion[SSD];
  // add in a sample weight for the observations of xstate - default this returns 1.0
  weight = weightXSample(sstate->r, next_sample, xx_observation);
  //printf("X weight %g \n",weight);

  // weight for the client behaviour observation - compares next_sample->x[BEH_CLIENT] with client_behaviour_observation->index
  xbc_weight = weightXbcSample(sstate->r, next_sample, client_behaviour_observation);
  //printf("xbc weight %g \n",xbc_weight);
  weight *= xbc_weight;



  // clent emotion weight 
  ce_weight=1.0;
  if (WEIGHT_CLIENT_EMOTION && the_client_emotion != NULL) {
    computeEmotion(sstate, next_sample->f+CLIENT*SSD, next_sample->tau+CLIENT*SSD, tmp_emotion);
    ce_weight = somaticTransform(WEIGHT, sstate->r, dyad_index, &(the_client_emotion->index), tmp_emotion, the_agent->entity_dicts[EMOTION]);
    // could also add weight here for an emotion inertia - i.e. that we don't believe the client's emotion will change quickly.
    // not sure if this is reasonable or needed though.
  }  else {
    ce_weight=1.0;
  }
  weight = weight*ce_weight;

  // agent emotion weight - so this means the agent itself will weight its own update by the emotion it displayed.
  // this is something like James' "sad because we cry" idea - that is, the emotion that is expressed determines the state to some degree
  ae_weight=1.0;
  if (WEIGHT_AGENT_EMOTION && the_agent_emotion != NULL) {
    computeEmotion(sstate, next_sample->f+AGENT*SSD, next_sample->tau+AGENT*SSD, tmp_emotion);
    ae_weight = somaticTransform(WEIGHT, sstate->r, dyad_index, &(the_agent_emotion->index), tmp_emotion, the_agent->entity_dicts[EMOTION]);
  } else {
    ae_weight=1.0;
  }
  weight = weight*ae_weight;

  return weight;
}

void sampleObservation(dyad *the_dyad,  agentState *the_agent, sample * the_sample,
		       behaviour * the_client_behaviour, emotion * the_client_emotion, xostate *xo)
{
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  double tmp_emotion[SSD];
  sampleX(sstate->r, &(the_sample->xx), xo);
  sampleXbc(sstate->r, the_sample->x[BEH_CLIENT], &(the_client_behaviour->index));
  //printf("sampling from %d and got %d\n",the_sample->x[BEH_CLIENT],the_client_behaviour->index);
  if (sstate->use_emotions) {
    computeEmotion(sstate, the_sample->f+CLIENT*SSD, the_sample->tau+CLIENT*SSD, tmp_emotion);
    // really should go one step further here and sample the_client_emotion from the denotative state returned by the somaticTransform
    somaticTransform(DENOTATIVE, sstate->r, dyad_index, &(the_client_emotion->index), tmp_emotion, the_agent->entity_dicts[EMOTION]);
  } else {
    emotion * emot_list=NULL;
    int index;
    double prob;
    getNullEmotion(the_dyad, the_agent, &index, &prob, CLIENT);
    the_client_emotion->index=index;
    the_client_emotion->prob=prob;
    free(emot_list);
  }
  
  return;
}

//*********************************************************************************
// main sample update loop
// does the bulk of the work in updating a sample and somatic transform weighting
// 1. possibly reset the identities from the interact means
// 2. get sentiment for the actual action taken if reset and copy into f
// 3. update the sentiment: draw from P(f'|...)
// 4. update the X sample - draw X_ba, X_bc, X_x (if defined) -- possibly also Xa, Xc if DO_DENOTATIVE_IDENTITY_UPDATES is defined, and 
// 5. WEIGHT somatic transform for identities -- this is G(f_a,X_a) and G(f_c,X_c) if not reset
// 6. DENOTATIVE somatic transform to recover identity labels from sentiment if resetting
// 7. WEIGHT somatic transform for behaviour (possibly if behaviour denotative has been filled in -- does not happen on getNextAction for example)
//*********************************************************************************
// sample the complete next state, and assign weights based on somatic transforms
// who_acted is either BEH_AGENT or BEH_CLIENT and indicates who is acting for this sample
void sampleNextState(dyad *the_dyad, d_entity_type who_acted, agentState * the_agent, reset_type rtype,
		     sample * current_sample, sample * next_sample, behaviour * agent_behaviour, emotion * the_agent_emotion,
		     behaviour * client_behaviour_observation, emotion * the_client_emotion)
{
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  double weight;
  sample tmp_sample;
  weight = 1.0;
  // clpy out current_sample_input so its not abused below on reset
  sample tmp_current_sample;
  memcpy(&tmp_current_sample,current_sample,sizeof(sample));
    
    

  // was done previously in sampleNextSEntiment -- copy in the interact mean if resetting
  sentimentParams * __agent_params = sstate->sentiParams;
  if (!(rtype==RESET_NONE) &&  !(__agent_params->interact_mean==NULL)) {
    if (rtype==RESET_CLIENT || rtype==RESET_BOTH) {
      for (int i=SSD*2; i<SSD*3; i++) 
	tmp_current_sample.f[i]=(gsl_rng_uniform_pos(sstate->r)*8.6)-4.3;
#ifdef DO_INTERACT
      //if (__agent_params->do_interact) 
      for (int i=SSD*2; i<SSD*3; i++) {
	tmp_current_sample.f[i]=gsl_vector_get(__agent_params->interact_mean,i);
	tmp_current_sample.tau[i]=tmp_current_sample.f[i];
      }
#endif
    }
    //      else // totally random
    
    if (rtype==RESET_AGENT || rtype==RESET_BOTH) {
      // totally random
      for (int i=0; i<SSD; i++) 
	tmp_current_sample.f[i]=(gsl_rng_uniform_pos(sstate->r)*8.6)-4.3;
#ifdef DO_INTERACT
      //if (__agent_params->do_interact) 
      for (int i=0; i<SSD; i++) {
	tmp_current_sample.f[i]=gsl_vector_get(__agent_params->interact_mean,i);
	tmp_current_sample.tau[i]=tmp_current_sample.f[i];
      }
#endif
    }
  }

  // we need to "shoehorn" in the behaviour that was taken if there was a reset
  // if the agent behaviour is not null, we sample the next sentiment based on it
  if (agent_behaviour != NULL) {
    // if there was an agent reset, then we use the behaviour that was actually performed in the sentiment distribution as well before the update
    if (rtype == RESET_AGENT || rtype == RESET_BOTH) 
      somaticTransform(CONNOTATIVE,sstate->r,dyad_index,&(agent_behaviour->index),tmp_current_sample.f+SSD*BEHAVIOUR,the_agent->entity_dicts[BEHAVIOUR]);

    sampleNextSentiment(sstate, BEH_AGENT, rtype, &tmp_current_sample, &tmp_sample);
  } else {
    memcpy(&tmp_sample, &tmp_current_sample, sizeof(sample));
  }
  // if the client behaviour is not null
  if (client_behaviour_observation != NULL) {
    // if there was a client reset, then we use the behaviour that was actually performed in the sentiment distribution as well before the update
    // so this means if the agent has not acted, but the client has, and we not resetting or resetting the agent, then we don't use
    // the sentiment for the behaviour actually performed here.
    // may want this instead: if (rtype == RESET_BOTH) --- so that behaviour is not shoehorned in when resetting the client - it is free to be re-identified.
    // RESET_FLAG_MARKER
    if (rtype == RESET_CLIENT || rtype == RESET_BOTH) 
      somaticTransform(CONNOTATIVE,sstate->r,dyad_index,&(client_behaviour_observation->index),tmp_sample.f+SSD*BEHAVIOUR,the_agent->entity_dicts[BEHAVIOUR]);
    sampleNextSentiment(sstate, BEH_CLIENT, rtype, &tmp_sample, next_sample);
  } else {
    memcpy(next_sample, &tmp_sample, sizeof(sample));
  }
  // sample the next sentiment - this does the work of computing the update for f and tau for who_acted acting (BEH_AGENT or BEH_CLIENT)
  // if no one acted, we still need to update 
  if (agent_behaviour == NULL && client_behaviour_observation == NULL) 
    sampleNextSentiment(sstate, who_acted, rtype, &tmp_current_sample, next_sample);


  // who_acted is the actor that acted for this step, so this is assigned to next_sample
  next_sample->who_acted=who_acted;
  
  // sample the next X state (including application-dependent XX state updates)
  // this copies the identity over directly unless DO_DENOTATIVE_IDENTITY_UPDATES is 1 in fvarsample.h,
  // in which case updateXacSample from xxAPP.c is called
  // this version "cheats" by sampling x_bc as client_behaviour_observation exactly
  updateXSample(sstate->r, the_dyad, next_sample, &tmp_current_sample, the_agent, agent_behaviour, client_behaviour_observation);
  // for the planner I needed a version that you can call without having client_behaviour_observation yet
  // to not do this, do the following instead: this will yield distributions over x states that are denotatively consistent with those
  // as specified in the APP file
  //updateXSample(sstate->r, next_sample,current_sample,the_agent,agent_behaviour,NULL);
 
  // now the weight for this sample is computed using the somatic transform for identity of client, agent, and  behaviour and 
  // compute weights for each element in the grammar using a somatic transform
  // should also be able to do this now when resetting as the weights should be close to one as the sentiment was sampled from the reset X id
  // however I am leaving the if statements in here for now

  /* MARCH 2021 */
  weight=1.0;
  // if we are doing denotative identity updates then this is always done
  if (DO_DENOTATIVE_IDENTITY_UPDATES || rtype==RESET_NONE || rtype==RESET_CLIENT) 
    weight *= somaticTransform(WEIGHT, sstate->r, dyad_index, (next_sample->x+ID_AGENT), next_sample->f+AGENT*SSD, the_agent->entity_dicts[AGENT]);
  if (DO_DENOTATIVE_IDENTITY_UPDATES || rtype==RESET_NONE || rtype==RESET_AGENT) 
    weight *= somaticTransform(WEIGHT, sstate->r, dyad_index, (next_sample->x+ID_CLIENT), next_sample->f+CLIENT*SSD, the_agent->entity_dicts[CLIENT]);
  // update the denotative state based on the new connotative state, but this for reset tests only
  // this is where new identities get added to the denotative state on reset
  // these moved from updateFullState March 2021
  // HERE WE NEED TO APPLY institutional filters by consulting xxAPP.c?
  // no not here because its only EPAs here...

  // if we are doing denotative identity updates then this is not necessary

  if (!DO_DENOTATIVE_IDENTITY_UPDATES) {
    if (rtype == RESET_CLIENT || rtype== RESET_BOTH) 
      somaticTransform(DENOTATIVE, sstate->r, dyad_index, (next_sample->x+ID_CLIENT), next_sample->f+CLIENT*SSD, the_agent->entity_dicts[CLIENT]);
    if (rtype == RESET_AGENT || rtype== RESET_BOTH) 
      somaticTransform(DENOTATIVE, sstate->r, dyad_index, (next_sample->x+ID_AGENT), next_sample->f+AGENT*SSD, the_agent->entity_dicts[AGENT]);
  }

  // behaviour somatic transform -- ONLY do this is x[who_acted] has been filled in -
  // this will not occur e.g. in getNextAction since x[who_acted] needs to be computed from f using a DENOTATIVE SOMATIC TRANSFORM
  //printf("who_acted %d next_sample->x[who_acted]: %d\n",who_acted,next_sample->x[who_acted]);
  // next_sample->x[who_acted] is the behaviour actually taken
  //
  double tmpweight=0.0;
  tmpweight = somaticTransform(WEIGHT, sstate->r, dyad_index, (next_sample->x+who_acted), next_sample->f+BEHAVIOUR*SSD, the_agent->entity_dicts[BEHAVIOUR]);
  //printf("rtype %d weight %g tmpweight: %g f=[%4.3g %4.3g %4.3g]\n",rtype,weight,tmpweight,next_sample->f[3],next_sample->f[4],next_sample->f[5]);
  weight *= tmpweight;
  next_sample->weight = weight;
}

// sets up the curlyC, curlyK and curlyH matrices - these are refactorings of the impression
// formation equations used in drawing a sentiment sample - basically everything not involving the behaviour
// term f_b is calculated
void setUpImpressions(samplerState *sstate, d_entity_type atype, double *tau)
{
  int *themap;
  sentimentParams * __agent_params = sstate->sentiParams;

  // at this spot, the dynamics are chosen to be either for agent or for client
  // so atype here acts somewhat like a "turn" variable, but only at this level
  if (atype == BEH_AGENT)
    themap = __agentmap;
  else // if (atype == BEH_CLIENT) 
    themap = __clientmap;

  // set up impression formation equations
  setUpImpressionFormation_bonly(__agent_params,themap,tau);
}

// sets up the curlyC, curlyK and curlyH matrices - these are refactorings of the impression
// formation equations used in drawing a sentiment sample - basically everything not involving the behaviour
// term f_b is calculated
// curlyK, curlyH and curlyC must be pre-alllocated using setUpParameters(...)
// this only computes curlyH,K,C for behaviour, not object and actor
int setUpImpressionFormation_bonly(sentimentParams *__agent_params, int * themap, double *tau)
{
  return (setUpImpressionFormation(__agent_params,themap,tau,0,0));
  
}
// two arguments doA and doC will set up the curlyH,K,C matrices so that optimal actor and object are also computable easily
// in computeInteractMean for example (or perhaps for use in resets)
int setUpImpressionFormation(sentimentParams *__agent_params, int * themap, double *tau, int doA, int doC)
{
  double ffval;
  int is_b,is_a,is_c;
  int i,j,k;
  // zero curlyC and set curlyK to identity to start
  gsl_matrix_set_identity(__agent_params->curlyK);
  gsl_matrix_set_identity(__agent_params->curlyKa);
  gsl_matrix_set_identity(__agent_params->curlyKc);
  gsl_matrix_set_zero(__agent_params->curlyH);
  gsl_matrix_set_zero(__agent_params->curlyHa);
  gsl_matrix_set_zero(__agent_params->curlyHc);
  gsl_vector_set_zero(__agent_params->curlyC);
  gsl_vector_set_zero(__agent_params->curlyCa);
  gsl_vector_set_zero(__agent_params->curlyCc);
  
  // the following computes the transient impression from the input fundamental by computing the features
  // and then multiplying by the matrix M
  for (k=0; k<VD; k++) {
    for (j=0; j<__agent_params->num_features; j++) {
      // would it be more efficient here to have  M a 2D array?
      ffval = __agent_params->M[j*VD+themap[k]];   // change k part of arg of M  here on client turn from 0,1,2<->6,7,8
      //printf("                 ffval: %4.3g\n",ffval);

      is_b = 0;
      for (i=0; i<__agent_params->m_num_features[j]; i++) {
	if (__agent_params->m_index[j][i] >= SSD && __agent_params->m_index[j][i] < SSD*2)  {
	   // a behaviour grammar term
	  // which behaviour grammar term is it?  (numbered 1,2,3 for e,p,a)
	  //assumes only one of these is present in each m_num_features[j] set so impression tau' is linear in fb' (which it is)
	  is_b = __agent_params->m_index[j][i]-(SSD-1);
	} else {
	   // behaviour grammar term is 0 (no behaviour term)
	  ffval *= tau[themap[__agent_params->m_index[j][i]]];        // change arg of tau here on client turn from 0,1,2<->6,7,8     
	}
      }
      //add to running sums to complete curlyC, curlyK
      if (is_b == 0) {
	gsl_vector_set(__agent_params->curlyC, k, gsl_vector_get(__agent_params->curlyC,k) + ffval);   
      } else {
	gsl_matrix_set(__agent_params->curlyK, k, SSD+is_b-1, gsl_matrix_get(__agent_params->curlyK, k, SSD+is_b-1) - ffval);
	gsl_matrix_set(__agent_params->curlyH, k, is_b-1, gsl_matrix_get(__agent_params->curlyH, k, is_b-1) + ffval);
      } 
    }
  }
  if (my_invert_matrix(__agent_params->curlyK,__agent_params->curlyKinv,VD))
    return 1;
  if (doA) {
    // Ha matrix 
    for (k=0; k<VD; k++) {
      for (j=0; j<__agent_params->num_features; j++) {
	// would it be more efficient here to have  M a 2D array?
	ffval = __agent_params->M[j*VD+themap[k]];   // change k part of arg of M  here on client turn from 0,1,2<->6,7,8
	is_a = 0;
	for (i=0; i<__agent_params->m_num_features[j]; i++) {
	  if (__agent_params->m_index[j][i] >= 0 && __agent_params->m_index[j][i] < SSD)  {
	  // an actor grammar term
	  // which behaviour grammar term is it?  (numbered 1,2,3 for e,p,a)
	  //assumes only one of these is present in each m_num_features[j] set so impression tau' is linear in fb' (which it is)
	    is_a = __agent_params->m_index[j][i]+1;
	  } else {
	    // behaviour grammar term is 0 (no behaviour term)
	    ffval *= tau[themap[__agent_params->m_index[j][i]]];        // change arg of tau here on client turn from 0,1,2<->6,7,8     
	  }
	}
	//printf("is_b : %d ffval: %4.3g\n",is_b,ffval);
      //add to running sums to complete curlyC, curlyK
	if (is_a == 0) {
	  gsl_vector_set(__agent_params->curlyCa, k, gsl_vector_get(__agent_params->curlyCa,k) + ffval);   
	} else {
	  gsl_matrix_set(__agent_params->curlyKa, k, is_a-1, gsl_matrix_get(__agent_params->curlyKa, k, is_a-1) - ffval);
	  //printf("adding to curlyH at position %d %d : %4.3g\n",k,is_b-1,ffval);
	  gsl_matrix_set(__agent_params->curlyHa, k, is_a-1, gsl_matrix_get(__agent_params->curlyHa, k, is_a-1) + ffval);
	} 
      }
    }
    if (my_invert_matrix(__agent_params->curlyKa,__agent_params->curlyKainv,VD))
      return 1;
  }
  if (doC) {
    // Hc matrix
    for (k=0; k<VD; k++) {
      for (j=0; j<__agent_params->num_features; j++) {
	// would it be more efficient here to have  M a 2D array?
	ffval = __agent_params->M[j*VD+themap[k]];   // change k part of arg of M  here on client turn from 0,1,2<->6,7,8
	is_c = 0;
	for (i=0; i<__agent_params->m_num_features[j]; i++) {
	  if (__agent_params->m_index[j][i] >= SSD*2 && __agent_params->m_index[j][i] < SSD*3)  {
	    // a client grammar term
	    // which behaviour grammar term is it?  (numbered 1,2,3 for e,p,a)
	    //assumes only one of these is present in each m_num_features[j] set so impression tau' is linear in fb' (which it is)
	    is_c = __agent_params->m_index[j][i]-(SSD*2-1);
	  } else {
	    // behaviour grammar term is 0 (no behaviour term)
	    ffval *= tau[themap[__agent_params->m_index[j][i]]];        // change arg of tau here on client turn from 0,1,2<->6,7,8     
	  }
	}
	//printf("is_b : %d ffval: %4.3g\n",is_b,ffval);
	//add to running sums to complete curlyC, curlyK
	if (is_c == 0) {
	  gsl_vector_set(__agent_params->curlyCc, k, gsl_vector_get(__agent_params->curlyCc,k) + ffval);   
	} else {
	  gsl_matrix_set(__agent_params->curlyKc, k, SSD*2+(is_c-1), gsl_matrix_get(__agent_params->curlyKc, k, SSD*2+(is_c-1)) - ffval);
	  //printf("adding to curlyH at position %d %d : %4.3g\n",k,is_b-1,ffval);
	  gsl_matrix_set(__agent_params->curlyHc, k, is_c-1, gsl_matrix_get(__agent_params->curlyHc, k, is_c-1) + ffval);
	} 
      }
    }
    if (my_invert_matrix(__agent_params->curlyKc,__agent_params->curlyKcinv,VD))
      return 1;
  }
  return 0;
}
void printInteractMean(samplerState *sstate)
{
  sentimentParams * __agent_params = sstate->sentiParams;
  //  if (!__agent_params->do_interact)
  //return;
  printf("interact optimal: ");
  for (int i=0; i<VD; i++) 
    printf("%4.3g ",gsl_vector_get(__agent_params->interact_mean,i));
  printf("\n");
}
void printInteractMeanCSV(samplerState *sstate, FILE *fpout, char *csv_header_line)
{
  char tmp_header_line[1024];

  sentimentParams * __agent_params = sstate->sentiParams;
  for (int i=0; i<VD; i++) 
    fprintf(fpout,"%4.3g, ",gsl_vector_get(__agent_params->interact_mean,i));


  for (int i=0; i<VD; i++) {
    sprintf(tmp_header_line, "interact mean %d,",i);
    strcat(csv_header_line,tmp_header_line);
  }
}
// does a full interact update
// INPUT:  from f (EPA for identities) + behaviour taken (SSD) + tau (current tau for identities)
// OUTPUT: resulting transient tau,
// optimal mean as a VD vector accessible after this call fro _sstate->sentiParams[dyad_index]->interact_mean
// main function takes behaviour * as input
double doFullInteractUpdate(samplerState *sstate, int iteration, agentState *the_agent, d_entity_type atype,
			    int agent_index, int client_index, behaviour *b, double *tau, double *f_out, double *tau_out, double *agent_act, double *client_act)
{
  double behaviour[SSD];
  if (b) {
    // get the EPA for this behaviour
    for (int i=0; i<SSD; i++) 
      behaviour[i]=gsl_vector_get(the_agent->entity_dicts[BEHAVIOUR]->theDictionary[b->index].epa,i);
    for (int i=0; i<SSD; i++)
      f_out[i]=gsl_vector_get(the_agent->entity_dicts[AGENT]->theDictionary[agent_index].epa,i);
    for (int i=0; i<SSD; i++)
      f_out[i+2*SSD]=gsl_vector_get(the_agent->entity_dicts[CLIENT]->theDictionary[client_index].epa,i);
  } else {
    printf("Error: behaviour must be specified in doFullInteractUpdate\n");
    return -1;
  }
  double defl=doFullInteractUpdate_helper(sstate, iteration,atype,f_out,behaviour,tau,tau_out,agent_act,client_act);
  return defl;
}
      
// helper function takes EPA Vector as behaviour
double doFullInteractUpdate_helper(samplerState *sstate, int iteration, d_entity_type atype,
				   double *f, double *behaviour, double *tau, double *tau_out, double *agent_act, double *client_act)
{
  double defl;
  double f_interact[VD];

  for (int i=0; i<VD; i++)
    f_interact[i]=f[i];
  for (int i=0; i<SSD; i++) 
    f_interact[i+SSD]=behaviour[i];

#ifdef VERYVERBOSE
  printf("fundamental (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_interact[i]);
  printf("\n");
#endif
  if (iteration==0) {  // for first iteration
    //printf("tau set to f on iteration 0\n");
    for (int i=0; i<VD; i++) 
      tau[i]=f_interact[i];
  } else {
    //copy only the behaviour part
    for (int i=0; i<SSD; i++) 
      tau[i+SSD]=f_interact[i+SSD];
  }
  computeInteractMean(sstate, atype, f_interact, tau);

#ifdef VERYVERBOSE
  printf("tau (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau[i]);
  printf("\n");
  printf("-----------------------------------------------------------------------------\n");
  printf("Interact optimal identites and behaviour (NOW):\n");
  printInteractMean(sstate);
#endif

  // function computes the transients in tau_out from the f_b and tau_a/tau_c
  defl  = formImpression(sstate, atype, f_interact, tau, tau_out);

    // this should be identical to that produced by interact for the event
#ifdef VERYVERBOSE
  printf("transient output (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_out[i]);
  printf("\n");
  printf("deflection (interact): %g\n",defl);
  printf("fundamental (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_interact[i]);
  printf("\n");

  printf("-----------------------------------------------------------------------------\n");
  printf("Interact optimal identities and behaviour (NEXT):\n");
#endif
  //computeInteractMean(sstate, dyad_index, BEH_AGENT, f_interact,tau_out_interact);
  // get optimal next behaviour (interact/agent)
  sentimentParams * __agent_params = sstate->sentiParams;
  computeInteractMean(sstate, BEH_AGENT, f_interact, tau_out);
  for (int i=0; i<SSD; i++) 
    agent_act[i]=gsl_vector_get(__agent_params->interact_mean,i+SSD);
#ifdef VERYVERBOSE
  printf("AGENT");
  printInteractMean(sstate);
#endif
  // get optimal next behaviour (interact/client)
  computeInteractMean(sstate, BEH_CLIENT, f_interact, tau_out);
  for (int i=0; i<SSD; i++) 
    client_act[i]=gsl_vector_get(__agent_params->interact_mean,i+SSD);
#ifdef VERYVERBOSE
  printf("CLIENT");
  printInteractMean(sstate);
#endif

  return defl;
}

// this does the standard interact impression formation. Takes the identities from tau and the behaviour from f to compute the new
// tau and f. Returns the total deflection
double formImpression(samplerState *sstate, d_entity_type atype, double * f, double * tau, double * tau_out)
{
  int *themap;
  sentimentParams * __agent_params = sstate->sentiParams;

  // at this spot, the dynamics are chosen to be either for agent or for client
  // so atype here acts somewhat like a "turn" variable, but only at this level
  if (atype == BEH_AGENT)
    themap = __agentmap;
  else // if (atype == BEH_CLIENT) 
    themap = __clientmap;

  // set up impression formation equations
  int status;
  status=setUpImpressionFormation_bonly(__agent_params,themap,tau);
  if (status)
    return 0.0;

  gsl_vector *workv = gsl_vector_alloc(VD);
  // copy out curlyC to workv
  gsl_vector_memcpy(workv, __agent_params->curlyC);

  // now compute the tau sample 
  gsl_vector_view ff = gsl_vector_view_array(f,VD);
  gsl_vector_view fb = gsl_vector_subvector(&ff.vector,SSD,SSD);
  gsl_blas_dgemv(CblasNoTrans, 1.0, __agent_params->curlyH, &fb.vector, 1.0, workv);
  // copy out curlyC to output mytout
  memcpy(tau_out,gsl_vector_ptr(workv,0),VD*sizeof(double));

  gsl_vector_free(workv);

  double deflection = 0;
  for (int k=0; k<VD; k++)
    deflection  += pow(tau_out[k]-f[k],2);
  return deflection;
}
// a helper function for sampling sentiments - legacy code that could be combined withe sampleNextSentiment
int setUpSampleNextSentimentRaw(samplerState *sstate, d_entity_type atype, reset_type rtype, double * f, double * tau)
{
  int *themap;

  sentimentParams * __agent_params = sstate->sentiParams;

  // at this spot, the dynamics are chosen to be either for agent or for client
  // so atype here acts somewhat like a "turn" variable, but only at this level
  if (atype == BEH_AGENT)
    themap = __agentmap;
  else // if (atype == BEH_CLIENT) 
    themap = __clientmap;

  
  gsl_matrix *siga, *sigf, *isiga, *isigf;
    // set up covariances
  if (rtype == RESET_NONE) {
    siga = __agent_params->sig_a;
    sigf = __agent_params->sig_f;
    isiga = __agent_params->isig_a;
    isigf = __agent_params->isig_f;
  } else {
    siga = __agent_params->sig_a_reset;
    isiga = __agent_params->isig_a_reset;
    if (rtype == RESET_AGENT) {
      sigf = __agent_params->sig_f_reset_agent;
      isigf = __agent_params->isig_f_reset_agent;
    } else if (rtype == RESET_CLIENT) {
      sigf = __agent_params->sig_f_reset_client;
      isigf = __agent_params->isig_f_reset_client;
    } else if (rtype == RESET_BOTH) {
      sigf = __agent_params->sig_f_reset_both;
      isigf = __agent_params->isig_f_reset_both;
    }
  }
  /*
  printMatrix(siga,"siga",VD,VD);
  printMatrix(sigf,"sigf",VD,VD);
  printMatrix(isiga,"isiga",VD,VD);
  printMatrix(isigf,"isigf",VD,VD);
  printMatrix(__agent_params->sig_f_reset_client,"sigf_reset_client",VD,VD);
  printMatrix(__agent_params->isig_f_reset_client,"isigf_reset_client",VD,VD);
  printMatrix(__agent_params->sig_a_reset,"siga_reset",VD,VD);
  */

  // set up impression formation equations
  int status;
  status=setUpImpressionFormation_bonly(__agent_params,themap,tau);
  if (status) 
    return status;

  gsl_matrix *work = gsl_matrix_alloc(VD,VD); 
  gsl_matrix *work2 = gsl_matrix_alloc(VD,VD);
  gsl_vector *workv = gsl_vector_alloc(VD);

#ifdef USING_INERTIA
    // curlyK^T*isig_a
    gsl_blas_dgemm (CblasTrans, CblasNoTrans, 1.0, __agent_params->curlyK, isiga, 0.0, work);

    // curlyK^T*sig_a*curlyK
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, work, __agent_params->curlyK, 0.0, work2);

    // curlyK^T*sig_a*curlyK + isig_f
    gsl_matrix_add(work2, isigf);

    // (curlyK^T*sig_a*curlyK + isig_f)^-1
    status = my_invert_matrix(work2,__agent_params->sig_n,VD);
    if (status)
      return status;

    // curlyK^T*isig_a
    gsl_blas_dgemm (CblasTrans, CblasNoTrans, 1.0, __agent_params->curlyK, isiga, 0.0, work2);
    // curlyK^T*isig_a*curlyC
    gsl_blas_dgemv (CblasNoTrans, 1.0, work2, __agent_params->curlyC, 0.0, workv);

    // curlyK^T*isig_a*curlyC + isig_f*f
    gsl_vector_view ff = gsl_vector_view_array(f,VD);
    //    printVector(&ff.vector,"f used is: ",VD);
    gsl_blas_dgemv (CblasNoTrans, 1.0, isigf, &ff.vector, 1.0, workv);

    // sig_n*(curlyK^T*isig_a*curlyC + isig_f*f);
    gsl_blas_dgemv (CblasNoTrans, 1.0, __agent_params->sig_n, workv, 0.0, __agent_params->mu_n);
    
#else
    // curlyKinv*sig_a
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,1.0, __agent_params->curlyKinv, siga, 0.0, work);
    // sig_n=curlyKinv*sig_a*curlyKinv^T
    gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, work, __agent_params->curlyKinv, 0.0, __agent_params->sig_n);
    
    // sig_n*curlyK^T
    gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, __agent_params->sig_n, __agent_params->curlyK, 0.0, work);
    
    // sig_n*curlyK^T*isig_a
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, work, isiga, 0.0, work2);
    
    // sig_n*curlyK^T*isig_a*curlyC
    gsl_blas_dgemv (CblasNoTrans, 1.0, work2, __agent_params->curlyC, 0.0, __agent_params->mu_n);
#endif
  // free temp storage
  gsl_matrix_free(work);
  gsl_matrix_free(work2);
  gsl_vector_free(workv);
  
  return status;
}
// a helper function for sampling sentiments - legacy code that could be combined withe sampleNextSentiment
// atype gives the type for the behaviour that you want to see
int computeInteractMean(samplerState *sstate, d_entity_type atype, double * f, double * tau)
{
  int *themap;
  int i,j;

  sentimentParams * __agent_params = sstate->sentiParams;

  // at this spot, the dynamics are chosen to be either for agent or for client
  // so atype here acts somewhat like a "turn" variable, but only at this level
  if (atype == BEH_AGENT)
    themap = __agentmap;
  else // if (atype == BEH_CLIENT) 
    themap = __clientmap;

  gsl_matrix *z = gsl_matrix_alloc(SSD,SSD);
  gsl_matrix *zinv = gsl_matrix_alloc(SSD,SSD);
  gsl_matrix *ftmp = gsl_matrix_alloc(VD,SSD);
  gsl_vector *tmpf = gsl_vector_alloc(VD);
  gsl_vector *tmpff = gsl_vector_alloc(VD);
  gsl_vector *tmpmean=gsl_vector_alloc(SSD);
  gsl_vector *tmpmean2=gsl_vector_alloc(SSD);


  // set up impression formation equations
  // for behaviour, actor and object
  //themap seems to make no difference here - so confusing
  
  int status = setUpImpressionFormation(__agent_params,themap,tau,1,1);
  if (status)
    return status;

  gsl_vector_view myff = gsl_vector_view_array(f,VD);

  //***********************************************************
  // First, compute optimal behaviour
  //***********************************************************
  //printf("getting optimal behaviour\n");
  //zero out the middle of f 
  for (i=0; i<VD; i++) 
    gsl_vector_set(tmpf,i,gsl_vector_get(&myff.vector,i));
  for (i=SSD; i<SSD+3; i++)
    gsl_vector_set(tmpf,i,0.0);
		   

  //tmpf is now [I -M']*I_b*g_b
  gsl_vector_sub (tmpf, __agent_params->curlyC);
  for (i=0; i<VD; i++) {
    for (j=0; j<SSD; j++) {
      gsl_matrix_set(ftmp,i,j,-1.0*gsl_matrix_get(__agent_params->curlyH,i,j));
      if (i==j+3)
	gsl_matrix_set(ftmp,i,j,1.0-1.0*gsl_matrix_get(__agent_params->curlyH,i,j));
    }
  }
  //ftmp is now [I -M']I_bS_b
  //compute ([I -M']I_bS_b)^T[I -M']I_b*S_b-->z
  gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.0,ftmp,ftmp,0.0,z);
  // invert z
  status = my_invert_matrix(z,zinv,SSD);
  if (status)
    return status;
  // compute ([I -M']I_bS_b)^T[I -M']I_b*g_b-->tmpmean
  gsl_blas_dgemv(CblasTrans,1.0,ftmp,tmpf,0.0,tmpmean);
  //compute -(z)^-1*([I -M']I_bS_b)^T[I -M']I_b*g_b --> tmpmean2
  gsl_blas_dgemv(CblasNoTrans,-1.0,zinv,tmpmean,0.0,tmpmean2);
  // copy tmpmean2 --> interact_mean
  for (i=0; i<SSD; i++)
    gsl_vector_set(__agent_params->interact_mean,i+SSD,gsl_vector_get(tmpmean2,i));


  //***********************************************************
  // compute optimal agent
  //***********************************************************
  //printf("getting optimal actor\n");
  gsl_vector *curlyC;
  gsl_matrix *curlyH;

  //zero out the top of f 
  for (i=0; i<VD; i++) 
    gsl_vector_set(tmpf,i,gsl_vector_get(&myff.vector,i));
  for (i=0; i<SSD; i++)
    gsl_vector_set(tmpf,i,0.0);
  // have to swap the curlyC matrices if going BEH_CLIENT because client is actually the actor...
  // so in curlyG -- the agent and client are supposed to be all swapped
  // all this is so that f,tau always remain with agent in agent spot and client in the client spot
  if (atype==BEH_AGENT) {
     curlyC=__agent_params->curlyCa;
     curlyH=__agent_params->curlyHa;
  } else {
     curlyC=__agent_params->curlyCc;
     curlyH=__agent_params->curlyHc;
  }
  //tmpf is now [I -M']*I_b*g_b
  gsl_vector_sub (tmpf,curlyC);
  for (i=0; i<VD; i++) {
    for (j=0; j<SSD; j++) {
      gsl_matrix_set(ftmp,i,j,-1.0*gsl_matrix_get(curlyH,i,j));
      if (i==j)
	gsl_matrix_set(ftmp,i,j,1.0-1.0*gsl_matrix_get(curlyH,i,j));
    }
  }
  //ftmp is now [I -M']I_bS_b
  //compute ([I -M']I_bS_b)^T[I -M']I_b*S_b-->z
  gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.0,ftmp,ftmp,0.0,z);
  // invert z
  status = my_invert_matrix(z,zinv,SSD);
  if (status)
    return status;
  // compute ([I -M']I_bS_b)^T[I -M']I_b*g_b-->tmpmean
  gsl_blas_dgemv(CblasTrans,1.0,ftmp,tmpf,0.0,tmpmean);
  //compute -(z)^-1*([I -M']I_bS_b)^T[I -M']I_b*g_b --> tmpmean2
  gsl_blas_dgemv(CblasNoTrans,-1.0,zinv,tmpmean,0.0,tmpmean2);
  // copy tmpmean2 --> interact_mean
  for (i=0; i<SSD; i++)
    gsl_vector_set(__agent_params->interact_mean,i,gsl_vector_get(tmpmean2,i));


  //***********************************************************
  // compute optimal client
  //***********************************************************
  //printf("getting optimal object\n");

    //zero out the bottom of f 
  for (i=0; i<VD; i++) 
    gsl_vector_set(tmpf,i,gsl_vector_get(&myff.vector,i));
  for (i=2*SSD; i<SSD*2+SSD; i++)
    gsl_vector_set(tmpf,i,0.0);
  // have to swap the curlyC matrices if going BEH_CLIENT because client is actually the actor...
  if (atype==BEH_AGENT) {
     curlyC=__agent_params->curlyCc;
     curlyH=__agent_params->curlyHc;
  } else {
     curlyC=__agent_params->curlyCa;
     curlyH=__agent_params->curlyHa;
  }
  //tmpf is now [I -M']*I_b*g_b
  gsl_vector_sub (tmpf,curlyC);
  for (i=0; i<VD; i++) {
    for (j=0; j<SSD; j++) {
      gsl_matrix_set(ftmp,i,j,-1.0*gsl_matrix_get(curlyH,i,j));
      if (i==j+SSD*2)
	gsl_matrix_set(ftmp,i,j,1.0-1.0*gsl_matrix_get(curlyH,i,j));
    }
  }
  //ftmp is now [I -M']I_bS_b
  //compute ([I -M']I_bS_b)^T[I -M']I_b*S_b-->z
  gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.0,ftmp,ftmp,0.0,z);
  // invert z
  status = my_invert_matrix(z,zinv,SSD);
  if (status)
    return status;
  // compute ([I -M']I_bS_b)^T[I -M']I_b*g_b-->tmpmean
  gsl_blas_dgemv(CblasTrans,1.0,ftmp,tmpf,0.0,tmpmean);
  //compute -(z)^-1*([I -M']I_bS_b)^T[I -M']I_b*g_b --> tmpmean2
  gsl_blas_dgemv(CblasNoTrans,-1.0,zinv,tmpmean,0.0,tmpmean2);
  // copy tmpmean2 --> interact_mean
  for (i=0; i<SSD; i++)
    gsl_vector_set(__agent_params->interact_mean,i+SSD*2,gsl_vector_get(tmpmean2,i));

  
  // free temp storage
  gsl_matrix_free(z);
  gsl_matrix_free(ftmp);
  gsl_vector_free(tmpf);
  gsl_vector_free(tmpff);
  gsl_vector_free(tmpmean);
  gsl_vector_free(tmpmean2);

  return status;
}

// the main function to sample from the posterior over fundamentals
// takes 5 arguments
// samplerState - contains sentiment Parameters and samples for the dyad given by dyad_index
// atype is either BEH_AGENT or BEH_CLIENT depending on who is acting for this sample
// reset type (can be "agent", "client" "both", or "none") depending on what reset is being attempted
// input sample current_sample
// output sample next_sample
int sampleNextSentiment(samplerState *sstate, d_entity_type atype,
			reset_type rtype, sample * current_sample, sample * next_sample)
{
  double * f = current_sample->f;
  double * tau = current_sample->tau;
  double * f_out = next_sample->f;
  double * tau_out = next_sample->tau;

  int status = setUpSampleNextSentimentRaw(sstate,atype,rtype,f,tau);
  if (!status) {
    // everything ok, actually draw the sample
    status = sampleASentiment(sstate,f_out,tau_out);
  } else {
    // there was an error in drawing the sample.
    // so we can go on, let's flag this up and then copy f,tau to f_out, tau_out directly
    memcpy(next_sample->f,current_sample->f,VD*sizeof(double));
    memcpy(next_sample->tau,current_sample->tau,VD*sizeof(double));
    printf("ERROR:  sample was not drawn correctly in sampleNextSentiment");
  }
  return status;
}

// computes the expected emotion for entity atype given a set of samples given by the dyad_index in sstate
// this computes the emotion for each state, weighted by the sample weight and averaged
// see computeEmotion(..)
// this computes \int_{f,\tau} \epsilon(f,\tau) b(f,\tau) dfd\tau
// as a set of samples y_i=\{f_i,\tau_i\} with weights w_i
// \sum_i \epsilon(f_i,\tau_i) w_i
int computeExpectedEmotion(dyad * the_dyad, double * e_out, c_entity_type atype)
{
  samplerState * sstate =  the_dyad->dsstate;
  sentimentParams * __agent_params = sstate->sentiParams;
  if ( __agent_params->num_features_modifiers == 0) 
    return 0;

  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  // this defines what part of f/tau we are going to be using
  int v_pos=atype*SSD;

  double tmp[SSD];
  for (int i=0; i<SSD; i++)
    e_out[i]=0.0;
  double totalweight=0.0;
  for (int i=0; i<numsamples; i++) {
    computeEmotion(sstate, the_samples[i].f+v_pos, the_samples[i].tau+v_pos, tmp);
    for (int j=0; j<SSD; j++)
      e_out[j]=e_out[j]+tmp[j]*the_samples[i].weight;
    totalweight += the_samples[i].weight;
  }
  for (int j=0; j<SSD; j++)
    e_out[j]/=totalweight;
  return 1;
  
}

// computes the emotion for entity atype given a set of samples given by the dyad_index in sstate
// by first computing the average fundamentals and transients, then computing the single emotion based on that
int computeEmotionFromAverage(dyad * the_dyad, double * e_out, c_entity_type who_acts)
{
  samplerState * sstate =  the_dyad->dsstate;
  sentimentParams * __agent_params = sstate->sentiParams;
  if ( __agent_params->num_features_modifiers == 0) 
    return 0;

  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  // this defines what part of f/tau we are going to be using
  int v_pos=(who_acts)*SSD;


  // gather summary statistics
  double f_avg[SSD];
  double tau_avg[SSD];

  for (int i=0; i<SSD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  double totalweight=0.0;
  // this loops through all the samples, and computes the weighted sum of fundamentals and transients,
  for (int i=0; i<numsamples; i++) {
      for (int j=0; j<SSD; j++) {
	f_avg[j] = f_avg[j]+the_samples[i].f[j+v_pos]*the_samples[i].weight;
	tau_avg[j] = tau_avg[j]+the_samples[i].tau[j+v_pos]*the_samples[i].weight;
      }
      totalweight += the_samples[i].weight;
  }
  for (int i=0; i<SSD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }
  for (int i=0; i<SSD; i++)
    e_out[i]=0.0;

  return computeEmotion(sstate, f_avg, tau_avg, e_out);
}
// get a default null emotion - this is used in planning when emotions are not being used
// it doesn't matter what index this returns, so long as the list is of length 1 and the prob is 1.0
int getNullEmotion(dyad *the_dyad, agentState *the_agent, int * index, double * prob, c_entity_type who_acts)
{
  *index=0;
  *prob=1.0;
  int num_emotions=1;
  return num_emotions;
}

// this computes the most likely next emotion by first computing a list of emotions with probabilty, then picking  the most probable one
// that is, it first computes the distirbution over denotative identities (see getNextEmotions)
// then picks the most likely one
int getNextMostLikelyEmotion(dyad *the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type whose_emotion, d_entity_type who_acted)
{
  int num_emotions = getNextEmotions(the_dyad, the_agent, emotion_list, whose_emotion, who_acted);
  // take the top one
  int i=0;
  int best_emot=-1;
  double max_prob=0.0;
  while (i<num_emotions) {
    if (i==0 || (*emotion_list)[i].prob > max_prob) {
      max_prob = (*emotion_list)[i].prob;
      best_emot = i;
    }
    i++;
  }
  // we only keep the top one
  free (*emotion_list);
  // re-allocate with one element
  (*emotion_list) = (emotion *) malloc(sizeof(emotion));
  (*emotion_list)[0].prob=1.0;
  (*emotion_list)[0].index=best_emot;
  num_emotions=1;
  return num_emotions;
  
}
// gets a list of emotions with normalized weights in *emotion_list corresponding to the currrent state in sstate->samples[dyad_index]
// and then return the most likely one in a list of length 1 with prob 1.0
int getCurrentMostLikelyEmotion(dyad *the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type who_acts)
{
  int num_emotions = getCurrentEmotions(the_dyad, the_agent, emotion_list, who_acts);
  // take the top one
  int i=0;
  int best_emot=-1;
  double max_prob=0.0;
  while (i<num_emotions) {
    //printf("%d %g\n",(*emotion_list)[i].index,(*emotion_list)[i].prob);
    if (i==0 || (*emotion_list)[i].prob > max_prob) {
      max_prob = (*emotion_list)[i].prob;
      best_emot = (*emotion_list)[i].index;
    }
    i++;
  }
  //printf("best emotion is %d\n",best_emot);
  // we only keep the top one
  free (*emotion_list);
  // re-allocate with one element
  (*emotion_list) = (emotion *) malloc(sizeof(emotion));
  (*emotion_list)[0].prob=1.0;
  (*emotion_list)[0].index=best_emot;
  num_emotions=1;
  return num_emotions;
  
}
// gets a list of emotions with normalized weights in *emotion_list corresponding to the currrent state in sstate->samples[dyad_index]
// and then return the a sample from this distribution
int sampleCurrentEmotion(dyad *the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type who_acts)
{
  behaviour * best_emot=NULL;
  int num_emotions = getCurrentEmotions(the_dyad, the_agent, emotion_list, who_acts);
  // sample one
  best_emot = drawRandomProb(the_dyad->dsstate->r,*emotion_list);
  // we only keep this one
  free (*emotion_list);
  // re-allocate with one element
  (*emotion_list) = (emotion *) malloc(sizeof(emotion));
  (*emotion_list)[0].prob=1.0;
  (*emotion_list)[0].index=best_emot->index;
  num_emotions=1;
  return num_emotions;
  
}

// gets a *sorted* list of emotions with normalized weights in *emotion_list corresponding to the next state in sstate->samples
int getNextEmotionList(dyad *the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type whose_emotion, d_entity_type who_acted)
{
  int num_emotions = getNextEmotions(the_dyad, the_agent, emotion_list, whose_emotion, who_acted);
  // sort
  qsort(*emotion_list, num_emotions, sizeof(denotativeState), denotativeStateCompare);
  return num_emotions;
}
// compute the next emotions that will be generated by default. No behaviour weights are used here as the action is undefined
// b(y') = \int_{y} P(y'|y,a=NULL) b(y)
// b(x') = then \int_{y'} G(\epsilon(y'),x')b(y')   where G is the emotion somatic transform
// now b(x') is a distribution over denotative emotions
int getNextEmotions(dyad *the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type whose_emotion, d_entity_type who_acted)
{
  // really don't need this - could just re-use a single sample
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  int numsamples=sstate->numsamples;
  sample * the_samples = sstate->samples;
  sample * new_samples = (sample *) malloc(numsamples*sizeof(sample));
  double totalweight = 0.0;
  //double aid_weight, cid_weight;
  int * counts_list=NULL;
  int v_pos=whose_emotion*SSD;
  double tmp[SSD];
  int num_emotions=0;
  int emot;
  for (int i=0; i<numsamples; i++)  {
    sampleNextState(the_dyad, who_acted, the_agent, RESET_NONE, the_samples+i, new_samples+i,  NULL, NULL, NULL, NULL);
    // possibly override the weight if we want this to be unconstrained by the denotative identity inertia
    //new_samples[i].weight=1.0;


    computeEmotion(sstate, new_samples[i].f+v_pos, new_samples[i].tau+v_pos, tmp);

    somaticTransform(DENOTATIVE, sstate->r, dyad_index, &emot, tmp, the_agent->entity_dicts[EMOTION]);
    if (new_samples[i].weight > 0) 
      num_emotions = insert_into_state_list(emotion_list, &counts_list, num_emotions, emot, new_samples[i].weight);
    totalweight += the_samples[i].weight;
  }
  for (int i=0; i<num_emotions; i++)  
    (*emotion_list)[i].prob /= totalweight;

  if (totalweight > 0)
    free(counts_list);
  free(new_samples);  // really don't need this - could just re-use a single sample

  return num_emotions;
}
//---------------------------------
// CORE EMOTION COMPUTING ROUTING
//---------------------------------
// computes an emotion from fundamental f and transient tau, each of which are SSD-arrays
// dREQ[i*SSD...(i+1)*SSD] is a SSD-vector (double array of length SSD) that contains the term that
// multiplies the features dreq_index[i][0...dreq_num_features[i]-1]
//where dreq_index[i][j] is an index into [ee,ep,ea...(more if SSD>3), fe,fp,fa...(more if SSD>3)]
int computeEmotion(samplerState *sstate, double *f, double *tau, double *e_out)
{

  int i, k, j;
  double tmp;
  sentimentParams * __agent_params = sstate->sentiParams;
  if ( __agent_params->num_features_modifiers == 0) 
    return 0;

  gsl_matrix *work = gsl_matrix_alloc(SSD,SSD); 
  gsl_matrix *work2 = gsl_matrix_alloc(SSD,SSD);
  gsl_vector *emot = gsl_vector_alloc(SSD);

  // compute normalizing factor - first gather E
  gsl_matrix_set_zero(work);
  for (k=0; k<__agent_params->num_features_modifiers; k++) {
    if (__agent_params->dreq_num_features[k]==1 && __agent_params->dreq_index[k][0] < SSD)  {  // part of E
      // the __agent_params->dreq_index[k][0]th row of E
      for ( i = 0; i < SSD; i++) 
	gsl_matrix_set(work,__agent_params->dreq_index[k][0],i,__agent_params->dREQ[k*SSD+i]);
    }
  }

  
  // now 
  gsl_matrix_set_zero(work2);
  // now add the I*Q terms
  for (k=0; k<__agent_params->num_features_modifiers; k++) {
    // interaction terms
    if (__agent_params->dreq_num_features[k]==2 &&  __agent_params->dreq_index[k][0]<SSD && __agent_params->dreq_index[k][1]>=SSD) {
      // the multiplies __agent_params->dreq_index[k][1]th term in f and gets added to the
      for ( i = 0; i < SSD; i++) {
	// the jth column of Q_{__agent_params->dreq_index[k][1]}
	j=__agent_params->dreq_index[k][0];
	tmp = gsl_matrix_get(work2, i, j);
	// was like this until December 15 2017:
	//gsl_matrix_set(work2,i,i,tmp+__agent_params->dREQ[k*SSD+i]*f[__agent_params->dreq_index[k][0]]);
	gsl_matrix_set(work2,i,j,tmp+__agent_params->dREQ[k*SSD+i]*f[__agent_params->dreq_index[k][1]-SSD]);
      }
    }
  }
   
  gsl_matrix_add(work,work2);

  //work2 is (E+I*f*Q)^{-1}
  if (my_invert_matrix(work,work2,SSD)) {
    gsl_matrix_free(work);
    gsl_matrix_free(work2);
    gsl_vector_free(emot);
    return 1;
  }

  // now compute tau-R*f-d and multiply by work2
  
  // compute emotion term
  for (i = 0; i < SSD; i++) 
    gsl_vector_set(emot,i,tau[i]);
  for (k=0; k<__agent_params->num_features_modifiers; k++) {
    if (__agent_params->dreq_num_features[k]==1 && __agent_params->dreq_index[k][0] >= SSD)  {   // part of R
      // the (__agent_params->dreq_index[k][0]-SSD)th row of R
      tmp = gsl_vector_get(emot,__agent_params->dreq_index[k][0]-SSD);  // tau
      for (j = 0; j < SSD; j++) 
	tmp -= __agent_params->dREQ[k*SSD+j]*f[j];   //-R*f TOCHECK - had + here and gave interesting results
      gsl_vector_set(emot,__agent_params->dreq_index[k][0]-SSD,tmp);
    }
    // vector d
    if (__agent_params->dreq_num_features[k]==0) {
      for (j = 0; j < SSD; j++) {
	tmp = gsl_vector_get(emot,j)-__agent_params->dREQ[k*SSD+j];   // TOCHECK - had + here and gave interesting results
	gsl_vector_set(emot,j,tmp);
      }
    }
  }
  //emot = work2*emot;
  for (i = 0; i < SSD; i++) {
    e_out[i]=0.0;
    for (j = 0; j < SSD; j++)
      e_out[i] = e_out[i] + gsl_matrix_get(work2,i,j)*gsl_vector_get(emot,j);  // TOCHECK: replace this with gsl_vector_get(emot,i) <--gives interesting results??
  }

  gsl_matrix_free(work);
  gsl_matrix_free(work2);
  gsl_vector_free(emot);
  
  return 0;
}
  
// gets a *sorted* list of emotions with normalized weights in *emotion_list corresponding to the currrent state in sstate->samples
int getCurrentEmotionList(dyad * the_dyad, agentState *the_agent, emotion ** emotion_list, c_entity_type who_acts)
{
  int num_emotions = getCurrentEmotions(the_dyad, the_agent, emotion_list, who_acts);
  // sort
  qsort(*emotion_list, num_emotions, sizeof(denotativeState), denotativeStateCompare);
  return num_emotions;
}
// gets a list of emotions with normalized weights in *emotion_list corresponding to the currrent state in sstate->samples[dyad_index]
// b(x) = then \int_{y} G(\epsilon(y),x)b(y)   where G is the emotion somatic transform
// now b(x) is a distribution over denotative emotions corresponding to the current state
int getCurrentEmotions(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type who_acts)
{
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  sentimentParams * __agent_params = sstate->sentiParams;
  if ( __agent_params->num_features_modifiers == 0) 
    return 0;

  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  int num_emotions = 0;
  int * counts_list=NULL;
  double totalweight = 0.0;
  // the position in the vector for the agent we are computing this for (AGENT or CLIENT for this dyad)
  int v_pos=(who_acts)*SSD;
  double tmp[SSD];
  int emot;
  double emot_weight;
  for (int i=0; i<numsamples; i++) {
     // only do ones that didn't reset  -- why was this done?
    // TODO: remove
    if (the_samples[i].reset_flag==0) {
      computeEmotion(sstate, the_samples[i].f+v_pos, the_samples[i].tau+v_pos, tmp);
      emot_weight=somaticTransform(DENOTATIVE, sstate->r, dyad_index, &emot, tmp, the_agent->entity_dicts[EMOTION]);
      
      if (WEIGHT_EMOTIONS_BY_DISTANCE==0) 
	emot_weight=1.0;
      
      num_emotions = insert_into_state_list(emotion_list, &counts_list, num_emotions, emot, the_samples[i].weight*emot_weight);
      totalweight += the_samples[i].weight*emot_weight;
    }
  }
  //now, if the number of emotions returned is 0, we need to try again this time with samples that were reset - basically this
  //has all transients equal to fundamentals so is the set of characteristic emotions instead
  if (num_emotions==0) {
    for (int i=0; i<numsamples; i++) {
      computeEmotion(sstate, the_samples[i].f+v_pos, the_samples[i].tau+v_pos, tmp);
      emot_weight=somaticTransform(DENOTATIVE, sstate->r, dyad_index, &emot, tmp, the_agent->entity_dicts[EMOTION]);
      if (WEIGHT_EMOTIONS_BY_DISTANCE) 
	emot_weight=1.0;
      num_emotions = insert_into_state_list(emotion_list, &counts_list, num_emotions, emot, the_samples[i].weight*emot_weight);
      totalweight += the_samples[i].weight*emot_weight;
    }
  }

  for (int i=0; i<num_emotions; i++)  
    (*emotion_list)[i].prob /= totalweight;

  free(counts_list);

  return num_emotions;
}

// general purpose emotion getting routine. This will call the appropriate helper function based on whether emotions
// are being used at all (sstate->use_emotions) and whether emotions are being used in planning or not sstate->emotion_planning
int getCurrentEmotion(dyad * the_dyad, agentState *the_agent, emotion ** emot_list, c_entity_type who_acts)
{
  samplerState * sstate = the_dyad->dsstate;
  int num_potential_emotions;
  int index;
  double prob;
  if (sstate->use_emotions) {
    if (sstate->emotion_planning)  
      num_potential_emotions = getCurrentEmotionList(the_dyad, the_agent, emot_list, AGENT);
    else 
      num_potential_emotions = getCurrentMostLikelyEmotion(the_dyad, the_agent, emot_list, AGENT);   // will be only a single emotion
  } else {
    num_potential_emotions = getNullEmotion(the_dyad, the_agent, &index, &prob, AGENT);   // will be only a single emotion: "no emotion"
    emot_list[0]->prob=prob;
    emot_list[1]->index=index;
  }
  return num_potential_emotions;
}

// predict next default action by sampling from the next state and taking the behaviour f_b
// then sampling from x_b given f_b with a DENOTATIVE  somatic transform
// on input, beh_list and count_list are expected to be null
// on output, behaviour_list is a list of behaviour indices
// sorted by their weights (give in .prob field)
// this computes
// b(x',y') = \int_{y} P(y',x'|y,x,a=NULL) b(y)
// b(x_b') = \int_{x_a',x_c'}\int_{y'} G(f_b',x')b(f_b',x')   where G is the BEHAVIOUR somatic transform
// now b(x_b') is a distribution over denotative behaviours x_b'
// 
// if sstate->default_action is set, then this returns the action list from the average state only
// see getNextActionListFromAverageState
int getNextAction(dyad *the_dyad, agentState * the_agent, identity ** behaviour_list,  d_entity_type who_acted)
{
  samplerState * sstate = the_dyad->dsstate;
  if (sstate->default_action) {
    // short circuit and get the default action only for the average state 
    getNextActionListFromAverageState(the_dyad, the_agent, behaviour_list, who_acted);
    return 1;
  }

  int num_potential_actions = 0;

  // really don't need this - could just re-use a single sample
  sample * new_samples = (sample *) malloc(sstate->numsamples*sizeof(sample));
  sample * the_samples = sstate->samples;
  double totalweight = 0.0;
  int * counts_list=NULL;
  for (int i=0; i<sstate->numsamples; i++)  {
    sampleNextState(the_dyad, who_acted, the_agent, RESET_NONE, the_samples+i, new_samples+i,  NULL, NULL, NULL, NULL);

    // here the institutional filters for action would still be active - so X[BEH_AGENT/CLIENT] at this point should include
    // the institutional filters - that is, new_samples[i].x[BEH_AGENT] at this point will be sampled from an institutionally
    // correct posterior. However, we might
    // this somatic transform is going to run through the whole dictionary but we want it to only do the identities that
    // are institutionally correct -- that is, that are supported by new_samples[i].x[BEH_AGENT] (which is a single identity) - so
    // in fact this DENOTATIVE Somatic Transform collapses into a WEIGHT somatic transform.
    // however, we could do a second step where the behaviour is "adjusted" to the nearest institutionally correct one,
    // copy out the behaviour of the agent that acted
    // INSTITUTTIONAL FILTERS: THIS USED TO BE A DENOTATIVE SOMATIC TRANSFORM
    // but now its weighted - is this going to ruin my samples?
    // places where there are changes
    // 1)  here
    // 2) in updateXSample  -- see there
    // 3) I changed defaultAPP to fake an institutional filter by setting the identities to 373 (rape)
    // REMOVED HERE NOW ALTOGETHER AS IT HAPPENS INSIDE sampleNextState
    // 
    //new_samples[i].weight += somaticTransform(WEIGHT, sstate, dyad_index,&(new_samples[i].x[who_acted]),
    //					      new_samples[i].f+SSD*BEHAVIOUR, the_agent->entity_dicts[BEHAVIOUR]);

    if (new_samples[i].weight > 0) 
	num_potential_actions = insert_into_state_list(behaviour_list, &counts_list, num_potential_actions,
						       new_samples[i].x[who_acted], new_samples[i].weight);
    totalweight += new_samples[i].weight;
  }
#ifdef VERBOSE
  printf("totalweight in getNextAction: %g\n",totalweight);
#endif
  // normalize
  for (int i=0; i<num_potential_actions; i++)  
    (*behaviour_list)[i].prob /= totalweight;
  if (num_potential_actions<1) {
    // special case must be handled - this means there was no denotative action with any denotative weight at all
    // which was close enough to the predicted distribution over f/tau computed in sampleNextState.
    // this means the denotative action set collectively has zero weight, so is therefore all equally likely given the
    // connotative state.
    
    for (int i=0; i<sstate->numsamples; i++)  
      num_potential_actions = insert_into_state_list(behaviour_list, &counts_list, num_potential_actions,
						     new_samples[i].x[who_acted], 1.0);
    for (int i=0; i<num_potential_actions; i++)  
      (*behaviour_list)[i].prob /= sstate->numsamples;
#ifdef VERBOSE
    printf("no potential actions found, reverting to denotative num_potential_actions %d\n",num_potential_actions);
#endif
    
  }

  // sort
  qsort(*behaviour_list, num_potential_actions, sizeof(denotativeState), denotativeStateCompare);
  
  if (totalweight > 0) 
    free(counts_list);
  
  free(new_samples);  // really don't need this - could just re-use a single sample

  return num_potential_actions;
}

// Computes the (Weighted by identity) average next state and extracts the average (predicted) behaviour f_b
// and then extracts a single best next behaviour using a DENOTATIVE somatic transform
// this will produce a prediction that is closer to that predicted by Interact
// returns the value in the list behaviour_list with one identity element with prob=1.0
void getNextActionListFromAverageState(dyad *the_dyad, agentState *the_agent, identity ** behaviour_list,  d_entity_type who_acted)
{
  (*behaviour_list) = (emotion *) malloc(sizeof(emotion));
  (*behaviour_list)[0].index = getNextActionFromAverageState(the_dyad, the_agent, who_acted);
  (*behaviour_list)[0].prob = 1.0;
}
// same as getNextActionListFromAverageState but returns the best action as an integer index 
int getNextActionFromAverageState(dyad *the_dyad, agentState *the_agent, d_entity_type who_acted)
{
  samplerState * sstate = the_dyad->dsstate;
  // really don't need this - could just re-use a single sample
  sample * new_samples = (sample *) malloc(sstate->numsamples*sizeof(sample));
  sample * the_samples = sstate->samples;
  double totalweight = 0.0;
  int the_behaviour;

  // for gathering summary statistics
  double *f_avg = (double *) malloc(VD*sizeof(double));
  double *tau_avg = (double *) malloc(VD*sizeof(double));

  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }

  for (int i=0; i<sstate->numsamples; i++)  {
    sampleNextState(the_dyad, who_acted, the_agent, RESET_NONE, the_samples+i, new_samples+i,  NULL, NULL, NULL, NULL);
    totalweight += new_samples[i].weight;
    
    for (int j=0; j<VD; j++) {
      f_avg[j] = f_avg[j]+new_samples[i].f[j]*new_samples[i].weight;
      tau_avg[j] = tau_avg[j]+new_samples[i].tau[j]*new_samples[i].weight;
    }
  }
  if (totalweight > 0.0 ){
    for (int i=0; i<VD; i++) {
      f_avg[i] = f_avg[i]/totalweight;
      tau_avg[i] = tau_avg[i]/totalweight;
    }
    somaticTransform(DENOTATIVE, sstate->r, the_dyad->dyad_index, &the_behaviour, f_avg+BEHAVIOUR*SSD, the_agent->entity_dicts[BEHAVIOUR]);

  } else {
    the_behaviour=-1;
  }
  free(f_avg);
  free(tau_avg);
  free(new_samples);  // really don't need this - could just re-use a single sample

  return the_behaviour;
}



// function to grow a list of denotativeStates id_list by adding index to it
// if index is in id_list already, then count_list is incremented at the same slot
// otherwise, id_list and count_list are grown by list_increment and count_list's new element is initialised to 1
// the weights are added to the "prob" field of each denotativeState.
// this can then be used with a sum of all weights to normalize
int insert_into_state_list(denotativeState ** id_list, int ** count_list, int id_list_length, int index, double weight)
{
  int i=0;
  identity *tmp_id_list;
  int * tmp_count_list;
  // amount by which to grow the lists each time
  int list_increment=1;

  while (i<id_list_length && ((*id_list)+i)->index != index)
    i++;
  if (i>=id_list_length) {
    // not found - add to identity list 
    if ((*id_list) == NULL) {
      id_list_length=1;
    } else {
      // save old lists
      tmp_id_list = *id_list;
      tmp_count_list = *count_list;
      id_list_length += list_increment;
    }
    (*id_list) = (identity *) malloc(id_list_length*sizeof(identity));
    (*count_list) = (int *) malloc(id_list_length*sizeof(int));
    // copy old lists to new lists
    for (int j=0; j<id_list_length-list_increment; j++) {
      ((*id_list)+j)->index = tmp_id_list[j].index;
      ((*id_list)+j)->prob = tmp_id_list[j].prob;
      (*count_list)[j] = tmp_count_list[j];
    }
    ((*id_list)+i)->index = index;
    ((*id_list)+i)->prob = 0.0;
    (*count_list)[i] = 0;
    // free temp storage
    if (id_list_length-list_increment > 0) {
      free(tmp_id_list);
      free(tmp_count_list);
    }
  }
  ((*id_list)+i)->prob += weight;
  (*count_list)[i]++;
  return id_list_length;
}
void printRawSample(sample * the_sample)
{
    printf("agent id: %d agent beh: %d client id: %d client_beh %d who_acted %d: ",the_sample->x[ID_AGENT],the_sample->x[BEH_AGENT],
	   the_sample->x[ID_CLIENT],the_sample->x[BEH_CLIENT],the_sample->who_acted);
    // XX:
    printRawXXstate(&(the_sample->xx));
    printf("f: ");
    for (int j=0; j<VD; j++) 
      printf("%4.3g ",the_sample->f[j]);
    printf("\ntau: ");
    for (int j=0; j<VD; j++) 
      printf("%4.3g ",the_sample->tau[j]);
    printf("\nweight: %4.3g ",the_sample->weight);
    printf("\n");
}
void printRawSampleCSV(sample * the_sample, FILE *fpout, char *csv_header_line, char*prefix)
{
  char tmp_header_line[1024];
  fprintf(fpout,"%4.3g,",the_sample->weight);
  for (int j=0; j<VD; j++) 
    fprintf(fpout,"%4.3g,",the_sample->f[j]);
  for (int j=0; j<VD; j++) 
    fprintf(fpout,"%4.3g,",the_sample->tau[j]);

  sprintf(tmp_header_line,"weight,");  
  strcat(csv_header_line,tmp_header_line);
  for (int i=0; i<VD; i++) {
    sprintf(tmp_header_line,"%s fundamental %d,",prefix,i);
    strcat(csv_header_line,tmp_header_line);
  }
  for (int i=0; i<VD; i++) {
    sprintf(tmp_header_line,"%s transient %d,",prefix,i);
    strcat(csv_header_line,tmp_header_line);
  }

}


void printEmotionsCSV(FILE *fpout, int iteration, char *csv_header_line, double * emot, char *bestemotlabel, char * actorname, char * clientname)
{
  char tmp_header_line[1024];
  if (iteration==0) {
    sprintf(tmp_header_line,"%s: emotion for %s,",actorname,clientname);
    strcat(csv_header_line,tmp_header_line);
  }
  fprintf(fpout,"\"");
  for (int j=0; j<SSD; j++) {
    fprintf(fpout,"%4.3g",emot[j]);
    if (j+1<SSD)
      fprintf(fpout,",");
  }
  fprintf(fpout,"\",");
  if (iteration==0) {
    sprintf(tmp_header_line,"%s: emotion for %s (label),",actorname,clientname);
    strcat(csv_header_line,tmp_header_line);
	      }
  fprintf(fpout,"%s,",bestemotlabel);
}



// gets the expected identities for agent and client
// this are returned in x_c_avg and x_a_avg;
int getIdentities(samplerState * sstate,
		  int *num_identities_agent, identity **x_a_avg,
		  int *num_identities_client, identity **x_c_avg)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  double totalweight;

  *num_identities_agent=0;
  *num_identities_client=0;
  // XX: int num_xx=0;
  int * count_list_agent = NULL;
  int * count_list_client = NULL;
  // XX: int * count_list_xx = NULL;

  // these start null -they will be allocated in insert_into_state_list
  *x_c_avg = NULL;
  *x_a_avg = NULL;
  totalweight=0.0;

  // this loops through all the samples, 
  // and builds arrays x_c_avg, x_a_avg, turn_avg that are of type denotativeState,
  // each element of these arrays is a denotative state encountered in the sample set, and the "prob" field of the denotativeState
  // is a sum of the weights of all those samples.
  // count_list_client, count_list_agent and count_list_turn are lists of counts
  // for each corresponding denotativeState, which are not even used in fact since the sample weights are used to normalize
  for (int i=0; i<numsamples; i++) {
    *num_identities_client = insert_into_state_list(x_c_avg,&count_list_client,*num_identities_client,
						    the_samples[i].x[ID_CLIENT],the_samples[i].weight);
    *num_identities_agent = insert_into_state_list(x_a_avg,&count_list_agent,*num_identities_agent,
						   the_samples[i].x[ID_AGENT],the_samples[i].weight);
    totalweight += the_samples[i].weight;
  }
  for (int i=0; i<*num_identities_client; i++) 
    (*x_c_avg)[i].prob /= totalweight;
  for (int i=0; i<*num_identities_agent; i++) 
    (*x_a_avg)[i].prob /= totalweight;

  qsort(*x_c_avg, *num_identities_client, sizeof(denotativeState), denotativeStateCompare);
  qsort(*x_a_avg, *num_identities_agent, sizeof(denotativeState), denotativeStateCompare);

  free(count_list_client);
  free(count_list_agent);
  return 1;
}

int printSamples(dyad *the_dyad, agentState * the_agent)
{
  return printSamplesCSV(the_dyad, the_agent,NULL,NULL,NULL,"");
}
// function to print raw samples to csv
// if max_num is less than numsamples, will randomly sample max_num samples
// if max_num is 0, will use all samples
// if max_num < 0, will not print any samples at all
void printRawSamplesCSV(samplerState *sstate, agentState * the_agent, agentState *the_client, FILE *fpout, char *csv_header_line, int max_num, char *prefix)
{
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  //double tmp_emotion[SSD];
  int j;
  if (max_num < 0) 
    return;
  if (max_num < 1) 
    max_num=numsamples;

  // XX: int * count_list_xx = NULL;
  //prints out the raw samples
  int output_csv=0;
  if (fpout)
    output_csv=1;
  
  char istr[64];
  for (int i=0; i<max_num; i++) {
    if (max_num==numsamples)
      j=i;
    else
      j=gsl_rng_uniform_int(sstate->r,max_num);   // may be repeats
    sprintf(istr, "%d", i);    
    strcat(prefix,istr);
    printRawSampleCSV(the_samples+j,fpout,csv_header_line, prefix);
  }  
}
// prints out variances of all samples to the CSV file. prints f variance, then tau variance
// only is meant to work with weights that are all 1.0
// because I am not sure how to normalize in the case of non-unitary weights
// returns zero on success
int printRawVariancesCSV(samplerState * sstate,  agentState * the_agent, agentState *the_client, FILE *fpout, char *csv_header_line, char *prefix)
{
  char tmp_header_line[1024];

  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  //prints out the raw samples
  int output_csv=0;
  if (fpout)
    output_csv=1;
  double *f_avg,*f2_avg,*tau_avg,*tau2_avg;
  double totalweight=0;
  f_avg = (double *) malloc(VD*sizeof(double));
  tau_avg = (double *) malloc(VD*sizeof(double));
  f2_avg = (double *) malloc(VD*sizeof(double));
  tau2_avg = (double *) malloc(VD*sizeof(double));

  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
    f2_avg[i] = 0.0;
    tau2_avg[i] = 0.0;
  }
  for (int j=0; j<numsamples; j++) {
    for (int k=0; k<VD; k++) {
      f_avg[k] = f_avg[k]+the_samples[j].f[k]*the_samples[j].weight;
      tau_avg[k] = tau_avg[k]+the_samples[j].tau[k]*the_samples[j].weight;
    }
    totalweight += the_samples[j].weight;
  }
  if ((int) totalweight != numsamples) 
    return (1);
  
  for (int i=0; i<VD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }
  // compute variances they easy way - with two passes
  totalweight=0.0;
  for (int j=0; j<numsamples; j++) {
    for (int k=0; k<VD; k++) {
      f2_avg[k]=f2_avg[k]+(the_samples[j].f[k]-f_avg[k])*(the_samples[j].f[k]-f_avg[k])*the_samples[j].weight;
      tau2_avg[k] = tau2_avg[k]+(the_samples[j].tau[k]-tau_avg[k])*(the_samples[j].tau[k]-tau_avg[k])*the_samples[j].weight;
    }
    totalweight += the_samples[j].weight;
  }
  if (totalweight<1.0) 
    return 1;

  for (int i=0; i<VD; i++) {
    f2_avg[i] = f2_avg[i]/(totalweight-1);
    fprintf(fpout,"%4.3g,",f2_avg[i]);
  }
  for (int i=0; i<VD; i++) {
    tau2_avg[i] = tau2_avg[i]/(totalweight-1.0);
    fprintf(fpout,"%4.3g,",tau2_avg[i]);
  }
  for (int i=0; i<VD; i++) {
    sprintf(tmp_header_line,"%s: fundamental variance %d,",prefix,i);
    strcat(csv_header_line,tmp_header_line);
  }
  for (int i=0; i<VD; i++) {
    sprintf(tmp_header_line,"%s: transient variance %d,",prefix,i);
    strcat(csv_header_line,tmp_header_line);
  }
  free(f_avg);
  free(tau_avg);
  free(f2_avg);
  free(tau2_avg);
  return 0;
}
// prints summary of samples out to CSV
// returns 1 if weights don't sum to > 1
// returns 0 on success
int printSamplesCSV(dyad *the_dyad, agentState * the_agent, agentState *the_client, FILE *fpout, char *csv_header_line, char * prefix)
{
  samplerState * sstate = the_dyad->dsstate;
  int dyad_index = the_dyad->dyad_index;
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  // TODO: add an expectation vector for the behaviour here as well - it may be a distribution (x_b)
  double * f_avg, * tau_avg;
  identity * x_c_avg, *x_a_avg;
  xxstate xx_avg;
  // TODO : add in average for behaviours
  //behaviour * x_bc_avg, * x_ba_avg;
  double totalweight;
  
  int emot;
  int num_identities_agent=0;
  int num_identities_client=0;

  int num_behaviours_agent=0;
  int num_behaviours_client=0;
  // XX: int num_xx=0;
  int * count_list_agent = NULL;
  int * count_list_client = NULL;
  // XX: int * count_list_xx = NULL;
  //prints out the raw samples
  int output_csv=0;
  if (fpout)
    output_csv=1;
    

#ifdef VERBOSE
  printf("number of samples %d roughening noise %g\n",sstate->numsamples, sstate->roughenNoise);
#endif
  // gather summary statistics
  f_avg = (double *) malloc(VD*sizeof(double));
  tau_avg = (double *) malloc(VD*sizeof(double));

  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  // these start null -they will be allocated in insert_into_state_list
  x_c_avg = NULL;
  x_a_avg = NULL;
  initializeXXstate(&xx_avg);
  totalweight=0.0;
  
  // this loops through all the samples, and computes the weighted sum of fundamentals and transients,
  // and builds arrays x_c_avg, x_a_avg, turn_avg that are of type denotativeState,
  // each element of these arrays is a denotative state encountered in the sample set, and the "prob" field of the denotativeState
  // is a sum of the weights of all those samples.
  // count_list_client, count_list_agent and count_list_turn are lists of counts
  // for each corresponding denotativeState, which are not even used in fact since the sample weights are used to normalize
  for (int i=0; i<numsamples; i++) {
      for (int j=0; j<VD; j++) {
	f_avg[j] = f_avg[j]+the_samples[i].f[j]*the_samples[i].weight;
	tau_avg[j] = tau_avg[j]+the_samples[i].tau[j]*the_samples[i].weight;
      }
      addToXXstate(&xx_avg,&(the_samples[i].xx),the_samples[i].weight);
      num_identities_client = insert_into_state_list(&x_c_avg,&count_list_client,num_identities_client,
						     the_samples[i].x[ID_CLIENT],the_samples[i].weight);
      num_identities_agent = insert_into_state_list(&x_a_avg,&count_list_agent,num_identities_agent,
						    the_samples[i].x[ID_AGENT],the_samples[i].weight);

      totalweight += the_samples[i].weight;

      if (the_samples[i].who_acted==BEH_AGENT) 
	num_behaviours_agent++;
      if (the_samples[i].who_acted==BEH_CLIENT) 
	num_behaviours_client++;
  }
#ifdef VERBOSE
  printf("the total weight: %4.3g\n",totalweight);
#endif
  if (totalweight < 1.0) 
    return 1;

  for (int i=0; i<VD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }
  for (int i=0; i<num_identities_client; i++) 
    x_c_avg[i].prob /= totalweight;
  for (int i=0; i<num_identities_agent; i++) 
    x_a_avg[i].prob /= totalweight;

  qsort(x_c_avg, num_identities_client, sizeof(denotativeState), denotativeStateCompare);
  qsort(x_a_avg, num_identities_agent, sizeof(denotativeState), denotativeStateCompare);

  divideXXstate(&xx_avg,totalweight);

  double *tmp_emotion = (double *) malloc(SSD*sizeof(double));
  computeEmotion(sstate, f_avg+AGENT*SSD, tau_avg+AGENT*SSD, tmp_emotion);

#ifdef VERBOSE
  printf("fundamental output average \nae\tap\taa\tbe\tbp\tba\tce\tcp\tca:\n");
  for (int i=0; i<VD; i++)
    printf("%4.3g\t",f_avg[i]);
  printf("\n\n");
    
  printf("transient output average \nae\tap\taa\tbe\tbp\tba\tce\tcp\tca:\n");
  for (int i=0; i<VD; i++)
    printf("%4.3g\t",tau_avg[i]);
  printf("\n\n");

  printf("client identity average labels (probabilities): \n");
  for (int i=0; i<num_identities_client; i++) 
    printf("%s (%4.3g) ",the_agent->entity_dicts[CLIENT]->theDictionary[x_c_avg[i].index].label,x_c_avg[i].prob);
  printf("\n\n");
  printf("agent identity average labels (probabilities): \n");
  for (int i=0; i<num_identities_agent; i++) 
    printf("%s (%4.3g) ",the_agent->entity_dicts[AGENT]->theDictionary[x_a_avg[i].index].label,x_a_avg[i].prob);
  printf("\n\n");
  printf("emotions from averaged fundamentals and transients: \n");

  printf("agent: \n");
  for (int i=0; i<SSD; i++)
    printf("%4.3g\t",tmp_emotion[i]);
#endif

  somaticTransform(DENOTATIVE, sstate->r, dyad_index, &emot, tmp_emotion, the_agent->entity_dicts[EMOTION]);
  computeEmotion(sstate, f_avg+CLIENT*SSD, tau_avg+CLIENT*SSD, tmp_emotion);

#ifdef VERBOSE
  printf("emotion label: %s\n",the_agent->entity_dicts[EMOTION]->theDictionary[emot].label);
  printf("\n");
  printf("client: \n");
  for (int i=0; i<SSD; i++)
    printf("%4.3g\t",tmp_emotion[i]);
#endif 
  somaticTransform(DENOTATIVE, sstate->r, dyad_index, &emot, tmp_emotion, the_agent->entity_dicts[EMOTION]);
#ifdef VERBOSE
  printf("emotion label: %s\n",the_agent->entity_dicts[EMOTION]->theDictionary[emot].label);
  printf("\n");
#endif 
  free(tmp_emotion);

  char tmp_header_line[1024];
  if (output_csv) {
    sprintf(tmp_header_line,"%s: agent name,%s: agent ids,%s: agent id probabilities,",prefix,prefix,prefix);
    strcat(csv_header_line,tmp_header_line);
    fprintf(fpout,"%s,",the_agent->name);
    fprintf(fpout,"\"");
    for (int i=0; i<num_identities_agent; i++) {
      fprintf(fpout,"%s",the_agent->entity_dicts[AGENT]->theDictionary[x_a_avg[i].index].label);
      if (i+1<num_identities_agent) {
	fprintf(fpout,",");
      }
    }
    fprintf(fpout,"\",\"");
    for (int i=0; i<num_identities_agent; i++) {
      fprintf(fpout,"%4.3g",x_a_avg[i].prob);
      if (i+1<num_identities_agent) {
	fprintf(fpout,",");
      }
    }
    fprintf(fpout,"\",");
    sprintf(tmp_header_line,"%s: client name,%s: client ids,%s: client id probabilities,",prefix,prefix,prefix);
    strcat(csv_header_line,tmp_header_line);
    fprintf(fpout,"%s,",the_client->name);
    fprintf(fpout,"\"");
    for (int i=0; i<num_identities_client; i++) {
      fprintf(fpout,"%s",the_agent->entity_dicts[CLIENT]->theDictionary[x_c_avg[i].index].label);
      if (i+1<num_identities_client) {
	fprintf(fpout,",");
      }
    }
    fprintf(fpout,"\",\"");
    for (int i=0; i<num_identities_client; i++) {
      fprintf(fpout,"%4.3g",x_c_avg[i].prob);
      if (i+1<num_identities_client) {
	fprintf(fpout,",");
      }
    }
    fprintf(fpout,"\",");
    for (int i=0; i<VD; i++) {
      sprintf(tmp_header_line,"%s: fundamental sentiment %d,",prefix,i);
      strcat(csv_header_line,tmp_header_line);
    }
    for (int i=0; i<VD; i++) {
      sprintf(tmp_header_line,"%s: transient %d,",prefix,i);
      strcat(csv_header_line,tmp_header_line);
    }
    for (int i=0; i<VD; i++) 
      fprintf(fpout,"%4.3g,",f_avg[i]);
    for (int i=0; i<VD; i++)
      fprintf(fpout,"%4.3g,",tau_avg[i]);

  }  
#ifdef VERBOSE  
  printf("number of samples that used agent behaviour %d\n",num_behaviours_agent);
  printf("number of samples that used client behaviour %d\n",num_behaviours_client);
  printf("xx average: \n");
#endif
  //XX:
  // TODO: also print to CSV
  printRawXXstate(&xx_avg);
  free(f_avg);
  free(tau_avg);
  free(x_c_avg);
  free(x_a_avg);
  free(count_list_client);
  free(count_list_agent);
  return 1;
}

// samples an observation from the next state if action agent_behaviour is taken and client_behaviour_obs is observed
// does this computation in a hacky way by computing the average xx state first, then updating the xx state and drawing a sample
int sampleNextXObservation(samplerState *sstate, int agent_behaviour, int client_behaviour_obs, xostate * xx_obs)
{
  double *f_avg = (double *) malloc(VD*sizeof(double));
  double *tau_avg = (double *) malloc(VD*sizeof(double));
  xxstate xx_avg, next_xx_avg;

  // get the expectation of the current state
  getExpectedSample(sstate, f_avg, tau_avg, &xx_avg);

  // get the next state based on that
  // assume everything works here , no error checking at all
  updateXXSample(sstate->r, &next_xx_avg, &xx_avg, agent_behaviour, client_behaviour_obs);
  
  sampleX(sstate->r, &next_xx_avg, xx_obs);
  free(f_avg);
  free(tau_avg);
  return 1;
}
int getExpectedSample(samplerState *sstate, double * f_avg, double *tau_avg, xxstate *xx_avg)
{
  double totalweight;
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;


  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  initializeXXstate(xx_avg);
  totalweight=0.0;
  
  // this loops through all the samples, and computes the weighted sum of fundamentals and transients,
  for (int i=0; i<numsamples; i++) {
    for (int j=0; j<VD; j++) {
      f_avg[j] = f_avg[j]+the_samples[i].f[j]*the_samples[i].weight;
      tau_avg[j] = tau_avg[j]+the_samples[i].tau[j]*the_samples[i].weight;
    }
    addToXXstate(xx_avg,&(the_samples[i].xx),the_samples[i].weight);
    totalweight += the_samples[i].weight;
  }
  for (int i=0; i<VD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }

  divideXXstate(xx_avg,totalweight);
  
  return 1;
}

// calculates the expected distance of a VD-length vector tau_avg (e.g. of transient impressions from Interact) from the current sample set transients
// result is  put into VD-length vector dist
// computes distance to tau if dotau neq 0, otherwise does distance to f
int calcExpectedDistanceToTransient(samplerState *sstate, double * tau_avg, double *dist, int dotau)
{
  double totalweight;
  sample * the_samples = sstate->samples;
  int numsamples = sstate->numsamples;

  for (int i=0; i<VD; i++) 
    dist[i] = 0.0;

  totalweight=0.0;
    // this loops through all the samples, and computes the weighted distance from tau_avg to tranients
  for (int i=0; i<numsamples; i++) {
    for (int j=0; j<VD; j++)
      if (dotau) 
	dist[j] += (tau_avg[j]-the_samples[i].tau[j])*(tau_avg[j]-the_samples[i].tau[j])*the_samples[i].weight;
      else
	dist[j] += (tau_avg[j]-the_samples[i].f[j])*(tau_avg[j]-the_samples[i].f[j])*the_samples[i].weight;
    totalweight += the_samples[i].weight;
  }
  for (int i=0; i<VD; i++) 
    dist[i] = sqrt(dist[i]/totalweight);
  return 1;
}

void printDeflections(samplerState * sstate, AEMeth ae_meth) {
  printDeflectionsCSV(sstate, ae_meth,NULL,NULL,"");
}

// print out all deflections, both expected deflections
// and deflections from the average state <-- this will be closer to what Interact will compute
void printDeflectionsCSV(samplerState * sstate, AEMeth ae_meth, FILE *fpout, char *csv_header_line, char *prefix) {
  double agdefl,behdefl,cldefl;
  int output_csv=0;
  char tmp_header_line[1024];
  if (fpout)
    output_csv=1;
  if (ae_meth==AVERAGE) {
    agdefl = computeDeflectionOfAverage(sstate, AGENT);
    behdefl = computeDeflectionOfAverage(sstate,BEHAVIOUR);
    cldefl = computeDeflectionOfAverage(sstate,CLIENT);
#ifdef VERBOSE
    printf("=====================================================================\n");
    printf("              AVERAGED method\n");
    printf("deflection of average sentiments AGENT: %4.3g\n",agdefl);
    printf("deflection of average sentiments BEHAVIOUR: %4.3g\n",behdefl);
    printf("deflection of average sentiments CLIENT: %4.3g\n",cldefl);
    printf("total deflection: %4.3g\n",agdefl+behdefl+cldefl);
#endif
    if (output_csv) {
      fprintf(fpout,"%4.3g,%4.3g,%4.3g,%4.3g,",agdefl,behdefl,cldefl,agdefl+behdefl+cldefl);
      sprintf(tmp_header_line,"%s: deflection agent, %s: deflection behaviour, %s: deflection client, %s: deflection total,",prefix,prefix,prefix,prefix);
      strcat(csv_header_line,tmp_header_line);
    }
  }
  if (ae_meth==EXPECTATION) {
    agdefl = computeDeflection(sstate, AGENT);
    behdefl= computeDeflection(sstate, BEHAVIOUR);
    cldefl = computeDeflection(sstate, CLIENT);
#ifdef VERBOSE
    printf("=====================================================================\n");
    printf("             EXPECTATION method\n");
    printf("deflection AGENT: %4.3g\n",agdefl);
    printf("deflection BEHAVIOUR: %4.3g\n",behdefl);
    printf("deflection CLIENT: %4.3g\n",cldefl);
    printf("total deflection: %4.3g\n",agdefl+behdefl+cldefl);
#endif
    if (output_csv) {
      fprintf(fpout,"%4.3g,%4.3g,%4.3g,%4.3g,",agdefl,behdefl,cldefl,agdefl+behdefl+cldefl);
      sprintf(tmp_header_line,"%s: deflection agent, %s: deflection behaviour,  %s: deflection client, %s: deflection total,",prefix,prefix, prefix,prefix);
      strcat(csv_header_line,tmp_header_line);
    }
  }
  agdefl = computeDeflectionVariance(sstate, AGENT);
  behdefl= computeDeflectionVariance(sstate, BEHAVIOUR);
  cldefl = computeDeflectionVariance(sstate, CLIENT);
#ifdef VERBOSE
  printf("=====================================================================\n");
  printf("             VARIANCES\n");
  printf("deflection variance AGENT: %4.3g\n",agdefl);
  printf("deflection variance BEHAVIOUR: %4.3g\n",behdefl);
  printf("deflection variance CLIENT: %4.3g\n",cldefl);
  printf("total deflection variance: %4.3g\n",agdefl+behdefl+cldefl);
#endif
  if (output_csv) {
    fprintf(fpout,"%4.3g,%4.3g,%4.3g,%4.3g,",agdefl,behdefl,cldefl,agdefl+behdefl+cldefl);
    sprintf(tmp_header_line,"%s: deflection variance agent, %s: deflection variance behaviour, %s: deflection variance client, %s: deflection variance total,",prefix,prefix,prefix,prefix);
    strcat(csv_header_line,tmp_header_line);
  }
  return;
}
