/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.3.8)                              #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@bayesact.ca   bayesact.ca                                           #
#  Date  : April 2021                                                                            #
#  Copyright (C) 2017-2021  Jesse Hoey                                                           #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
 * FVarSampler code - 
 *  Using GSL -> www.gnu.org/software/gsl
#************************************************************************************************/
#ifndef __FVARSAMPLE
#define __FVARSAMPLE
#define GSL_RANGE_CHECK_OFF
#define GSL_VERSION_2
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
// GNU scientific library
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>


// some structs defined later on used in APP files
typedef struct dyadType dyad;
typedef struct agState agentState;
typedef struct sentimentParameters sentimentParams;
typedef struct moduleState samplerState;

// **************************************************************************************************************
// INCLUDES 
// **************************************************************************************************************
// application specific state defined in a file
// CHANGE this to create a new application
// also change in Makefile
// if using institutionalAPP or bayesactAPP (who do some institutionalfiltering)
// need to set DO_DENOTATIVE_IDENTITY_UPDATES to 1 in fvarsample.h as well
// may also need to change DEFLECTION_WEIGHT in forvalue.h
// Pick one of the following:
// defaultAPP expects DO_DENOTATIVE_IDENTITY_UPDATES to be 0 but will function (I think) if it is 1
/*
#include "pdbayesactAPP.h"
#include "bayesactAPP.h"
#include "institutionalAPP.h"
#include "defaultAPP.h"
#include "pdAPP.h"
#include "bayesactAPP.h"
#include "ultimatumAPP.h"
*/
#include "bayesactAPP.h"

// functions to read a csv file
#include "tinycsv.h"

#include "mvnorm.h"

// **************************************************************************************************************
// CONSTANTS
// **************************************************************************************************************

// prints out additional information - little used
// must be on though if we are doing USER simulation...
#define VERBOSE
// a higher level of verbosity
//#define VERYVERBOSE


// max and min functions
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

// maximum number of agents that can be initialized  - is machine dependent 
#define MAX_NUM_AGENTS 10

// maximum number of interactions (dyads) that can be initialized
#define MAX_NUM_INTERACTIONS 10*MAX_NUM_AGENTS

// maximum number of dictionaries 
#define MAX_NUM_DICTIONARIES 10

// sentiment space dimensionality. This is normally 3 for E,P,A
#define SSD 3

// number of terms in the connotative grammar. A-B-C : 3.  A-B-C-S: 4
#define NGT 3

// number of terms in the denotative grammar. A-B-C : 3.  A-BA-BC-C: 4 ...
#define NGTD 4

// number of entity types (connotative)
#define NET_C 4  

// number of entity types (denotative)
#define NET_D 4 

// number of dimensions of the full vector space
#define VD NGT*SSD

// used in a few places - max number of characters that a label can have
#define MAX_LABEL_LENGTH 256

// number of institution slots
#define NUM_INSTITUTIONS 14
static char INSTITUTION_LIST[NUM_INSTITUTIONS][10] = {"overt","surmised","lay","business","law","politics","academe","medicine","religion","family","sexual","monadic","group","corporal"};

// mappings for agent and client turn
static int __agentmap[9] = {0,1,2,3,4,5,6,7,8};
static int __clientmap[9] = {6,7,8,3,4,5,0,1,2};

// the weight reset threshold - samples will be re-identified if they have a weight less than this
// THIS NUMBER IS IMPORTANT!! If a samples weight falls below this value, it will be re-identified
// 1e-69 is chosen fairly arbitririly as smaller than a single precision minimum of 10-38, but bigger
// than the double-precision xbminimum of 10^-308.
// making this number BIGGER (e.g. 1e-09) would mean more resetting
// making this number SMALLER (e.g. 1e-300) would mean less resetting
// but I have no idea what effect this has over its range.
// The OUTPUT_ALL_WEIGHTS flag will cause ALL_WEIGHTS_SAVE_FILE to be written which shows you
// all the weights for an iteration. This may allow calibration of this threshold to some degree.
// define NEVER_RESET if you want to gaurantee no resets, and also set this to 0.0
#define RESET_THRESHOLD 1e-32

//#define ALWAYS_RESET_BOTH
//#define NEVER_RESET


// whether to keep track of an interact state for each dyad
// this must be off if any actor in any dyad starts unaware of the identity of the other actor in the same dyad
// another way to handle this is to have a "nobody" identity that can be used in place of nothing
// (really its more like "everybody" so this could be the mean of the identity files)
// if DO_INTERACT is off then resets are done randomly across -4.3:4.3
#define DO_INTERACT 

// define this if you want a file written out every iteration called all_weights_save.csv which is your csv of the current set of weights for each sample
// this can be useful for figuring out what RESET_THRESHOLD should be. Basically on a disruptive action, you want RESET_THRESHOLD to be higher than the
// sum of all the weights, so look for roughly the mean of this distribution for a disuptive event (one requiring re-identification)x
//#define OUTPUT_ALL_WEIGHTS
#define ALL_WEIGHTS_SAVE_FILE "../output/all_weights_save.txt"

// the variance on the EPA means for dictionaries that don't include it
// THIS NUMBER IS ALSO IMPORTANT (if using MEAN dictionaries)
// this is \gamma value by default
// can override this on the command line *globally* with -ga -- if dictionaries are of type MEAN
// TODO: make it so gamma is can be also specified per agent to override MEAN dictionaries
// can also override it per agent by changing their dictionaries
// this is gamma 
#define DTYPE_MEAN_VARIANCE 0.1

// can override this in the sim file
#define DEFAULT_ALPHA 0.1
#define DEFAULT_BETA 0.01
#define DEFAULT_AGENT_DELTA 0.1   // This is in bayesactAPP.c 
#define DEFAULT_NUMSAMPLES 1000
// others (flags) are in inputparser.h

// fraction of samples to draw from the persona, if defined (in simulation specification file)
#define PERSONA_FRACTION 0.2


// if defined, will do denotative updates of X_a and and X_c (must be defined in updateXacSample in <xx>APP.c
#define DO_DENOTATIVE_IDENTITY_UPDATES 1

// if defined, will use inertia when sampling from fvar (I think this is actually necessary)
#define USING_INERTIA 1

// select action switch for getNextAction - set to 0 if not doing this
// made obsolete March 2021 - basically this was a *global* flag to indicate whether you wanted to get
// a distribution for the optimal action to take out of getNextAction, or get a distribution aroudn the action
// that was actually taken, but in the end you just call getNextAction with a behaviour specified or not
//#define SELECT_ACTION 0

// agent weights its beliefs with its own/ it's client's displayed emotion?
// normally this is off/on
// has no effect if use_emotions flag is set to off in e.g. bayesactsim.c
// or if agent_emotion/client_emotion is NULL when calling update
#define WEIGHT_CLIENT_EMOTION 1

// agent weights its beliefs with its own displayed emotion?
// normally this is on
// has no effect if use_emotions flag is set to off in e.g. bayesactsim.c
// or if client_emotion is NULL when calling update
#define WEIGHT_AGENT_EMOTION 0

// weight denotative emotions by distance from label
// when computing an emotion, we can weight each sample by the distance to that emotion
// so the emotional states that are more distant have smaller weight
// otherwise, they are all weighted equally
#define WEIGHT_EMOTIONS_BY_DISTANCE 1

// used to ensure the sampling process doesn't break if no samples are weighty enough.
#define MIN_SAMPLE_TOTAL_WEIGHT 1e-300


// **************************************************************************************************************
// STRUCTURES
// **************************************************************************************************************
// defines the set of grammar terms - for vanilla ACT this is A-B-O
// but if NGT=4 it could be A-B-O-S, or
// in general A-B-O-X where X is any other state
// the connotative entities 
// should be NET_C of these
typedef enum {
  AGENT=0,
  BEHAVIOUR,
  CLIENT,    // would add SETTING, here if using settings
  EMOTION,   // used for dynamics as well
} c_entity_type;

// the denotative entities 
// should be NET_D of these
typedef enum {
  ID_AGENT=0,
  ID_CLIENT,
  BEH_AGENT,
  BEH_CLIENT,
} d_entity_type;



// different types resets
typedef enum  {
  RESET_NONE=0,
  RESET_AGENT=1,
  RESET_CLIENT=2,
  RESET_BOTH=3
} reset_type;

// different types of dictionaries.
// MEAN: no variance information - as packaged with interact e.g. fbehaviours.dat
// SD: variances only  - each line is a EPA mean followed by EPA variances (not Standard deviations!)
// COV: full covariance - each line is a EPA mean followed by flattened EPA covariance
// SAMPLE: distribution specified as a set of raw EPA samples  (not implemented yet)
typedef enum  {
  MEAN=0,
  SD,
  COV,
  SAMPLE,
  GAUSS_MIX
} dict_type;

// different types of somatic transforms
// - WEIGHT means both denotative and connotative are specified, and a weight is computed for their congruence
// - CONNOTATIVE means only denotative is specified and the corresponding connotative is what is computed
// - DENOTATIVE means only connotative is specified and the corresponding  denotative is what is computed
typedef enum {
  WEIGHT,
  CONNOTATIVE,
  DENOTATIVE
} SATtype;


// different methods for computing actions and emotions
typedef enum {
  AVERAGE,
  EXPECTATION,
  PLANTREE
} AEMeth;


// different ways of querying the user will be specified
typedef enum  {
  BAYESACT=0,  // not used anymore -- run an events file with all "*"
  USER,
  EVENTS
} query_type;


//structure that defines a connotative sentiment
typedef struct senti sentiment;
struct senti {
  char label[256];   // the label 
  gsl_vector * epa;   // epa vector - the mean
  dict_type dtype;
  // overt surmised ... lay business law politics academe medicine religion family sexual ... monadic group corporal
  // could be shortened to 1 bit per institution
  int the_institutions[NUM_INSTITUTIONS];
  
  double var;                // for dtype: NONE
  gsl_vector * variance;     // for dtype: SD
  gsl_matrix * covariance;    // for dtype: COV
  gsl_matrix * cholesky;
};

// the structure that defines an denotative sentiment  (more like a state really)
// has fields for:
// -a string label
// -an index to a dictionary
// -a probability of this identity
// so if held in an array, this is a denotative identity set
typedef struct denState denotativeState;
struct denState {
  char label[MAX_LABEL_LENGTH];
  int index;
  double prob;
};

// some aliases for denotative sentiments
typedef denotativeState identity;
typedef denotativeState behaviour;
typedef denotativeState emotion;

// a dictionary maps labels to distributions over sentiments
// - num_items : number of entries in the dictionary (e.g. identity labels)
// - dtype: the type of dictionary is how the distribution over each identity is specified
// - fname: a file name for reading this dictionary from
// - theDictionary: an array of sentiments, the actual dictinoary
typedef struct dictionaryStruct dictionary;
struct dictionaryStruct {
  int num_items;
  double gamma;
  dict_type dtype;
  c_entity_type dict_what;
  char fname[MAX_LABEL_LENGTH];
  sentiment * theDictionary;
  // a set of converters for institutional filters applied to this this dictionary
  int *num_ic_items;
  int **list_ic_converter;
};

// structure that defines an agent
struct agState {
  //samplerState *interactionStates;
  char name[MAX_LABEL_LENGTH];
  char dyn_fname[MAX_LABEL_LENGTH];
  char mod_fname[MAX_LABEL_LENGTH];

  // an array of dictionaries that corresponds to entity_types
  dictionary * entity_dicts[NET_C];

  // parameters for this agent
  int numsamples;
  double alphaa, alphab, alphac;
  double betaa, betac;
  double deltaa, deltac;
  int do_resets;
  int default_action;
  int emotion_planning;
  int use_emotions;


  // the set of persona ids
  identity *persona_ids;
  int num_persona_ids;
  // arrays of type names and file names for those dictionaries.
  // this information is also replicted in the dictionary as dtype and fname
  char entity_dict_types[NET_C][MAX_LABEL_LENGTH];
  char entity_dicts_fname[NET_C][MAX_LABEL_LENGTH];
};


//structure that defines a dyad (interaction)
struct dyadType {
  // index of the dyad
  int dyad_index;

  // index of symmetric dyad (can be -1 if it doesn't exist)
  int mirror_dyad_index;

  samplerState *dsstate;
  // index into the array of agents of the agent in this dyad
  int agent_index;
  // index into the array of agents of the client in this dyad
  int client_index;
  // number of agent and client ids for this dyad
  // this is for persona 
  int num_agent_ids, num_client_ids;
  // the identity sets for each agent and client in this dyad
  identity *agent_ids, *client_ids;

  unsigned long random_seed;
  
  // the interact actions to take in this dyad the next time it is used
  // first index is for AGENT/CLIENT
  double interact_beh[2][SSD];

  // final thing a dyad may have is an institution list
  int institution_list[NUM_INSTITUTIONS];
};

// the structure that defines a sample
// has fields for
// *two connotative variables:
//  while these could be of type "sentiment", they are double arrays to save space
//   -fundamental sentiments f 
//   -transient impressions tau
// *NGTD denotative variables. While these could be of type "denotativeState",
//  they are ints to save space. with NGTD there are 
//   -ID_AGENT: self identity x_a                | 
//   -ID_CLIENT: other identity x_c              |- all these are indices into some dictionary.  
//   -BEH_AGENT: behaviour x_ba for agent        |
//   -BEH_CLIENT: behaviour x_bc for client      |
// *weight -- the weight of the sample
// ******
// NOTE: sample has a statically defined size (based on VD and NGTD) on purpose
//       because I copy samples using (shallow) memcpy
typedef struct samp sample;
struct samp {
  //fundamental sentiment
  double f[VD];
  double tau[VD];
  // denotative identities and behaviours
  int x[NGTD];

  //XX: additional state goes here -
  // xxstate type is defined in *APP.h files - application specific for now
  xxstate xx;

  // to save who actually acted for this sample
  // could be client or agent. This causes unbelievable confusion in the code
  // and really I should have just had two fields for behaviour sentiments,
  // one for agent and one for client. 
  d_entity_type who_acted;

  // weight for the sample
  double weight;

  // a flag to set for resets
  int reset_flag;
};



  
// the structure for all the necessary parameters for an agent
struct sentimentParameters {

  //flag to indicate if this is currently allocated
  //allows us to have multiple (but not too many) agents allocated at once
  int is_allocated;
  
  // matrices to hold the covariances 
  gsl_matrix * sig_a, * isig_a;
  gsl_matrix * sig_f, * isig_f;

  gsl_matrix * sig_a_reset, * isig_a_reset;
  gsl_matrix * sig_f_reset_agent, * isig_f_reset_agent;
  gsl_matrix * sig_f_reset_client, * isig_f_reset_client;
  gsl_matrix * sig_f_reset_both, * isig_f_reset_both;

  // dynamics matrices - curly K, curly H, and curly C in BayesACT paper
  gsl_matrix *curlyK;
  gsl_matrix *curlyKinv;
  gsl_matrix *curlyH;
  gsl_matrix *curlyKa;
  gsl_matrix *curlyKainv;
  gsl_matrix *curlyHa;
  gsl_matrix *curlyKc;
  gsl_matrix *curlyKcinv;
  gsl_matrix *curlyHc;
  gsl_vector *curlyC;
  gsl_vector *curlyCa;
  gsl_vector *curlyCc;


  // number of features in the M matrix for impression formation
  int num_features;
  // arrays of indices for the impression formation features
  int ** m_index;
  int * m_num_features;
  // the dynamics matrix M
  double * M;

  // number of features in the modifiers dynamics equations
  int num_features_modifiers;
  // arrays of indices for the impression formation features
  int ** dreq_index;
  int * dreq_num_features;
  // the modifier dynamics matrix d,R,E and Q
  double * dREQ;

  // the mean and covariance of the posterior distribution over f
  gsl_vector * mu_n;
  gsl_matrix * sig_n;

  // the mean as computed by interact
  gsl_vector * interact_mean;
  // whether it was possible (or desired) to compute the interact mean
  // could do it this way if we want different dyads to be able to use/not-use interact resets
  // for example, the ability to specify an agent who does not know the identity of another agent
  // precludes using interact, so the DOINTERACT flag must be off
  //int do_interact;

};

// the structure allowing us to have multiple interactions, 
// has an array of sentimentParams structures, plus
// the current number of agents.
// this was written to capture everything we need to access the module state from Python
struct moduleState {
  // the number of sentiParams objects we will have
  int current_num_interactions;
  sentimentParams *sentiParams;

  sample * samples;
  // alpha - strengths of the affect control principle for agent, behaviour and client
  double alphaa;
  double alphab;
  double alphac;
  // beta for agent and client - inertial strength
  double betaa;
  double betac;
  // number of samples to use for the particle filter
  // more means more precise estimates
  int numsamples;

  double roughenNoise;

  // will do resets
  int do_resets;

  // count the number of resets
  int num_resets;
  int num_resets_successful;
  int num_resets_agent;
  int num_resets_client;
  int num_resets_both;
  int num_hard_resets;
  // will return the default action only (most aligned action)
  // when getNextAction is called
  int default_action;

  // whether to use emotions at all
  int use_emotions;

  // whether to plan over emotions
  int emotion_planning;

  // random number generator
  const gsl_rng_type * T;
  gsl_rng * r;
  // the random number seed
  unsigned long int random_seed;
};


// an event between two members of a dyad
// each event consists of a behaviour and an emotion for each member of the dyad dyad_index
// dyad_index must be >= 0 and indicate the dyad to update for this event
// behaviour_index=-1 means "none"
// behaviour_index=-2 means "default" and will be selected at simluation run time
// emotion_index=-1 means "no emotion" (must exist in the dictinary
// emotion_index=-2 means "default" and will be selected at simluation run time
typedef struct eventType event;
struct eventType {
  int dyad_index;
  int agent_index;
  int client_index;
  // index is for AGENT/CLIENT
  int behaviour_index[2];
  int emotion_index[2];
  int disrupt_action[2];
  char agent_x_obs[MAX_LABEL_LENGTH];
  char client_x_obs[MAX_LABEL_LENGTH];
};
// **************************************************************************************************************
// FUNCTIONS
// **************************************************************************************************************

// helper functions
double euclidean_distance(double *epa1, double *epa2, int nd);
double cosine_distance(double *epa1, double *epa2, int nd);
int lookUpLabel(char * slabel, dictionary * the_dictionary);
int lookUp(denotativeState * s, dictionary * dict);
int lookUpActor(char *label);
int lookUpEntity(char *label);
int lookUpDictType(char *label);
int lookUpGeneric(char *label, int num_labels, char **label_list);

int insert_into_state_list(denotativeState ** id_list, int ** count_list, int id_list_length, int index, double weight);

int denotativeStateCompare( const void *s1, const void * s2);


// pretty-printing
void printMatrix(gsl_matrix * m, char * name, int sizex, int sizey);
void printVector(gsl_vector * m, char * name, int sizex);
void printRawSample(sample * the_sample);
void printRawSampleCSV(sample * the_sample,FILE *fpout, char*, char *);
int printRawVariancesCSV(samplerState * sstate, agentState * the_agent, agentState *the_client, FILE *fpout, char *, char *);
void printRawSamplesCSV(samplerState * sstate, agentState * the_agent, agentState *the_client, FILE *fpout, char *, int max_num, char *);
int printSamples(dyad *, agentState *agent);
int printSamplesCSV(dyad *, agentState *agent, agentState *client, FILE *fpout, char *csv_header_line, char *prefix);
void printDeflections(samplerState * sstate,  AEMeth);
void  printDeflectionsCSV(samplerState * sstate, AEMeth, FILE*, char *, char *prefix);
void printEmotionsCSV(FILE *fpout, int iteration, char *csv_header_line, double *emot, char *bestemotlabel,char *actorname, char * clientname);

// main functions for sampling
void roughenSample(samplerState *sstate, sample samp, double roughenNoise);
void roughenAllSamples(samplerState *sstate, double roughenNoise);
void reSample(sample * samples, sample * new_samples, int numsamples);
void reSampleInPlace(sample * samples, int numsamples);
int drawSample(gsl_rng *r, sample * samples, int numsamples);
int setUpDyadicInteraction(dyad *, agentState *);

denotativeState * drawRandomProb(gsl_rng *, denotativeState * id);
denotativeState * drawWeightedProb(gsl_rng *, denotativeState * id,double weight);
int drawRandomProbDouble(gsl_rng *r, double *prob);
int getExpectedSample(samplerState *sstate, double * f_avg, double *tau_avg, xxstate *xx_avg);
int calcExpectedDistanceToTransient(samplerState *sstate, double * tau_avg, double *dist, int dotau);
int sampleNextXObservation(samplerState *sstate, int agent_behaviour, int client_behaviour_obs, xostate * xx_obs);


void getSampleSet(dyad *, identity *, int, identity *, int, agentState *);
double updateFullState(dyad *the_dyad, agentState *the_agent, behaviour *, identity *);
double updateFullStateWithXXObs(dyad *the_dyad, agentState *the_agent, behaviour *, identity *, xostate *);
double updateFullStateTwoActors(dyad *the_dyad, agentState *the_agent,
				behaviour * the_agent_behaviour, behaviour * the_client_behaviour_observation);
double updateFullStateTwoActorsWithEmotion(dyad *the_dyad, agentState * the_agent,
					   behaviour * agent_behaviour, behaviour * client_behaviour_observation,
					   emotion * agent_emotion, emotion * client_emotion);
double updateFullStateTwoActorsWithXXObs(dyad *the_dyad, agentState *the_agent,
				behaviour * the_agent_behaviour, behaviour * the_client_behaviour_observation, xostate *xx_observation);
double updateFullStateTwoActorsWithEmotionWithXXObs(dyad *the_dyad, agentState * the_agent,
					   behaviour * agent_behaviour, behaviour * client_behaviour_observation, xostate *xx_observation, 
					   emotion * agent_emotion, emotion * client_emotion);
void updateXSample(gsl_rng *r, dyad *the_dyad, sample *, sample *, agentState *the_agent, behaviour *agent_behaviour, behaviour * client_behaviour_observation);
double weightXSample(gsl_rng *r, sample * next_sample, xostate * xx_observation);
double weightXbcSample(gsl_rng *r, sample * next_sample, behaviour *client_behaviour_observation);

double sampleWeightNextSentimentWithEmotion(dyad *the_dyad, reset_type rtype, agentState *the_agent, d_entity_type *the_actor,
					    sample * current_sample, sample * next_sample,
					    behaviour * agent_behaviour, behaviour * client_behaviour_observation, xostate *xx_observation,
					    emotion * the_agent_emotion, emotion * the_client_emotion);

void sampleNextState(dyad *the_dyad, d_entity_type the_actor, agentState * the_agent, reset_type rtype,
		     sample * current_sample, sample * next_sample, behaviour * agent_behaviour, emotion * the_agent_emotion,
		     behaviour * client_behaviour_observation, emotion * the_client_emotion);
double computeSampleWeight(samplerState *sstate, int dyad_index, agentState * the_agent, 
			   sample * next_sample, xostate * xx_observaiton, 
			   behaviour * client_behaviour_observation,
			   emotion * the_client_emotion, emotion * the_agent_emotion);
void sampleObservation(dyad *,  agentState *the_agent, sample * the_sample,
		       behaviour * the_client_behaviour, emotion * the_client_emotion, xostate *xo);
int sampleNextSentiment(samplerState *sstate, d_entity_type, reset_type, sample *, sample *);
int setUpSampleNextSentimentRaw(samplerState *sstate, d_entity_type, reset_type, double * f, double * tau);
double formImpression(samplerState *sstate, d_entity_type atype, double * f, double * tau, double * tau_out);
void setUpImpressions(samplerState *sstate, d_entity_type atype, double *tau);
int setUpImpressionFormation_bonly(sentimentParams *__agent_params, int * themap, double *tau);
int setUpImpressionFormation(sentimentParams *__agent_params, int * themap, double *tau, int, int);
double doFullInteractUpdate(samplerState *sstate, int, agentState *the_agent, d_entity_type atype, int, int,
			    behaviour *b, double *tau, double *f_out, double *tau_out, double*, double*);
double doFullInteractUpdate_helper(samplerState *sstate, int, d_entity_type atype, double *f,
				   double *behaviour, double *tau, double *tau_out, double *, double *);
int sampleSentiment(samplerState *sstate, int dyad_index, double *f_out, double *tau_out);
int sampleASentiment(samplerState *sstate, double *myfout, double *mytout);

int computeInteractMean(samplerState *sstate, d_entity_type atype, double * f, double * tau);
void printInteractMean(samplerState *sstate);
void printInteractMeanCSV(samplerState *sstate, FILE*, char*);

int getIdentities(samplerState * sstate,
		  int *num_identities_agent, identity **x_c_avg,
		  int *num_identities_client, identity **x_a_avg);

void resampleFromPersona(samplerState *sstate, int dyad_index, agentState *agent, double fraction);

// next action predicitons
int getNextAction(dyad *, agentState *, identity **, d_entity_type);

int getNextActionFromAverageState(dyad *the_dyad, agentState * the_agent, d_entity_type);
void getNextActionListFromAverageState(dyad *the_dyad, agentState *the_agent, identity ** behaviour_list,  d_entity_type);

// emotions;
int computeExpectedEmotion(dyad *the_dyad, double * e_out, c_entity_type atype);
int computeEmotionFromAverage(dyad *the_dyad, double * e_out, c_entity_type atype);
int computeEmotion(samplerState *, double *f, double *tau, double *e_out);


int getNextEmotionList(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list,
		       c_entity_type whose_emotion, d_entity_type who_acted);
int getNextMostLikelyEmotion(dyad* the_dyad,  agentState *the_agent, emotion ** emotion_list,
			     c_entity_type whose_emotion, d_entity_type who_acted);
int getNextEmotions(dyad  *the_dyad,  agentState *the_agent, emotion ** emotion_list,
		    c_entity_type whose_emotion, d_entity_type who_acted);

int getCurrentEmotion(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type atype);
int getCurrentEmotionList(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type atype);
int getCurrentMostLikelyEmotion(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type atype);
int getCurrentEmotions(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type atype);
int sampleCurrentEmotion(dyad *the_dyad,  agentState *the_agent, emotion ** emotion_list, c_entity_type atype);
double getExpectedEmotionMap(dyad *the_dyad,  agentState *the_agent, c_entity_type atype, double *e0, double *e1);

int getNullEmotion(dyad *the_dyad,  agentState *the_agent, int*, double*, c_entity_type atype);

//the somatic transform
double somaticTransform(SATtype sat_type, gsl_rng *r, int dyad_index, int * denotative_index, double * connotative_epa, dictionary * dict);
double computeDeflection(samplerState *sstate, c_entity_type atype);
double computeDeflectionVariance(samplerState *sstate, c_entity_type atype);
double computeDeflectionOfAverage(samplerState *sstate, c_entity_type atype);
double computeDeflectionOfSample(sample * the_sample, c_entity_type atype);

// memory set up and tear down
void initializeBayesACT(dyad *, agentState *) ;
int setupParameters(samplerState * sstate, char * dyn_fname, char * mod_fname);
int freeParameters(dyad * the_dyad);
int freeParametersFromIndex(samplerState * sstate);
void freeBayesACT(samplerState *sstate);
#endif
