#************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
#  pdbot python test class - tests the prisoner's dilemma, a canonical game for BayesACT         #
#*************************************************************************************************

import time
from libbayesact import *

#a PDBot

class PDBot(object):
    def __init__(self,*args,**kwargs):

        self.numsamples=kwargs.get("numsamples",1000)
        self.num_plan_samples=kwargs.get("num_plan_samples",500)
        self.timeout=kwargs.get("timeout",1)
        self.simfile=kwargs.get("simfile","testsim-pd.txt")
        # whether to use planning or not - this has to be set to 1 right now
        self.use_planning=0
        # whether to plan over emotions or not - could be an input argument
        self.emotion_planning=0

        #this index must be used for all future calls to this sampler
        self.index = initializeState(self.numsamples, self.num_plan_samples, self.timeout, self.simfile, self.emotion_planning)
        if self.index < 0:
            raise NameError("ERROR: too many agents initialized")
        print(100*"*")
        print("new agent index "+str(self.index))
        # if multiple dyads are specified in simfile then we can refer to specific ones from now on
        self.the_dyad = 0
        #ADDED BY NEIL BUDNARAIN ------------
        #EPA Values set after each move
        self.epavals = 1
        #------------------------------------

    #not sure if this should be called "__del__" instead  - it seems this is dangerous, so
    #I'm going to put it as a regular member function that needs to be explicitly called
    def freeMe(self):
        freeState(self.index)

    #query the bot for its next (default only) emotion
    def get_emotion(self):
        #self.agent_emotion = getNextMostLikelyEmotion(self.index,self.the_dyad)
        self.agent_emotion = sampleCurrentEmotion(self.index,self.the_dyad)
        return self.agent_emotion

    #query the bot for its next action (play)
    def get_play(self):

        (self.agent_play, self.agent_emotion) = getNextActionEmotion(self.index, self.the_dyad, self.use_planning)
        #ADDED BY NEIL BUDNARAIN--------------
        self.epavals = self.agent_emotion
        #-------------------------------------
        return (self.agent_play,self.agent_emotion)

    #update the bot with the play made by the client - "give 2" or "take 1" (or see action_names)
    #there is no error checking here ... will fail badly if anything other than an action name is given
    def make_play(self,client_play,client_emotion):
        xx_obs = 0
        if (client_play=="give 2"):
            xx_obs += 2
        if (self.agent_play=="take 1"):
            xx_obs += 1
        print(str(xx_obs),client_play,client_emotion,self.agent_play,self.agent_emotion)
        numnodes = updateState(self.index,self.the_dyad,str(xx_obs),client_play,client_emotion,self.agent_play,self.agent_emotion,self.use_planning)
        if numnodes < 0:
            print("failure - emotion label not found - state did not update")

    def get_identities(self):
        (self.agent_id, self.client_id) = getIdentities(self.index,self.the_dyad)
        return (self.agent_id, self.client_id)

    def get_deflections(self):
        return getDeflections(self.index,self.the_dyad);
    def get_random_seed(self):
        return getRandomSeed(self.index,self.the_dyad);

    def get_state(self):
        return getState(self.index,self.the_dyad);
    def get_extended_state(self):
        return getExtendedState(self.index,self.the_dyad);

def doSimulation(pd_agent):
    done = False
    start = time.clock()
    iter=0
    while not done:
        ## get default emotion to display at start
        (agent_emotion) = pd_agent.get_emotion()
        print("pd-bot is feeling "+agent_emotion+" and is now thinking ...")
        (agent_action,agent_emotion) = pd_agent.get_play()
        
        client_action = input("your action (give 2 or take 1, anything else stops play): ")


        if client_action == "give 2" or client_action == "take 1":
            client_emotion = input("your emotion (label from dictionary): ")
            print(100*"*")
            print("pd-bot's action is to: ",agent_action)
            print("pd-bot's emotion is: ",agent_emotion)
            pd_agent.make_play(client_action,client_emotion)
        else:
            done = True
        (agent_id,client_id) = pd_agent.get_identities()
        print("agent thinks it is a  :  "+agent_id)
        print("and thinks client is a: "+client_id)
        # test get_state method
        stateDictionary=pd_agent.get_state()
        print("deflection of agent: "+str(stateDictionary["agentdeflection"]));
        print("deflection of behaviour: "+str(stateDictionary["behaviourdeflection"]))
        print("deflection of client: "+str(stateDictionary["clientdeflection"]))
        print("moving on to next game...")
    elapsed = time.clock()
    elapsed = elapsed - start
    print("Time spent is: "+str(elapsed))
    

#test section
if __name__ == "__main__":

    #tests with individual agents ------------------------------------------
    pd_agent = PDBot(simfile="testsim-pd2a.txt",numsamples=1000)   #, num_plan_samples=500, timeout=4)

    doSimulation(pd_agent);

    # now let's free that guy 
    pd_agent.freeMe()

    #-----------------tests with allocating multiple samplers ------------------------------------------
    i=0
    pd_agents={}
    try:
        while i<20:  # max 10 samplers allowed
            print("initializing sampler "+str(i))
            pd_agents[i] = PDBot(simfile="testsim-pd2.txt",numsamples=500, num_plan_samples=100, timeout=2)
            i = i+1
    except NameError:
        print("whoops!")
        pass
    while (i>5):
        i = i-1
        #free the last few agents and try again
        print("freeing agent "+str(i))
        pd_agents[i].freeMe()
        
    #try re-allocating up to the next to last because then we're going to allocate one more to play with
    while (i<9):
        print(100*"%")
        print("initializing sampler "+str(i))
        pd_agents[i] = PDBot(simfile="testsim-pd2.txt",numsamples=500, num_plan_samples=100, timeout=2)
        i = i+1


    #tests with individual agents ------------------------------------------
    pd_agent = PDBot(simfile="testsim-pd.txt",numsamples=1000, num_plan_samples=500, timeout=4)

    doSimulation(pd_agent);


    # now let's free that guy 
    pd_agent.freeMe()

    #and free all the ones we allocated before
    i=0
    while (i<10):
        pd_agents[i].freeMe()
        i = i + 1

    # and add another back in
    pd_agent = PDBot(simfile="testsim-pd2.txt",numsamples=500, num_plan_samples=100, timeout=2)
    
    doSimulation(pd_agent);

    # add a second agent
    pd_agent2 = PDBot(simfile="testsim-pd.txt",numsamples=1000, num_plan_samples=500, timeout=4)

    # run a simulation with both agents at once
    done = False
    start = time.clock()
    iter=0
    while not done:
        print("AGENT 1: pd-bot is thinking ...")
        (agent_action,agent_emotion) = pd_agent.get_play()
        (agent2_action,agent2_emotion) = pd_agent2.get_play()
        
        client_action = input("AGENT 1: your action (give 2 or take 1, anything else stops play): ")


        if client_action == "give 2" or client_action == "take 1":
            client_emotion = input("AGENT 1: your emotion (label from dictionary): ")
            print(100*"*")
            print("AGENT 1: pd-bot's action is to: ",agent_action)
            print("AGENT 1: pd-bot's emotion is: ",agent_emotion)
            pd_agent.make_play(client_action,client_emotion)
        else:
            done = True

        client2_action = input("AGENT 2: your action (give 2 or take 1, anything else stops play): ")
                    

        if client2_action == "give 2" or client2_action == "take 1":
            client2_emotion = input("AGENT 2: your emotion (label from dictionary): ")
            print(100*"*")
            print("AGENT 2: pd-bot's action is to: ",agent2_action)
            print("AGENT 2: pd-bot's emotion is: ",agent2_emotion)
            pd_agent2.make_play(client2_action,client2_emotion)
        else:
            done = True
    print("moving on to next game...")
    elapsed = time.clock()
    elapsed = elapsed - start
    print("Time spent is: "+str(elapsed))
    

