/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
# Python Wrapper functions - this allows the functions defined 
# "initializeState", "freeState", "updateState",  and "getNextActionEmotion"
# and others as added to the array libbayesact_methods[] (see below)
# to be used by a python3 program that imports then from the module libbayesact 
#************************************************************************************************/

#include <Python.h>
#include "inputparser.h"
#include "fvarsample.h"

// functions to do planning
#include "forvalue.h"


// maximum number of agents that can be initialized 
#define MAXNUMSAMPLERS 100

#define SAMPLE_EMOTIONS

// the wrapper for all sampler state info needed
// each allocated structure of this type represents
// a complete set of dyads in an group of agents
typedef struct moduleStateWrapper samplerStateWrapper;
struct moduleStateWrapper {
  PyObject *error;
  treeNode *planTree;
  int num_plan_samples;
  int numsamples;
  int timeout;
  dyad * dyads;
  int num_dyads;
  agentState *agents;
  int num_agents;
  dictionary * dictionaries;
  int num_dictionaries;
};

// **********************************************************
// the set of sampler states - this is the main module state
// basically any Python program including libbayesact will get
// allocated (in PyInit_libbayesact) memory accessible by the
// C-code that is wrapped in this structure. The functions
// in this Python wrapper code then access that memory and
// manipulate it. In our case, its an array of samplerStateWrapper
// objects. Each one of these will be defined by a 'testsim.txt' type
// file containing a set of dyads that can interact. This is
// captured in a samplerStateWrapper structure that is referenced
// using an index into the array samplerStates
// ***********************************************************
typedef struct setOfSamplerStates samplerStateSet;
struct setOfSamplerStates {
  PyObject *error;
  // the number of sentiParams objects we will have
  // is always 1 now so this could really be removed
  int __CURRENT_NUM_SAMPLERS;
  samplerStateWrapper *samplerStates;
};


// convenience function to get the module state
#define GETSTATE(m) ((samplerStateSet *)PyModule_GetState(m))


// frees a sampler state - takes a single argument: the index of the sampler state to free
// returns 0
static PyObject* py_freeState(PyObject* self, PyObject* args)
{

  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 

  int dyad_index;
  if (!PyArg_ParseTuple(args, "i", &dyad_index)) 
    return NULL;
  dyad *the_dyad = sstate_wrap->dyads+dyad_index;
  //get the module state
  //samplerState for this sampler
  samplerState * sstate = the_dyad->dsstate;
  
  //printf("freeing %d dictionaries\n",sstate_wrap->num_dictionaries);
  for (int j=0; j<sstate_wrap->num_dictionaries; j++) 
    freeSentimentDictionary(sstate_wrap->dictionaries+j);
  free(sstate_wrap->dictionaries);
  //printf("freeing %d dyads\n",sstate_wrap->num_dyads);
  freeBayesACT(sstate);
  freeParameters(the_dyad);
  //printf("freeing agents\n");
  free(sstate_wrap->agents);
  //printf("freeing plan tree\n");
  freeTree(sstate_wrap->planTree,NULL);
  //printf("freeing bayesACT\n");
  
  return Py_BuildValue("i",0);
}
static PyObject* py_setTimeout(PyObject* self, PyObject* args)
{
  int timeout;
  
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  
  if (!PyArg_ParseTuple(args, "i", &timeout))
    return NULL;
  //get the module state
  //samplerState for this sampler
  sstate_wrap->timeout = timeout;
  return Py_BuildValue("i",0);
}
// initialize a sampler state 
// returns the index of the sampler created, or -1 if all MAXNUMSAMPLERS samplers are allocated
static PyObject* py_initializeState(PyObject* self, PyObject* args)
{
  int  num_plan_samples, timeout;
  char *simfile;

  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 

  // now create the new sampler based on the input arguments
  if (!PyArg_ParseTuple(args,"iis", &num_plan_samples, &timeout, &simfile))
    return NULL;
  //set the module state
  sstate_wrap->num_plan_samples = num_plan_samples;
  sstate_wrap->timeout = timeout;
  // planning stuff
  // the plan tree
  sstate_wrap->planTree=NULL;

  // ================================================================================================
  // basic parameters  gamma override
  // ================================================================================================
  double gamma=-1.0;
  
  printf("reading from %s\n", simfile);
  printf("number of plan samples %d\n",num_plan_samples);
  // ================================================================================================
  // agents and dictionaries 
  // ================================================================================================
  sstate_wrap->agents = (agentState *) malloc(MAX_NUM_AGENTS*sizeof(agentState));
  sstate_wrap->dictionaries = (dictionary *) malloc(MAX_NUM_DICTIONARIES*sizeof(dictionary));
  sstate_wrap->num_dictionaries=0;
  
  // ================================================================================================
  // read the simulation input file - this initializes everything
  // including all agents, dictionaries, and interactions
  // ================================================================================================
  sstate_wrap->dyads = (dyad *) malloc(MAX_NUM_INTERACTIONS*sizeof(dyad));
  // these are ignored
  query_type query_user;
  int max_iterations;
  event *events = NULL;
  printf("setting up agents and dyads\n");
  // this function does all the work - reads in the simulation setup, initializes all dictionaries and dyads
  // dyads have indices agent_index and client_index refeingr to agents in the list agents
  setUpAgentsDyads(simfile, sstate_wrap->agents, &(sstate_wrap->num_agents),
		   sstate_wrap->dyads, &(sstate_wrap->num_dyads), &query_user, &max_iterations, gamma,
		   sstate_wrap->dictionaries, &(sstate_wrap->num_dictionaries), &events);

  
  return Py_BuildValue("i", 0);
}

// This function takes two arguments: sampler state index and dyad index, 
// and computes the next (default) emotion to take for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns a string for the emotion label
static PyObject* py_getNextMostLikelyEmotion(PyObject* self, PyObject* args)
{
  int best_emot;
  int the_dyad=0;
  emotion * emotion_list;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;

  agentState *the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;

  getNextMostLikelyEmotion(sstate_wrap->dyads+the_dyad, the_agent, &emotion_list, AGENT, BEH_AGENT);
  best_emot = emotion_list[0].index;
  free(emotion_list);
  // here is how we access index and string values for action and emotion and return it
  return Py_BuildValue("s",the_agent->entity_dicts[EMOTION]->theDictionary[best_emot].label);
}

// This function takes two arguments: sampler state index and dyad index, 
// and returns a sample from the distribution over next emotions for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns a string for the emotion label
static PyObject* py_sampleCurrentEmotion(PyObject* self, PyObject* args)
{
  int best_emot;
  int the_dyad=0;
  emotion * emotion_list;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;

  agentState *the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;

  sampleCurrentEmotion(sstate_wrap->dyads+the_dyad, the_agent, &emotion_list, AGENT);
  best_emot = emotion_list[0].index;
  free(emotion_list);
  // here is how we access index and string values for action and emotion and return it
  return Py_BuildValue("s",the_agent->entity_dicts[EMOTION]->theDictionary[best_emot].label);
}

// This function takes two arguments: sampler state index and dyad index, 
// and computes the most likely current emotion for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns a string for the emotion label
static PyObject* py_getCurrentMostLikelyEmotion(PyObject* self, PyObject* args)
{
  int best_emot;
  int the_dyad=0;
  emotion * emotion_list;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;

  agentState *the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;

  getCurrentMostLikelyEmotion(sstate_wrap->dyads+the_dyad, the_agent, &emotion_list, AGENT);
  best_emot = emotion_list[0].index;
  free(emotion_list);
  // here is how we access index and string values for action and emotion and return it
  return Py_BuildValue("s",the_agent->entity_dicts[EMOTION]->theDictionary[best_emot].label);
}

// This function takes two arguments: sampler state index and dyad index, 
// and computes the next (default) action to take for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns a string for the behaviour label
static PyObject * py_getNextMostLikelyAction(PyObject * self, PyObject * args)
{
  int best_act;
  int the_dyad=0;
  behaviour * action_list;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;

  agentState *the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;

  getNextAction(sstate_wrap->dyads+the_dyad, the_agent, &action_list, BEH_AGENT);
  best_act = action_list[0].index;
  free(action_list);
  // here is how we access index and string values for action and emotion and return it
  return Py_BuildValue("s",the_agent->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label);
}

// This function takes three arguments: sampler state index and dyad index, as well as a flag to indicate
// whether emotions should be used during the planning process (e.g. plan over emotions to be more manipulative)
// and computes the best next action and emotion to take for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns a Python tuple of strings (action, emotion)
static PyObject* py_getNextActionEmotion(PyObject* self, PyObject* args)
{
  int best_act, best_emot;
  int the_dyad=0;
  int use_planning;


  // for convenience
  agentState *the_agent, *the_client;

  if (!PyArg_ParseTuple(args, "ii",&the_dyad,&use_planning))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;
  samplerState *sstate = sstate_wrap->dyads[the_dyad].dsstate;
  
  // which method to use for action and emotion selection - could be an explicit input argument,
  // but right now we either use EXPECTATION or PLANTREE (if use_planning is set)
  AEMeth ae_method = EXPECTATION;
  
  if (use_planning)
    ae_method = PLANTREE;

  int num_plan_samples = sstate_wrap->num_plan_samples;
  int numsamples = sstate_wrap->numsamples;

  printf("the dyad: %d\n",the_dyad);

  the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;
  the_client = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].client_index;
  if (ae_method==PLANTREE) {
    // get the next emotions and actions with the planTree
    printf("PLANNING: getting next action to take\n");
    int a_sample_index;
    if (sstate_wrap->planTree == NULL) {
      printf("PLANNNING: initializing with %d samples\n",num_plan_samples);
      initializeTreeNode(num_plan_samples, &(sstate_wrap->planTree));
      for (int i=0; i<num_plan_samples; i++) {
	a_sample_index = drawSample(sstate->r, sstate->samples, numsamples);
	memcpy(sstate_wrap->planTree->samples+i,sstate->samples+a_sample_index,sizeof(sample));
      }
    }
    printf("PLANNNING: initializing with %d samples\n",num_plan_samples);
    seekNextAction(sstate_wrap->dyads+the_dyad, the_agent, BEH_AGENT,
		   sstate_wrap->planTree, sstate_wrap->num_plan_samples, &best_act, &best_emot, sstate_wrap->timeout);
    printTree(sstate_wrap->planTree, the_agent);
    // here is how we access index and string values for action and emotion and return it
  }
  if (ae_method == EXPECTATION) {
    behaviour * action_list;
    getNextAction(sstate_wrap->dyads+the_dyad, the_agent, &action_list, BEH_AGENT);
    best_act = action_list[0].index;
    free(action_list);
    
    emotion * emotion_list;
#ifdef SAMPLE_EMOTIONS
    sampleCurrentEmotion(sstate_wrap->dyads+the_dyad, the_agent, &emotion_list, AGENT);
#else
    getCurrentMostLikelyEmotion(sstate_wrap->dyads+the_dyad, the_agent, &emotion_list, AGENT);
    //num_emotions=computeEmotionFromAverage(sstate_wrap->sstate, sstate_wrap->dyads[the_dyad].dyad_index, &emotion_list, AGENT);
#endif
    best_emot = emotion_list[0].index;
    free(emotion_list);
  }
  if (ae_method == AVERAGE) {
    best_act = getNextActionFromAverageState(sstate_wrap->dyads+the_dyad, the_agent, BEH_AGENT);

    double *emot = (double *) malloc(SSD*sizeof(double));;
    for (int i=0; i<SSD; i++) 
      emot[i] = 0.0;
    computeExpectedEmotion(sstate_wrap->dyads+the_dyad, emot, AGENT);
    // look up labels
    somaticTransform(DENOTATIVE, sstate->r, sstate_wrap->dyads[the_dyad].dyad_index, &best_emot, emot, the_agent->entity_dicts[EMOTION]);
    free(emot);
  }
  return Py_BuildValue("ss",the_agent->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label,
		       the_agent->entity_dicts[EMOTION]->theDictionary[best_emot].label);
}

// get a variety of things computed from the state, such as the
// distribution over identities, the emotions, 
static PyObject* py_getExtendedState(PyObject* self, PyObject* args)
{
  int the_dyad=0;
  if (!PyArg_ParseTuple(args, "i",&the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates;
  samplerState * sstate = sstate_wrap->dyads[the_dyad].dsstate;
  // return a dictionary with all state
  PyObject *d = PyDict_New();

  int numsamples = sstate->numsamples;
  sample * the_samples = sstate->samples;
  agentState *the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;


  // some structures to collect the information we need
  double * f_avg, * tau_avg;
  identity * x_c_avg, *x_a_avg;
  xxstate xx_avg;
  double totalweight;

  int num_identities_agent=0;
  int num_identities_client=0;

  int num_behaviours_agent=0;
  int num_behaviours_client=0;
  int * count_list_agent = NULL;
  int * count_list_client = NULL;

  PyObject *client_emot_lst = PyList_New(sstate->numsamples*SSD);
  if (!client_emot_lst)
    return NULL;
  PyObject *agent_emot_lst = PyList_New(sstate->numsamples*SSD);
  if (!agent_emot_lst)
    return NULL;

  int k=0,l=0;
  PyObject *num;

  double tmp_emotion[SSD];
  for (int i=0; i<numsamples; i++) {
    computeEmotion(sstate_wrap->dyads[the_dyad].dsstate, the_samples[i].f+CLIENT*SSD, the_samples[i].tau+CLIENT*SSD, tmp_emotion);

    for (int j=0; j<SSD; j++) {
      num = PyFloat_FromDouble(tmp_emotion[j]);
      if (!num) {
        Py_DECREF(client_emot_lst);
        return NULL;
      }
      PyList_SET_ITEM(client_emot_lst, k, num);   // reference to num stolen
      k++;
    }
    computeEmotion(sstate_wrap->dyads[the_dyad].dsstate, the_samples[i].f+AGENT*SSD, the_samples[i].tau+AGENT*SSD, tmp_emotion);
    for (int j=0; j<SSD; j++) {
      num = PyFloat_FromDouble(tmp_emotion[j]);
      if (!num) {
        Py_DECREF(agent_emot_lst);
        return NULL;
      }
      PyList_SET_ITEM(agent_emot_lst, l, num);   // reference to num stolen
      l++;
    }
  }
  

  // gather summary statistics
  f_avg = (double *) malloc(VD*sizeof(double));
  tau_avg = (double *) malloc(VD*sizeof(double));

  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  // these start null -they will be allocated in insert_into_state_list
  x_c_avg = NULL;
  x_a_avg = NULL;
  initializeXXstate(&xx_avg);
  //XX: xx_avg = NULL;
  totalweight=0.0;
  
  // this loops through all the samples, and computes the weighted sum of fundamentals and transients,
  // and builds arrays x_c_avg, x_a_avg, turn_avg that are of type denotativeState,
  // each element of these arrays is a denotative state encountered in the sample set, and the "prob" field of the denotativeState
  // is a sum of the weights of all those samples.
  //count_list_client, count_list_agent and count_list_turn are lists of counts
  // for each corresponding denotativeState, which are not even used in fact since the sample weights are used to normalize
  for (int i=0; i<numsamples; i++) {
      for (int j=0; j<VD; j++) {
	f_avg[j] = f_avg[j]+the_samples[i].f[j]*the_samples[i].weight;
	tau_avg[j] = tau_avg[j]+the_samples[i].tau[j]*the_samples[i].weight;
      }
      addToXXstate(&xx_avg,&(the_samples[i].xx),the_samples[i].weight);
      num_identities_client = insert_into_state_list(&x_c_avg,&count_list_client,num_identities_client,
						     the_samples[i].x[ID_CLIENT],the_samples[i].weight);
      num_identities_agent = insert_into_state_list(&x_a_avg,&count_list_agent,num_identities_agent,
						    the_samples[i].x[ID_AGENT],the_samples[i].weight);

      totalweight += the_samples[i].weight;

      if (the_samples[i].who_acted==BEH_AGENT) 
	num_behaviours_agent++;
      if (the_samples[i].who_acted==BEH_CLIENT) 
	num_behaviours_client++;
  }
  for (int i=0; i<VD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }
  for (int i=0; i<num_identities_client; i++) 
    x_c_avg[i].prob /= totalweight;
  for (int i=0; i<num_identities_agent; i++) 
    x_a_avg[i].prob /= totalweight;

  qsort(x_c_avg, num_identities_client, sizeof(denotativeState), denotativeStateCompare);
  qsort(x_a_avg, num_identities_agent, sizeof(denotativeState), denotativeStateCompare);

  divideXXstate(&xx_avg,totalweight);
  
  PyObject *f_avg_lst = PyList_New(VD);
  if (!f_avg_lst)
    return NULL;
  for (int i=0; i<VD; i++) {
    num = PyFloat_FromDouble(f_avg[i]);
    if (!num) {
      Py_DECREF(f_avg_lst);
      return NULL;
    }
    PyList_SET_ITEM(f_avg_lst, i, num);   // reference to num stolen
  }
  PyObject *tau_avg_lst = PyList_New(VD);
  if (!tau_avg_lst)
    return NULL;
  for (int i=0; i<VD; i++) {
    num = PyFloat_FromDouble(tau_avg[i]);
    if (!num) {
      Py_DECREF(tau_avg_lst);
      return NULL;
    }
    PyList_SET_ITEM(tau_avg_lst, i, num);   // reference to num stolen
  }

  PyObject *client_ids_lst = PyList_New(num_identities_client);
  if (!client_ids_lst)
    return NULL;
  for (int i=0; i<num_identities_client; i++) {
    num = PyTuple_New(2);
    if (!num) {
      Py_DECREF(client_ids_lst);
      return NULL;
    }
    PyTuple_SET_ITEM(num, 0, Py_BuildValue("s",the_agent->entity_dicts[CLIENT]->theDictionary[x_c_avg[i].index].label));
    PyTuple_SET_ITEM(num, 1, PyFloat_FromDouble(x_c_avg[i].prob));
    PyList_SET_ITEM(client_ids_lst, i, num);   // reference to num stolen
  }

  PyObject *agent_ids_lst = PyList_New(num_identities_agent);
  if (!agent_ids_lst)
    return NULL;
  for (int i=0; i<num_identities_agent; i++) {
    num = PyTuple_New(2);
    if (!num) {
      Py_DECREF(agent_ids_lst);
      return NULL;
    }
    PyTuple_SET_ITEM(num, 0, Py_BuildValue("s",the_agent->entity_dicts[AGENT]->theDictionary[x_a_avg[i].index].label));
    PyTuple_SET_ITEM(num, 1, PyFloat_FromDouble(x_a_avg[i].prob));
    PyList_SET_ITEM(agent_ids_lst, i, num);   // reference to num stolen
  }

  /*
  printf("emotions from averaged fundamentals and transients: \n");
  //double *tmp_emotion = (double *) malloc(SSD*sizeof(double));
  printf("agent: \n");
  computeEmotion(sstate, dyad_index, f_avg+AGENT*SSD, tau_avg+AGENT*SSD, tmp_emotion);
  for (int i=0; i<SSD; i++)
    printf("%4.3g\t",tmp_emotion[i]);
  printf("\n\n");
  printf("client: \n");
  computeEmotion(sstate, dyad_index, f_avg+CLIENT*SSD, tau_avg+CLIENT*SSD, tmp_emotion);
  for (int i=0; i<SSD; i++)
    printf("%4.3g\t",tmp_emotion[i]);
  printf("\n\n");
  //free(tmp_emotion);
  
  printf("number of samples that used agent behaviour %d\n",num_behaviours_agent);
  printf("number of samples that used client behaviour %d\n",num_behaviours_client);
  //XX:
  printf("xx average: \n");
  printRawXXstate(&xx_avg);
  */

  PyDict_SetItem(d, Py_BuildValue("s","agent_emot"), agent_emot_lst);
  PyDict_SetItem(d, Py_BuildValue("s","client_emot"), client_emot_lst);
  PyDict_SetItem(d, Py_BuildValue("s","f_avg"), f_avg_lst);
  PyDict_SetItem(d, Py_BuildValue("s","tau_avg"), tau_avg_lst);
  PyDict_SetItem(d, Py_BuildValue("s","agent_ids"), agent_ids_lst);
  PyDict_SetItem(d, Py_BuildValue("s","client_ids"), client_ids_lst);


  free(f_avg);
  free(tau_avg);
  free(x_c_avg);
  free(x_a_avg);
  free(count_list_client);
  free(count_list_agent);



  return d;

}

// gets all the state information in a dictionary which is returned.
// the dictionary keys are:
// agentdeflection:  deflection for agent identity
// behaviourdeflection:  deflection for client identity
// clientdeflection: deflection for behaviour
//
static PyObject* py_getState(PyObject* self, PyObject* args)
{
  int the_dyad=0;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  samplerState * sstate = sstate_wrap->dyads[the_dyad].dsstate;
  // return a dictionary with all state
  PyObject *d = PyDict_New();

  
  // first, deflections
  double agdefl,behdefl,cldefl;
  agdefl = computeDeflection(sstate,  AGENT);
  behdefl= computeDeflection(sstate,  BEHAVIOUR);
  cldefl = computeDeflection(sstate,  CLIENT);

  PyDict_SetItem(d, Py_BuildValue("s","agentdeflection"), Py_BuildValue("d",agdefl));
  PyDict_SetItem(d, Py_BuildValue("s","behaviourdeflection"), Py_BuildValue("d",behdefl));
  PyDict_SetItem(d, Py_BuildValue("s","clientdeflection"), Py_BuildValue("d",cldefl));
  PyDict_SetItem(d, Py_BuildValue("s","numsamples"), Py_BuildValue("i",sstate->numsamples));

  PyObject *f_lst = PyList_New(sstate->numsamples*VD);
  if (!f_lst)
    return NULL;

  PyObject *tau_lst = PyList_New(sstate->numsamples*VD);
  if (!tau_lst)
    return NULL;
  PyObject *x_lst = PyList_New(sstate->numsamples*NGTD);
  if (!x_lst)
    return NULL;
  // weight list
  PyObject *weight_lst = PyList_New(sstate->numsamples);
  if (!weight_lst)
    return NULL;

  // actor list
  PyObject *actor_lst = PyList_New(sstate->numsamples);
  if (!actor_lst)
    return NULL;
  // reset list
  PyObject *reset_lst = PyList_New(sstate->numsamples);
  if (!reset_lst)
    return NULL;

  int k=0,l=0;
  PyObject *num;
  for (int i = 0; i < sstate->numsamples; i++) {
    for (int j=0; j<VD; j++) {
      num = PyFloat_FromDouble(sstate->samples[i].f[j]);
      if (!num) {
        Py_DECREF(f_lst);
        return NULL;
      }
      PyList_SET_ITEM(f_lst, k, num);   // reference to num stolen

      num = PyFloat_FromDouble(sstate->samples[i].tau[j]);
      if (!num) {
        Py_DECREF(tau_lst);
        return NULL;
      }
      PyList_SET_ITEM(tau_lst, k, num);   // reference to num stolen
      k++;
    }
    for (int j=0; j<NGTD; j++) {
      num = Py_BuildValue("i",sstate->samples[i].x[j]);
      if (!num) {
        Py_DECREF(x_lst);
        return NULL;
      }
      PyList_SET_ITEM(x_lst, l, num);   // reference to num stolen
      l++;
    }
    num = PyFloat_FromDouble(sstate->samples[i].weight);
    if (!num) {
      Py_DECREF(weight_lst);
      return NULL;
    }
    PyList_SET_ITEM(weight_lst, i, num);
    
    num = Py_BuildValue("i",sstate->samples[i].who_acted);
    if (!num) {
      Py_DECREF(actor_lst);
      return NULL;
    }
    PyList_SET_ITEM(actor_lst, i, num);

    num = Py_BuildValue("i",sstate->samples[i].reset_flag);
    if (!num) {
      Py_DECREF(reset_lst);
      return NULL;
    }
    PyList_SET_ITEM(reset_lst, i, num);
  }

  PyDict_SetItem(d, Py_BuildValue("s","f"), f_lst);
  PyDict_SetItem(d, Py_BuildValue("s","tau"), tau_lst);
  PyDict_SetItem(d, Py_BuildValue("s","x"), x_lst);
  PyDict_SetItem(d, Py_BuildValue("s","weight"), weight_lst);
  PyDict_SetItem(d, Py_BuildValue("s","actor"), actor_lst);
  PyDict_SetItem(d, Py_BuildValue("s","reset"), reset_lst);
  

  
  return d;
}
static PyObject* py_getDeflections(PyObject* self, PyObject* args)
{
  int the_dyad;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  samplerState * sstate = sstate_wrap->dyads[the_dyad].dsstate;

  //samplerState for this sampler
  double agdefl,behdefl,cldefl;
  agdefl = computeDeflection(sstate, AGENT);
  behdefl= computeDeflection(sstate, BEHAVIOUR);
  cldefl = computeDeflection(sstate, CLIENT);

  // return a tuple of four doubles, one for each deflection
  return Py_BuildValue("ddd",agdefl,behdefl,cldefl);
}
  
static PyObject* py_getRandomSeed(PyObject* self, PyObject* args)
{
  int the_dyad;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  samplerState * sstate = sstate_wrap->dyads[the_dyad].dsstate;
  // return a tuple of four doubles, one for each deflection
  return Py_BuildValue("l",sstate->random_seed);
}
  
static PyObject* py_getIdentities(PyObject* self, PyObject* args)
{
  agentState *the_agent;
  int the_dyad;
  if (!PyArg_ParseTuple(args, "i", &the_dyad))
    return NULL;
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  samplerState * sstate = sstate_wrap->dyads[the_dyad].dsstate;
  identity *x_c_avg, *x_a_avg;
  int num_identities_agent, num_identities_client;
  getIdentities(sstate, &num_identities_agent, &x_a_avg, &num_identities_client, &x_c_avg);

  the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;

  int a_index = x_a_avg[0].index;
  int c_index = x_c_avg[0].index;
  free(x_c_avg);
  free(x_a_avg);
  // return a string with the identities for agent and client
  return Py_BuildValue("ss",the_agent->entity_dicts[AGENT]->theDictionary[a_index].label,
		       the_agent->entity_dicts[CLIENT]->theDictionary[c_index].label);
  
}

// updates the state
// This function takes seven arguments: dyad index first, followed by five strings:
// the xx observation  (if not present, pass in "")
// the client action and client emotion,
// then agent action and agent emotion
// and finally, whether or not to use the planning engine (a binary flag)
// and updates the sampler state (computes the posterior) for the agent specified by the dyad index
// in the sampler corresponding to the sampler index. Returns the number of nodes in the plan tree after pruning
// just because I couldn't think of anything else to return
static PyObject* py_updateState(PyObject* self, PyObject* args)
{
  behaviour agent_behaviour, client_behaviour;
  emotion agent_emotion, client_emotion;
  int agent_acted=0, client_acted=0;
  int numnodes=0;
  int gotit;
  int use_planning;
  xostate xx_observation;
  char *xxobs;
  char *cbeh, *cemot;
  char *abeh, *aemot;
  // for convenience
  agentState *the_agent, *the_client;
  int the_dyad;

  //get the module state
  if (!PyArg_ParseTuple(args, "isssssi", &the_dyad, &xxobs, &cbeh, &cemot, &abeh, &aemot, &use_planning))
    return NULL;
  printf("the dyad: %d\n",the_dyad);
  //get the module state
  samplerStateSet *ssset = GETSTATE(self);
  //samplerState for this sampler
  samplerStateWrapper * sstate_wrap = ssset->samplerStates; 
  dyad *the_dyad_for_realsies = sstate_wrap->dyads+the_dyad;
  memcpy(client_behaviour.label,cbeh,MAX_LABEL_LENGTH*sizeof(char));
  memcpy(client_emotion.label,cemot,MAX_LABEL_LENGTH*sizeof(char));
  memcpy(agent_behaviour.label,abeh,MAX_LABEL_LENGTH*sizeof(char));
  memcpy(agent_emotion.label,aemot,MAX_LABEL_LENGTH*sizeof(char));
  
  printf("calling update with %s %s %s %s\n",agent_behaviour.label,client_behaviour.label,agent_emotion.label,client_emotion.label);
  //printf("timeout is %d\n",sstate_wrap->timeout);
  printf("xx observation :  %s\n",xxobs);

  the_agent = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].agent_index;
  the_client = sstate_wrap->agents+sstate_wrap->dyads[the_dyad].client_index;


  gotit=lookUp(&agent_behaviour,the_agent->entity_dicts[BEHAVIOUR]);
  if (gotit>=0) 
    agent_acted=1;
  gotit=lookUp(&agent_emotion,the_agent->entity_dicts[EMOTION]);

  if (gotit<0) 
    return Py_BuildValue("i",-1);

  gotit = lookUp(&client_behaviour,the_agent->entity_dicts[BEHAVIOUR]);

  if (gotit>=0) 
    client_acted=1;

  gotit=lookUp(&client_emotion,the_agent->entity_dicts[EMOTION]);

  if (gotit<0) 
    return Py_BuildValue("i",-1);

  printf("....................%d %d %d %d\n",agent_behaviour.index,client_behaviour.index,agent_emotion.index,client_emotion.index);
  // if null then we use NULL
  if (strcmp(xxobs,"")==0)  {
    updateFullStateTwoActorsWithEmotionWithXXObs(the_dyad_for_realsies, the_agent,
						 (agent_acted ? &agent_behaviour : NULL),
						 (client_acted ? &client_behaviour : NULL),
						 NULL, &agent_emotion, &client_emotion);
  } else {
    // else we convert to the  observation
    gotit = stringToXXObs(xxobs,&xx_observation);
    if (gotit<0) 
      return Py_BuildValue("i",-1);
    printRawXXObs(&xx_observation);
    updateFullStateTwoActorsWithEmotionWithXXObs(the_dyad_for_realsies, the_agent,
						 (agent_acted ? &agent_behaviour : NULL),
						 (client_acted ? &client_behaviour : NULL),
						 &xx_observation, &agent_emotion, &client_emotion);
  }
  if (use_planning) {
    printf("pruning tree\n");
    // comented out temporarily while working on making pywrapper work with 2.3.8
    //pruneTree(&(sstate_wrap->planTree), &agent_behaviour, &agent_emotion, &client_behaviour, &client_emotion, &xx_observation);
    //numnodes = countTreeNodes(sstate_wrap->planTree);
    printf("number of nodes in planTree after pruning: %d\n",numnodes);
  }

  // a hack to reduce the timeout after the first iteration
  //if (sstate_wrap->timeout > 3) {
   //sstate_wrap->timeout = 3;
  // }
  
  return Py_BuildValue("i",&numnodes);
}

// Bind Python function names to our C functions
static PyMethodDef libbayesact_methods[] = {
  {"initializeState", py_initializeState, METH_VARARGS},
  {"setTimeout",py_setTimeout, METH_VARARGS},
  {"freeState", py_freeState, METH_VARARGS},
  {"updateState", py_updateState, METH_VARARGS},
  {"getIdentities",py_getIdentities, METH_VARARGS},
  {"getRandomSeed",py_getRandomSeed, METH_VARARGS},
  {"getDeflections",py_getDeflections, METH_VARARGS},
  {"getState",py_getState, METH_VARARGS},
  {"getExtendedState",py_getExtendedState, METH_VARARGS},
  {"getNextActionEmotion", py_getNextActionEmotion, METH_VARARGS},
  {"getNextMostLikelyEmotion", py_getNextMostLikelyEmotion, METH_VARARGS},
  {"sampleCurrentEmotion", py_sampleCurrentEmotion, METH_VARARGS},
  {"getCurrentMostLikelyEmotion", py_getCurrentMostLikelyEmotion, METH_VARARGS},
  {"getNextMostLikelyAction", py_getNextMostLikelyAction, METH_VARARGS},
  {NULL, NULL}
};

// some other functions that are needed (?) 
static int libbayesact_traverse(PyObject *m, visitproc visit, void *arg) {
    Py_VISIT(GETSTATE(m)->error);
    return 0;
}
// leave the rest of the memory freeing to the user using py_freeState
static int libbayesact_clear(PyObject *m) {
    Py_CLEAR(GETSTATE(m)->error);
    return 0;
}

// module definition
static struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT,
        "libbayesact",
        NULL,
        sizeof(samplerStateSet),
        libbayesact_methods,
        NULL,
        libbayesact_traverse,
        libbayesact_clear,
        NULL
};

#define INITERROR return NULL

// initialization function
PyMODINIT_FUNC PyInit_libbayesact(void) {
  PyObject *module = PyModule_Create(&moduledef);
  if (module == NULL)
    INITERROR;
  samplerStateSet *ssset = GETSTATE(module);

  ssset->__CURRENT_NUM_SAMPLERS = 1;  // this is fixed and could be removed
  ssset->samplerStates = PyMem_RawMalloc(sizeof(samplerStateWrapper));

  
  ssset->error = PyErr_NewException("libbayesact.Error", NULL, NULL);
  if (ssset->error == NULL) {
    Py_DECREF(module);
    INITERROR;
  }

  return module;
}
