/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
# Test main function. Demonstrates reading a simulation file, setting up the simulation and then #
# running the simulation in user-query mode or automatic and printing the average state          #
#************************************************************************************************/

#include "inputparser.h"
#include "fvarsample.h"

// functions to do planning
#include "forvalue.h"

int main(int argc, char **argv)
{
  // whether to use the default action at each step, or to sample from the action distribution
  int default_action=0;

  // whether to allow resets in identity to be done
  // hard resets when all weights are 0 are always done
  int do_resets=0;
  

  // use emotions at all? 
  int use_emotions=1;
  // wether to use emotions during planning or not
  int emotion_planning=0;

  int use_xx_state=0;
  xostate xx_observation;
  // this is the simulation file that describes what is to be done - see testsim.txt
  char *simfile="testsim.txt";

  // number of samples to use for the particle filter
  // more means more precise estimates
  int num_samples=100;


  
  // planning stuff
  // the plan tree
  treeNode *planTree=NULL;
  // the number of samples to use for planning - set to zero to not do any
  int num_plan_samples=0;
  int best_act, best_emot;
  int timeout=0;


  int thearg=1;
  while (thearg<argc) {
    if (strncmp(argv[thearg],"-ep",3)==0) {
      emotion_planning=1;
    } else if (strncmp(argv[thearg],"-e",2)==0) {
      use_emotions=1;
    } else if (strncmp(argv[thearg],"-r",2)==0) {
      do_resets=1;
    } else if (strncmp(argv[thearg],"-d",2)==0) {
      default_action=1;
    } else if (strncmp(argv[thearg],"-x",2)==0) {
      use_xx_state=1;
    } else if (strncmp(argv[thearg],"-h",2)==0) {
      printf("usage: testbayesact [-e] [-ep] [-x] [-h] [-n num] [-p num] [-t num][simfile]\n\t -r : allow identity resets (re-identification, default no)\n\t -e : use emotions (default no)\n\t -x : use additional x state and observations (default no)\n\t -d : use default action only (default no)\n\t -h : print usage\n\t -n <num> : use num samples for belief monitoring (default %d)\n\t -p <num> : use num samples for plan tree generation (default no plan tree)\n\t -t <num> : timeout (in seconds) for planning (default 0 or 1 rollout)\n\t -ep: plan over emotions as well (default no)\n\t simfile: simulation description file\n",num_samples);
      exit(0);
    } else if (strncmp(argv[thearg],"-n",2)==0) {
      thearg++;
      num_samples=atoi(argv[thearg]);
    } else if (strncmp(argv[thearg],"-t",2)==0) {
      thearg++;
      timeout=atoi(argv[thearg]);
    } else if (strncmp(argv[thearg],"-p",2)==0) {
      thearg++;
      num_plan_samples=atoi(argv[thearg]);
    } else {
      simfile = argv[thearg];
    }
    thearg++;
  }

  
  printf("reading from %s\n", simfile);
  printf("number of samples %d\n",num_samples);
  printf("number of plan samples %d\n",num_plan_samples);
  if (use_emotions)
    printf("using emotions\n");
  if (emotion_planning)
    printf("planning over emotions is on\n");
  if (default_action)
    printf("will use default actions only\n");
  // ================================================================================================
  // agents and dictionaries 
  // ================================================================================================
  agentState * agents = (agentState *) malloc(MAX_NUM_AGENTS*sizeof(agentState));
  int num_agents;
  dictionary * dictionaries = (dictionary *) malloc(MAX_NUM_DICTIONARIES*sizeof(dictionary));
  int num_dictionaries=0;
  // for convenience
  agentState *the_agent, *the_client;
  
  samplerState *sstate;
  
  // ================================================================================================
  // read the simulation input file - this initializes everything
  // including all agents, dictionaries, and interactions
  // ================================================================================================
  dyad * dyads = (dyad *) malloc(MAX_NUM_INTERACTIONS*sizeof(dyad));
  int num_dyads;
  // if 0, default behaviours are taken, else behaviours are queried from the user
  query_type query_user=USER;
  int max_iterations=50;
  event *events = NULL;
  //dict_type dtype=MEAN;
  printf("setting up agents and dyads\n");
  double gamma=-1.0;
  // this function does all the work - reads in the simulation setup, initializes all dictionaries and dyads
  // dyads have indices agent_index and client_index refeingr to agents in the list agents
  setUpAgentsDyads(simfile, agents, &num_agents, dyads, &num_dyads, &query_user, &max_iterations, gamma,
		   dictionaries, &num_dictionaries, &events);

  // ================================================================================================
  // some bookkeeping and settings for the simulation
  // ================================================================================================
  // some buffers for behaviours 
  char input_buffer[MAX_LABEL_LENGTH];
  behaviour the_behaviours[2];
  // need to have these indices so we can easily assign NULL
  int acted[2];
  
  // other parameters of the simulation
  // we will iterate over these two 
  c_entity_type actor_iter[2]={AGENT,CLIENT};
  d_entity_type who_acted[2]={BEH_AGENT,BEH_CLIENT};
  // these will eventually become properties of "Agent"
  char  * agent_names[2];
  double totalweight;
  int iteration=0;
  int num_potential_actions, num_potential_emotions, best_action_from_average;

  // number of best actions to show
  int num_to_show=5;
  int gotit;
  behaviour * beh_list;
  emotion * emot_list=NULL;
  double *emot = (double *) malloc(SSD*sizeof(double));;
  for (int i=0; i<SSD; i++) 
    emot[i] = 0.0;
    
  emotion emotState[2];
  emotion tmp_emot;
  int index;
  double emot_prob;

  
  int the_dyad=0;
  // ================================================================================================
  // main simulation loop
  // ================================================================================================
  while (iteration < max_iterations) {

    // dyad selection step
    printf("available dyads:\n");
    for (int j=0; j<num_dyads; j++) 
      printf("%d: %s - %s\n",j,agents[dyads[j].agent_index].name,agents[dyads[j].client_index].name);
    printf("select dyad to simulate (enter for same as last time %d): ",the_dyad);
    fgets(input_buffer,MAX_LABEL_LENGTH,stdin);

    // remove trailing newline
    input_buffer[strcspn(input_buffer, "\n")] = 0;
    if (strlen(input_buffer)!=0) 
      the_dyad=atoi(input_buffer);
    
    // could also here find the opposite dyad - the one with
    //dyads[the_opposite_dyad].agent_index = dyads[the_dyad].client_index
    //dyads[the_opposite_dyad].client_index = dyads[the_dyad].agent_index
    // and simulate that as well simultaneously

    the_agent = agents+dyads[the_dyad].agent_index;
    the_client = agents+dyads[the_dyad].client_index;
    printf("agent index %d client index %d\n",dyads[the_dyad].agent_index,dyads[the_dyad].client_index);
    printf("agent %s  client %s\n",the_agent->name,the_client->name);
    agent_names[0]=the_agent->name;
    agent_names[1]=the_client->name;
    printf("***************************************************************** iteration %d\n",iteration);
    printSamples(dyads+the_dyad, the_agent);
    printf("------------------------------------------------------------------\n");

    sstate = dyads[the_dyad].dsstate;
    // compute and print deflections
    printf("deflections:\n");

    printDeflections(sstate,AVERAGE);
    printf("------------------------------------------------------------------\n");

    // compute emotions
    printf("emotions:\n");
    for (int tt=0; tt<2; tt++) {
      // compute the expected emotion based on the state - the expectation of emotion
      // now this is folded into getNextEmotion - this option will be used if sstate->use_emotion=0
      // and is in the function getNextMostLikelyEmotion() in fvarsample.c

      // we still want to *see* the emotions if mimicking interact, just not use them in an update
      // ********** AVERAGED method
      if (use_emotions) {  
	computeExpectedEmotion(dyads+the_dyad, emot, actor_iter[tt]);
	// look up labels
	somaticTransform(DENOTATIVE, sstate->r, dyads[the_dyad].dyad_index, &index, emot, the_agent->entity_dicts[EMOTION]);
      } else {
	getNullEmotion(dyads+the_dyad, the_agent, &index, &emot_prob, actor_iter[tt]);
      }
      printf("emotion for %s:\n",agent_names[tt]);
      emotState[tt].index = index;
      // get label
      strcpy(emotState[tt].label,the_agent->entity_dicts[EMOTION]->theDictionary[emotState[tt].index].label);
      printf("=====================================================================\n");
      printf("           AVERAGED method\n");
      for (int j=0; j<SSD; j++)
	printf("%4.3g ",emot[j]);
      printf("  label: %s\n",emotState[tt].label);

      // now get a full list of emotions that could be displayed - this uses a somatic transform per sample, and
      // aggregates the results in a distribution over the denotative emotion state
      printf("=====================================================================\n");
      printf("           EXPECTATION method\n");
      num_potential_emotions = getCurrentEmotionList(dyads+the_dyad, the_agent, &emot_list, actor_iter[tt]);
      if (num_potential_emotions > 0)  {
	printf("....................................................\n");
	printf("got %d potential current emotions for %s, top %d are:\n",num_potential_emotions,agent_names[tt],num_to_show);
	printf("rank\t*\tlabel (probability)\n");
	printf("____________________________________________________________________\n");

	//show top 5 - emot_list is sorted on return from getNextAction
	for (int i=0; i< MIN(num_to_show,num_potential_emotions); i++)  
	  printf("%d\t*\t%s (%4.3g)\n",i+1,the_agent->entity_dicts[EMOTION]->theDictionary[emot_list[i].index].label,emot_list[i].prob);
      }
      printf("------------------------------------------------------------------\n");
      free(emot_list);
      emot_list=NULL;
      // get the next emotions
      /*
      num_potential_emotions = getNextEmotionList(sstate, dyads[the_dyad].dyad_index, the_agent, &emot_list, actor_iter[tt], turn_iter[tt]);
      if (num_potential_emotions > 0)  {
	printf("....................................................\n");
	printf("got %d potential next emotions for %s, top %d are:\n",num_potential_emotions,agent_names[tt],num_to_show);
	printf("rank\t*\tlabel (probability)\n");
	printf("____________________________________________________________________\n");

	//show top 5 - emot_list is sorted on return from getNextAction
	for (int i=0; i< MIN(num_to_show,num_potential_emotions); i++)  
	  printf("%d\t*\t%s (%4.3g)\n",i+1,the_agent->entity_dicts[EMOTION]->theDictionary[emot_list[i].index].label,emot_list[i].prob);
      }
      */
    }

    // get the next emotions and actions with the planTree
    // 
    if (num_plan_samples > 0) {
      printf("PLANNING: getting next action to take\n");
      int a_sample_index;
      if (planTree == NULL) {
	printf("PLANNNING: initializing with %d samples\n",num_plan_samples);
	initializeTreeNode(num_plan_samples, &planTree);
	for (int i=0; i<num_plan_samples; i++) {
	  a_sample_index = drawSample(sstate->r, sstate->samples, num_samples);
	  memcpy(planTree->samples+i, sstate->samples+a_sample_index, sizeof(sample));
	}
      }
      //printTree(planTree, the_agent);
      printf("PLANNNING: seeking next action for %d %d\n",the_dyad,dyads[the_dyad].dyad_index);
      seekNextAction(dyads+the_dyad, the_agent, BEH_AGENT, planTree, num_plan_samples, &best_act, &best_emot, timeout);
      printf("PLANNING: best action to take %d (%s) best emotion %d (%s) has value %g\n", best_act, the_agent->entity_dicts[BEHAVIOUR]->theDictionary[best_act].label, best_emot, the_agent->entity_dicts[EMOTION]->theDictionary[best_emot].label, planTree->value);

      treeNode *temp = planTree->children;
      double giveval=0,takeval=0;
      while (temp != NULL) {
	if (temp->index.act_index==0) 
	  giveval += temp->value;
	else
	  takeval += temp->value;
	temp = temp->siblings;
      }
      printf("give value: %g  take value: %g  ratio: %g\n",giveval,takeval,giveval/takeval);
      
      printf("see the plan tree y/[N]?");
      fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
      // remove trailing newline
      input_buffer[strcspn(input_buffer, "\n")] = 0;
      if (strlen(input_buffer)>0)  {
	printTree(planTree, the_agent);
	//printTreeTikZ(planTree, the_agent);
      }
      printf("number of nodes in planTree after planning: %d\n",countTreeNodes(planTree));
    }

    
    printf("Getting next actions:\n");

    // predict next default action by sampling from the next state and taking the behaviour
    // then sampling from x_b given f_b with a DENOTATIVE  somatic transform
    // iterate over agent/client and get possible actions for each, query user for action to take
    // either (but not both) may be not specified
    // this is an *alternative* to the planning method above - simpler and faster, but does not take
    // rewards into account
    for (int tt=0; tt<2; tt++) {
      beh_list=NULL;
      // this will mimic what Interact computes 
      printf("=====================================================================\n");
      printf("                 AVERAGED method\n");
      best_action_from_average = getNextActionFromAverageState(dyads+the_dyad, the_agent, who_acted[tt]);
      if (best_action_from_average >= 0) 
	printf("best next action from average state: %s\n",the_agent->entity_dicts[BEHAVIOUR]->theDictionary[best_action_from_average].label);
      printf("\n");

      printf("=====================================================================\n");
      printf("                 EXPECTATION method\n");
      //printf("potential next actions for %s\n",turn_labels[turn_iter[tt]]);
      num_potential_actions = getNextAction(dyads+the_dyad,the_agent,
					    &beh_list, who_acted[tt]);
      //check if this (Agent or Client) have any potential actions
      if (num_potential_actions > 0)  {
	printf("....................................................\n");
	printf("got %d potential next actions for %s, top %d are:\n",num_potential_actions,agent_names[tt],num_to_show);
	printf("rank\t*\tlabel (probability)\n");
	printf("____________________________________________________________________\n");

	//show top 5 - beh_list is sorted on return from getNextAction
	for (int i=0; i< MIN(num_to_show,num_potential_actions); i++)  
	  printf("%d\t*\t%s (%4.3g)\n",i+1,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[beh_list[i].index].label,beh_list[i].prob);
	// show best behaviour to take
	printf("------------------------------------------------------------------\n");

      
	acted[tt]=0;
	// see bayesactsim.c for a version that can handle query_user=EVENTS
	if (query_user==BAYESACT) {
	  // run the default option - expectation mode
	  the_behaviours[tt].index = beh_list[0].index;
	  strcpy(the_behaviours[tt].label,the_agent->entity_dicts[BEHAVIOUR]->theDictionary[beh_list[0].index].label);
	  acted[tt]=1;
	} else {
	  // or  query the user for which behaviour to take
	  gotit=-1;
	  while (gotit<0) {
	    printf("behaviour to take for %s acting on %s (enter for none):  ",agent_names[tt], agent_names[1-tt]); 
	    fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	    // remove trailing newline
	    input_buffer[strcspn(input_buffer, "\n")] = 0;
	    if (strlen(input_buffer)==0) {
	      if (tt==1 && !acted[0]) {
		printf("you said %s didn't act, so %s must! try again\n",agent_names[tt],agent_names[1-tt]);
	      } else {
		gotit=1;
	      }
	    } else {
	      strcpy(the_behaviours[tt].label,input_buffer);
	      gotit=lookUp(&the_behaviours[tt],the_agent->entity_dicts[BEHAVIOUR]);
	      if (gotit>=0) 
		acted[tt]=1;
	    }
	  }
	}
	// query the user for which emotion to display
	if (use_emotions) {
	  gotit = -1;
	  while (gotit<0) {
	    printf("emotion to display for %s (enter for %s):  ",agent_names[tt],emotState[tt].label);
	    fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	    // remove trailing newline
	    input_buffer[strcspn(input_buffer, "\n")] = 0;
	    if (strlen(input_buffer)==0) {
	      gotit=1;
	    } else {
	      strcpy(tmp_emot.label,emotState[tt].label);
	      strcpy(emotState[tt].label,input_buffer);
	      gotit=lookUp(&emotState[tt],the_agent->entity_dicts[EMOTION]);
	      if (gotit<0)  
		strcpy(emotState[tt].label,tmp_emot.label);
	    }
	  }
	  printf("emotion displayed by %s is: %s\n",agent_names[tt],emotState[tt].label);
	}
	if (acted[tt]) {
	  printf("Behaviour taken by %s is : %s\n", agent_names[tt], the_behaviours[tt].label);
	} else {
	  printf("%s doesn't act\n",agent_names[tt]);
	}
      }
      free(beh_list);
    }
    // query the user for the xx (additional state) observation
    if (use_xx_state) {
      gotit = -1;
      while (gotit<0) {
	printf("observation of xx:  ");
	fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	// remove trailing newline
	input_buffer[strcspn(input_buffer, "\n")] = 0;
	gotit = stringToXXObs(input_buffer,&xx_observation);
      }
      printRawXXObs(&xx_observation);
    }

    // The state update function. If either actor has not acted, their action is NULL and will not be performed
    // the samples with that NULL actor as turn will be discarded
    printf("performing the update - choose the same dyad again to see the results\n");
   
    if (use_emotions)  {
      if (use_xx_state) {
	// most general case
	totalweight = updateFullStateTwoActorsWithEmotionWithXXObs(dyads+the_dyad, the_agent,
							  (acted[0] ? &the_behaviours[0] : NULL),
							  (acted[1] ? &the_behaviours[1] : NULL),
							  &xx_observation,
							  emotState, emotState+1);
      } else {
	totalweight = updateFullStateTwoActorsWithEmotion(dyads+the_dyad, the_agent,
							  (acted[0] ? &the_behaviours[0] : NULL),
							  (acted[1] ? &the_behaviours[1] : NULL),
							  emotState, emotState+1);
      }	
    } else {
      if (use_xx_state) {
	totalweight = updateFullStateTwoActorsWithXXObs(dyads+the_dyad,the_agent,
							(acted[0] ? &the_behaviours[0] : NULL),
							(acted[1] ? &the_behaviours[1] : NULL), &xx_observation);
      } else {
	totalweight = updateFullStateTwoActors(dyads+the_dyad, the_agent,
					       (acted[0] ? &the_behaviours[0] : NULL),
					       (acted[1] ? &the_behaviours[1] : NULL));
      }
    }

    // prune the planTree - if this fails then planTree will return NULL and
    // will be re-initialized on the next pass above
    if (num_plan_samples > 0) {
      pruneTree(&planTree, &the_behaviours[0], emotState, &the_behaviours[1], emotState+1, &xx_observation);
      //printTree(planTree, the_agent);
      printf("number of nodes in planTree after pruning: %d\n",countTreeNodes(planTree));
    }    
    iteration++;
  }
  
  // ================================================================================================
  // clean up
  // ================================================================================================
  free(emot);
  for (int j=0; j<num_dictionaries; j++) 
    freeSentimentDictionary(dictionaries+j);
  free(dictionaries);
  for (int j=0; j<num_dyads; j++)
    freeParameters(dyads+j);
  free(dyads);
  free(agents);
  freeBayesACT(sstate);
}
