#************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
#  generic agent python test class - tests generic (no additional X state, no planning) simulation
#  with BayesACT
#*************************************************************************************************

# import the main libbayesact module - this will allocate the basic memory structures needed
from libbayesact import *

# a generic set of interactions defined by a group of agents as defined in a "testsim.txt" like file
# this python class essentially wraps the state created in BayesACT samplerStateWrapper, and defines
# whatever functions are needed to access and manipulate that state.  Each function that we want to
# use here has to be defined in pywrapper.c so it can manipulate the actual agent states that exist
# in the C memory space.
class agentGroup(object):
    def __init__(self,*args,**kwargs):

        # inputs are the number of samples
        self.num_samples=kwargs.get("num_samples",1000)
        # the simulation file
        self.simfile=kwargs.get("simfile","testsim.txt")

        # planning elements (not used right now) -----------------------------
        # if using the planning engine, you have to get the next action to take by calling getNextActionEmotion
        # which will create and update the plan tree, otherwise, we use the functions getMostLikely{Action/Emotion}.
        # which simply computes an expectation
        # Also, when updating the state, we have to prune the planTree when using planning
        self.use_planning = 0
        # number of samples to use for planning -- this is not used yet
        self.num_plan_samples=kwargs.get("num_plan_samples",500)
        # planning timeout -- again, not used yet
        self.timeout=kwargs.get("timeout",1)
        # whether to plan over emotions or not - could be an input argument.
        self.emotion_planning=0
        # ---------------------------------------------------------------------

        # set up the main state 
        # this index must be used for all future calls to this sampler
        # multiple sampler states (corresponding to different groups of agents) can be initialized
        # with multiple calls to this function, which we now wrap in agentGroup, so will be created by
        # creating multiple agent groups in whatever python code uses this class
        self.index = initializeState(self.num_samples, self.num_plan_samples, self.timeout, self.simfile, self.emotion_planning)
        if self.index < 0:
            raise NameError("ERROR: too many samplers initialized")
        print(100*"*")
        print("new sampler index "+str(self.index))
        
        # if multiple dyads are specified in simfile then we can refer to specific ones from now on
        self.the_dyad = 0

    #free the state
    def freeGroup(self):
        freeState(self.the_dyad)

    #query the agent for its next (default only) emotion
    def get_emotion(self):
        self.agent_emotion = getNextMostLikelyEmotion(self.the_dyad)
        return self.agent_emotion

    #query the agent for its next (default only) action
    def get_action(self):
        self.agent_action = getNextMostLikelyAction(sself.the_dyad)
        return self.agent_action

    #query the bot for its next action and emotion to take
    #returns a tuple of (action, emotion) strings
    def get_action_emotion(self):
        # either do this by consulting (and building) a plan tree, which can also
        # include planning over emotions
        if self.use_planning:
            return getNextActionEmotion(self.the_dyad, self.emotion_planning)
        else:
            # or by getting the default (singled most aligned) action and emotion
            return (self.get_action(),self.get_emotion())

    def get_deflections(self):
        return getDeflections(self.the_dyad);
    def get_random_seed(self):
        return getRandomSeed(self.the_dyad);

    def get_state(self):
        return getState(self.index,self.the_dyad);
    def get_extended_state(self):
        return getExtendedState(self.index,self.the_dyad);
    
    #update the agent with the actions and emotions taken by agent and client
    #there is no error checking here ... will fail badly if anything other than an action name is given
    def update(self,agent_action,agent_emotion,client_action,client_emotion):
        numnodes = updateState(self.index, self.the_dyad, "", client_action, client_emotion, agent_action, agent_emotion, self.use_planning)
        if numnodes < 0:
            print("failure - emotion label not found - state did not update")

    # get the current most likely identities for agent and client
    # could make a more general version here that would return a ranked order list of possibilities
    def get_identities(self):
        (self.agent_id, self.client_id) = getIdentities(self.index,self.the_dyad)
        return (self.agent_id, self.client_id)

    ## TODO: add functions to
    ## get client action and emotion  (default at least)
    ## select different dyads for simulation
    ## more functions to access the state (e.g. identity distributions instead of point estimates)
    ## ...

#test section
if __name__ == "__main__":

    # create a new agent group based on a testsim.txt file. Default settings for number of samples
    # are defined above, but you can also define them here as arguments to this function
    my_group = agentGroup(simfile="testsim.txt", num_samples=100)

    the_seed=my_group.get_random_seed()
    print("seed to gsl random number generator is : "+str(the_seed))
    done = False
    iter=0
    while not done:
        # ask agent what it will do next
        (agent_action, agent_emotion) = my_group.get_action_emotion()
        print("agent's action is to: ",agent_action)
        print("agent's emotion is: ",agent_emotion)
        
        # request client action and emotion from keyboard input (could be from a file)
        client_action = input("your action (from dictionary, enter to stop play): ")
        if client_action=="":
            break 
        client_emotion = input("your emotion (label from dictionary, enter to stop play): ")
        if client_emotion=="":
            break
        
        # actually perform the update
        my_group.update(client_action,client_emotion,agent_action,agent_emotion)

        # get some information
        (agent_id,client_id) = my_group.get_identities()
        print("agent thinks it is a  :  "+agent_id)
        print("and thinks client is a: "+client_id)

        (defl_agent,defl_beh,defl_client)=my_group.get_deflections()
        print("deflection of agent: "+str(defl_agent));
        print("deflection of behaviour: "+str(defl_beh));
        print("deflection of client: "+str(defl_client));
        
        stateDictionary=my_group.get_state()
        print("deflection of agent: "+str(stateDictionary["agentdeflection"]));
        print("deflection of behaviour: "+str(stateDictionary["behaviourdeflection"]))
        print("deflection of client: "+str(stateDictionary["clientdeflection"]))
        numsamples=stateDictionary["numsamples"]
        print("number of samples: "+str(numsamples))
        print("fundamental sentiment=[",end="")
        f=stateDictionary["f"]
        for v in f:
            print(str(v)+", ",end="")
        print("]")
        print("transient impression=[",end="")
        tau=stateDictionary["tau"]
        for v in tau:
            print(str(v)+", ",end="")
        print("]")
        print("x=[",end="")
        x=stateDictionary["x"]
        for v in x:
            print(str(v)+", ",end="")
        print("]")
        print("weight=[",end="")
        weight=stateDictionary["weight"]
        for v in weight:
            print(str(v)+", ",end="")
        print("]")
        
        print("actor=[",end="")
        actor=stateDictionary["actor"]
        for v in actor:
            print(str(v)+", ",end="")
        print("]")
        
        print("reset=[",end="")
        reset=stateDictionary["reset"]
        for v in reset:
            print(str(v)+", ",end="")
        print("]")
        

        extendedStateDictionary=my_group.get_extended_state()
        print("agent emotion=[",end="")
        f=extendedStateDictionary["agent_emot"]
        for v in f:
            print(str(v)+", ",end="")
        print("]")
        
        print("client emotion=[",end="")
        f=extendedStateDictionary["client_emot"]
        for v in f:
            print(str(v)+", ",end="")
        print("]")
        
        print("f_avg=[",end="")
        f=extendedStateDictionary["f_avg"]
        for v in f:
            print(str(v)+", ",end="")
        print("]")
        
        print("tau_avg=[",end="")
        f=extendedStateDictionary["tau_avg"]
        for v in f:
            print(str(v)+", ",end="")
        print("]")
        
        print("client_ids=[",end="")
        f=extendedStateDictionary["client_ids"]
        for v in f:
            print("("+str(v[0])+", "+str(v[1])+"), ",end="")
        print("]")
        
        print("agent_ids=[",end="")
        f=extendedStateDictionary["agent_ids"]
        for v in f:
            print("("+str(v[0])+", "+str(v[1])+"), ",end="")
        print("]")
        

    my_group.freeGroup()
