/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#*************************************************************************************************
#                           ***********************************                                  #
#************************************************************************************************/

#include "fvarsample.h"

#define MAX_ITERATIONS 50

int main(int argc, char **argv)
{

  // ================================================================================================
  // basic parameters
  // ================================================================================================
  // the basic parameters for a simulation
  // alpha - strength of the affect control principle
  double alpha=1.0;
  // beta for agent and client - inertial strength 
  double betaa=0.001;
  double betac=0.001;

  //inertial strength is very large if mimicking interact
  // this means the identities stay very much the same
#ifdef MIMIC_INTERACT
  alpha=1.0;   // not sure if this should be very small as well - seems to work some of the time, not always???
  betaa=0.0001;
  betac=0.0001;
#endif

  
  // number of samples to use for the particle filter
  // more means more precise estimates
  int numsamples=500;
  // make very large to mimic interact
#ifdef MIMIC_INTERACT
  numsamples=10000;
#endif

  readSimulationInput("testsim.txt");
  
  // ================================================================================================
  // files for dynamics for transient impression formation and emotions
  // ================================================================================================
  // dynamics file
  char *dyn_fname = "tdynamics-male.dat";
  // modifiers dynamics file
  // set to null if not using emotions
  char *mod_fname = NULL;
  mod_fname = "temotions-male.dat";

  // ================================================================================================
  // read in dictionaries for identities, behaviours and modifiers
  // ================================================================================================
  //identity dictionary
  dictionary id_dictionary;
  int num_identities;
  num_identities = readSentimentDictionary(COV,"identityEPAandCOV.csv",&id_dictionary);
#ifdef MIMIC_INTERACT
  // we use identities with fixed, tiny variances if mimicking interact
  // alternate - interact dictionary format can also be used
  num_identities = readSentimentDictionary(MEAN,"fidentities.dat",&id_dictionary);
#endif
  printf("num identities: %d\n",num_identities);

  // behaviour dictionary
  dictionary beh_dictionary;
  int num_behaviours;
  num_behaviours = readSentimentDictionary(COV,"behaviourEPAandCOV.csv",&beh_dictionary);
  // for PDBOT
  //num_behaviours = readSentimentDictionary(COV,"pdbehaviours.csv",&beh_dictionary);
#ifdef MIMIC_INTERACT
  // we use identities with fixed, tiny variances if mimicking interact
  // alternate - interact dictionary format can also be used
  num_identities = readSentimentDictionary(MEAN,"fbehaviours.dat",&beh_dictionary);
#endif
  printf("num behaviours: %d\n",num_behaviours);


  // modifier dictionary
  dictionary mod_dictionary;
  int num_modifiers;
  num_modifiers = readSentimentDictionary(COV,"modifierEPAandCOV.csv",&mod_dictionary);
#ifdef MIMIC_INTERACT
  // we use identities with fixed, tiny variances if mimicking interact
  // alternate - interact dictionary format can also be used
  num_modifiers = readSentimentDictionary(MEAN,"fmodifiers.dat",&mod_dictionary);
#endif
  printf("num modifiers: %d\n",num_modifiers);

  // ================================================================================================
  // set up intial agent and client identities
  // ================================================================================================
  // create a distribution over some identities
  // here we use 2 identities, but this could be an arbitrary number
  int num_client_ids=3;
  int num_agent_ids=2;
  identity * agent_id;
  identity * client_id;
  agent_id = (identity *) malloc(num_agent_ids*sizeof(identity));
  client_id = (identity *) malloc(num_client_ids*sizeof(identity));
  agent_id[0].label="mother";   agent_id[0].prob=0.2;
  agent_id[1].label="secretary";  agent_id[1].prob=0.8;
  client_id[0].label="playboy";   client_id[0].prob=0.1;
  client_id[1].label="executive";  client_id[1].prob=0.2;
  client_id[2].label="boss";  client_id[2].prob=0.7;
    /* FOR PDBOT
  agent_id[0].label="friend";   agent_id[0].prob=0.7;
  agent_id[1].label="scrooge";  agent_id[1].prob=0.3;
  client_id[0].label="scrooge";   client_id[0].prob=0.3;
  client_id[1].label="friend";  client_id[1].prob=0.7;
    */
  // to have an unknown identity use the following
  //num_client_ids=-1;
  //client_id=NULL;

  // look up these labels to fill in the indices in the dictionary
  // don't do any error checking here - what if they are not in the dictionary?
  for (int i=0; i<num_client_ids; i++)
    lookUp(&client_id[i],&id_dictionary);
  for (int i=0; i<num_agent_ids; i++)
    lookUp(&agent_id[i],&id_dictionary);

  // ================================================================================================
  // set up initial turn state for dyad
  // ================================================================================================
  denotativeState dyad_turn[2];
  // turn at time t' is the one to act next
  dyad_turn[0].label="agent";  dyad_turn[0].index=AGENT;  dyad_turn[0].prob=0.9;   // this means agent is next
  dyad_turn[1].label="client"; dyad_turn[1].index=CLIENT; dyad_turn[1].prob=1.0-dyad_turn[0].prob;
  
  
  denotativeState dyad2_turn[2];
  // turn at time t' is the one to act next
  dyad2_turn[0].label="agent";  dyad2_turn[0].index=AGENT;  dyad2_turn[0].prob=0.1;   // this means client is next
  dyad2_turn[1].label="client"; dyad2_turn[1].index=CLIENT; dyad2_turn[1].prob=1.0-dyad2_turn[0].prob;
  
  
  // ================================================================================================
  // initialize the sampler state and agent
  // ================================================================================================
  printf("**************************** starting BayesACT\n");
  samplerState *sstate = initializeBayesACT(numsamples,alpha,betaa,betac,VD);

  //set up an interaction between agent and client
  int dyad_index = setUpDyadicInteraction(sstate, dyn_fname, mod_fname, agent_id, num_agent_ids, client_id, num_client_ids, dyad_turn, &id_dictionary);

  // we will model each interaction using a different dyad index, 
  // i.e. if we have sally, tom, and dave, then we will have 6 total "Dyadic Interactions" created by setUpDyadicInteraction
  // sally-tom, sally-dave are sally's interecations with tom and dave
  // etc..
  // then we can do BayesACT of Self as a separate set of functions that groups these interactions into agents
  // i.e. sally will be an agent with two interactions : sally-tom, sally-dave
  // although we will need some way of re-initializaing sally's identity of self (or modifying it at least) to take into
  // account carry-over from one to the next. i.e. if Sally is interacting with Tom for a while, her f_a will change
  // then she starts interacting with Dave, she "loads up" her last used model of Dave f_c, and her last used model of herself
  // when interacting with Dave, f_a, but then modifies this f_a based on the f_a she had upon leaving her interaction with Tom
  // Also, if Sally observes Dave and Tom interacting, she may modify her models of each of them.  Thus,
  // thus, the agent Sally could have a third interaction for Dave-Tom,
  // BAsically what this means is I need to also have the situational self sentiment and fundamental self-sentiment implemented!
  int dyad2_index = setUpDyadicInteraction(sstate, dyn_fname, mod_fname, client_id, num_client_ids, agent_id, num_agent_ids, dyad2_turn, &id_dictionary);

  // this means we need a general "mixing" function that takes two sets of samples (A and B) and merges them with some weights (w_A, w_B).
  // This will be easy to write - simply take two unweighted sets, change all the weights in set A,B to w_A, w_B respectively, and then re-sample
  
  // ================================================================================================
  // some bookkeeping and settings for the simulation
  // ================================================================================================
  // some buffers for behaviours 
  char input_buffer[MAX_LABEL_LENGTH];
  behaviour the_behaviours[2];
  for (int j=0; j < 2 ; j++) 
    the_behaviours[j].label=(char *) malloc(MAX_LABEL_LENGTH*sizeof(char));
  // need to have these indices so we can easily assign NULL
  int acted[2];
  
  // other parameters of the simulation
  // we will iterate over these two 
  agent_type turn_iter[2]={AGENT,CLIENT};
  // these will eventually become properties of "Agent"
  char  * agent_names[2]={"sally","bob"};
  double totalweight;
  int iteration=0;
  int num_potential_actions, best_action_from_average;

  // number of best actions to show
  int num_to_show=5;
  // if 0, default behaviours are taken, else behaviours are queried from the user
  int query_user=1;
  int gotit;
  identity * beh_list;
  double *emot = (double *) malloc(SSD*sizeof(double));
  int index;
  

  // ================================================================================================
  // main simulation loop
  // ================================================================================================
  while (iteration < MAX_ITERATIONS) {
    printf("***************************************************************** iteration %d\n",iteration);
    printf("state:\n");
    printSamples(sstate, dyad_index, &id_dictionary);
    printf("------------------------------------------------------------------\n");
    printSamples(sstate, dyad2_index, &id_dictionary);
    printf("------------------------------------------------------------------\n");

    // compute deflections
    printf("deflections:\n");

    printDeflections(sstate, dyad_index);
    printf("------------------------------------------------------------------\n");
    
    printf("emotions:\n");
    for (int tt=0; tt<2; tt++) {
      computeExpectedEmotion(sstate, dyad_index, emot, turn_iter[tt]);
      // look up labels - still have to read in the emotions dictionary
      somaticTransformNEW(DENOTATIVE, sstate->r, dyad_index, &index, emot, &mod_dictionary);
      // printf("emotion for %s:\n",turn_labels[turn_iter[tt]]);
      printf("emotion for %s:\n",agent_names[tt]);
      for (int j=0; j<SSD; j++)
	printf("%4.3g ",emot[j]);
      printf("  label: %s\n",mod_dictionary.theDictionary[index].label);
    }
    printf("------------------------------------------------------------------\n");
    printf("Getting next actions:\n");

    // predict next default action by sampling from the next state and taking the behaviour
    // then sampling from x_b given f_b with a DENOTATIVE  somatic transform
    for (int tt=0; tt<2; tt++) {
      beh_list=NULL;
      printf("....................................................\n");
      //printf("potential next actions for %s\n",turn_labels[turn_iter[tt]]);
      printf("potential next actions for %s\n",agent_names[tt]);
      num_potential_actions = getNextAction(sstate, dyad_index, &beh_dictionary, &id_dictionary, &beh_list, turn_iter[tt]);
      printf("got %d potential next actions, top %d are:\n",num_potential_actions,num_to_show);
      printf("rank\t*\tlabel (probability)\n");
      printf("____________________________________________________________________\n");

      //show top 5 - beh_list is sorted on return from getNextAction
      for (int i=0; i< MIN(num_to_show,num_potential_actions); i++)  
	printf("%d\t*\t%s (%4.3g)\n",i+1,beh_dictionary.theDictionary[beh_list[i].index].label,beh_list[i].prob);
      // show best behaviour to take
      printf("------------------------------------------------------------------\n");
      // this will mimic what Interact computes 
      best_action_from_average = getNextActionFromAverageState(sstate, dyad_index, &beh_dictionary, &id_dictionary, &beh_list, turn_iter[tt]);
      if (best_action_from_average >= 0) 
	printf("best next action from average state: %s\n",beh_dictionary.theDictionary[best_action_from_average].label);
      printf("\n");
      
      //can assign the best one like this:
      acted[tt]=0;
      if (num_potential_actions > 0)  {
	if (!query_user) {
	  the_behaviours[tt].index = beh_list[0].index;
	  the_behaviours[tt].label = beh_dictionary.theDictionary[beh_list[0].index].label;
	  acted[tt]=1;
	} else {
	  // or  query the user for which behaviour to take
	  gotit=-1;
	  while (gotit<0) {
	    // printf("behaviour to take for %s (enter for none):  ",turn_labels[turn_iter[tt]]);
	    printf("behaviour to take for %s (enter for none):  ",agent_names[tt]); 
	    fgets(input_buffer,MAX_LABEL_LENGTH,stdin);
	    // remove trailing newline
	    input_buffer[strcspn(input_buffer, "\n")] = 0;
	    if (strlen(input_buffer)==0) {
	      if (tt==1 && !acted[0]) {
		//printf("you said %s didn't act, so %s must! try again\n",turn_labels[turn_iter[0]],turn_labels[turn_iter[1]]);
		printf("you said %s didn't act, so %s must! try again\n",agent_names[0],agent_names[1]);
	      } else {
		gotit=1;
	      }
	    } else {
	      strcpy(the_behaviours[tt].label,input_buffer);
	      gotit=lookUp(&the_behaviours[tt],&beh_dictionary);
	      if (gotit>=0) 
		acted[tt]=1;
	    }
	  }
	}
	if (acted[tt]) {
	  //printf("Behaviour taken by %s is : %s\n", turn_labels[turn_iter[tt]], the_behaviours[tt].label);
	  printf("Behaviour taken by %s is : %s\n", agent_names[tt], the_behaviours[tt].label);
	  //printf("best index %d for behaviour label %s\n",the_behaviours[tt].index,the_behaviours[tt].label);
	} else {
	  printf("%s doesn't act\n",turn_labels[turn_iter[tt]]);
	}
      }
      free(beh_list);
    }

    // The state update function. If either actor has not acted, their action is NULL and will not be performed
    // the samples with that NULL actor as turn will be discarded
    totalweight = updateFullStateTwoActors(sstate, dyad_index, &beh_dictionary, &id_dictionary,
					   (acted[0] ? &the_behaviours[0] : NULL),
					   (acted[1] ? &the_behaviours[1] : NULL));
    totalweight = updateFullStateTwoActors(sstate, dyad2_index, &beh_dictionary, &id_dictionary,
					   (acted[1] ? &the_behaviours[1] : NULL),
					   (acted[0] ? &the_behaviours[0] : NULL));
    iteration++;
  }
  
  // ================================================================================================
  // clean up
  // ================================================================================================
  free(agent_id);
  free(client_id);
  for (int j=0; j < 2 ; j++) 
    free(the_behaviours[j].label);
  free(emot);
  freeSentimentDictionary(&id_dictionary);
  freeSentimentDictionary(&beh_dictionary);
  freeSentimentDictionary(&mod_dictionary);

  freeParameters(sstate, dyad_index);
  //freeParameters(sstate, dyad2_index);
  freeBayesACT(sstate);
}
